# GENERATED FILE, DO NOT EDIT.
# Run './gen-api-data.sh > api_data.py' instead.

API_URL = "https://invalid-test-only.archive.softwareheritage.org/api/1"

API_DATA = {
    "content/sha1_git:fe95a46679d128ff167b7c55df5d02356c5a1ae1/": r"""
{
  "length": 151810,
  "status": "visible",
  "checksums": {
    "sha1_git": "fe95a46679d128ff167b7c55df5d02356c5a1ae1",
    "sha1": "dc2830a9e72f23c1dfebef4413003221baa5fb62",
    "blake2s256": "791e07fcea240ade6dccd0a9309141673c31242cae9c237cf3855e151abc78e9",
    "sha256": "b5c7fe0536f44ef60c8780b6065d30bca74a5cd06d78a4a71ba1ad064770f0c9"
  },
  "data_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:fe95a46679d128ff167b7c55df5d02356c5a1ae1/raw/",
  "filetype_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:fe95a46679d128ff167b7c55df5d02356c5a1ae1/filetype/",
  "language_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:fe95a46679d128ff167b7c55df5d02356c5a1ae1/language/",
  "license_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:fe95a46679d128ff167b7c55df5d02356c5a1ae1/license/"
}
    """,  # NoQA: B950  # NoQA: B950
    "directory/977fc4b98c0e85816348cebd3b12026407c368b6/": r"""
[
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "58471109208922c9ee8c4b06135725f03ed16814",
    "name": ".bzrignore",
    "perms": 33188,
    "status": "visible",
    "length": 582,
    "checksums": {
      "sha1_git": "58471109208922c9ee8c4b06135725f03ed16814",
      "sha1": "e2d79ae437210941840f49966497cc348c7e817f",
      "sha256": "2b7001f4819e898776b45b2fa3411018b7bc24e38afbb351691c32508eb2ae5d"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:58471109208922c9ee8c4b06135725f03ed16814/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "2106da61725973b81a63a817ec6f245706af4353",
    "name": ".codecov.yml",
    "perms": 33188,
    "status": "visible",
    "length": 453,
    "checksums": {
      "sha1_git": "2106da61725973b81a63a817ec6f245706af4353",
      "sha1": "f47aabb47381119cf72add7633bc095ca2cd030d",
      "sha256": "4f0475fac23bcd3ebceceecffb0d4facc5a413f6d9a0287185fb75638b8e9c69"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:2106da61725973b81a63a817ec6f245706af4353/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "82694d81f276b2c59a0a93a4f678e1852e625052",
    "name": ".gitattributes",
    "perms": 33188,
    "status": "visible",
    "length": 255,
    "checksums": {
      "sha1_git": "82694d81f276b2c59a0a93a4f678e1852e625052",
      "sha1": "5daeb59981fe561fd5b4eb8761cd836a814ee540",
      "sha256": "15051f8c9b4cbfc18d9dd5cf7a4fa96fa8d7ff1143413b4d3d9f71f795de780b"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:82694d81f276b2c59a0a93a4f678e1852e625052/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "b266746952a70c852d4f760b431c28affc6853e2",
    "name": ".github",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/b266746952a70c852d4f760b431c28affc6853e2/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "f436777c7cf099c43c9f04ad31fe1551b4f6079d",
    "name": ".gitignore",
    "perms": 33188,
    "status": "visible",
    "length": 1433,
    "checksums": {
      "sha1_git": "f436777c7cf099c43c9f04ad31fe1551b4f6079d",
      "sha1": "4bc3fb4c5a1b998e258f3f83ca694fb20e25921b",
      "sha256": "ab427a61741fadbd0aa5eeb0f384dc74b5630fc878869d994a9c6d972d4854a9"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:f436777c7cf099c43c9f04ad31fe1551b4f6079d/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "eb19a6c88d28d05588db25d21525ee2e19c22666",
    "name": ".hgeol",
    "perms": 33188,
    "status": "visible",
    "length": 1060,
    "checksums": {
      "sha1_git": "eb19a6c88d28d05588db25d21525ee2e19c22666",
      "sha1": "f731523ff2ffbaf91d5f34000af8ec5875d5d63b",
      "sha256": "426003d52cd302a75d4d1bf56ce0a3aa5740d06dfa9267f4901cfd1b038b6169"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:eb19a6c88d28d05588db25d21525ee2e19c22666/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "68c607f2e8d420c8dfd0748efcd3b3b5447def16",
    "name": ".hgignore",
    "perms": 33188,
    "status": "visible",
    "length": 1358,
    "checksums": {
      "sha1_git": "68c607f2e8d420c8dfd0748efcd3b3b5447def16",
      "sha1": "173cafb7f9d6e034287f24f0a8353580c7f05e1e",
      "sha256": "191069e0e1aa07b0345c82d3d06226b6bc7b55b4eb83e5279b43a8139a8b6f73"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:68c607f2e8d420c8dfd0748efcd3b3b5447def16/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "8f51c2ced49aed46d8b480280b630ea4264c57c3",
    "name": ".hgtags",
    "perms": 33188,
    "status": "visible",
    "length": 8917,
    "checksums": {
      "sha1_git": "8f51c2ced49aed46d8b480280b630ea4264c57c3",
      "sha1": "f3e049dcf85963f453fd2efd3af819bc8e8eabcd",
      "sha256": "1d8f5495b39ace89c290f45383936127256f80d9789971fd35208aaf18ab7cb8"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:8f51c2ced49aed46d8b480280b630ea4264c57c3/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "482e241c4121b49c12d1fbca71cc7e89d9f30014",
    "name": ".mention-bot",
    "perms": 33188,
    "status": "visible",
    "length": 606,
    "checksums": {
      "sha1_git": "482e241c4121b49c12d1fbca71cc7e89d9f30014",
      "sha1": "23cbb97a62f86f68425e0b368eb824fca332c28e",
      "sha256": "4f09c1b2d9242f035d5b6340a5a441e1e94cfc5eec720a4c4ae8974752e4e6c0"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:482e241c4121b49c12d1fbca71cc7e89d9f30014/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "e7e8694530ca21a6d7a19da3fab687a3e9d79e9c",
    "name": ".travis.yml",
    "perms": 33188,
    "status": "visible",
    "length": 3512,
    "checksums": {
      "sha1_git": "e7e8694530ca21a6d7a19da3fab687a3e9d79e9c",
      "sha1": "6f1625beca4e860503213da4c929a656b6569187",
      "sha256": "3ffd00254adef8ce24f8cb44fffcd2d02e72004ae9ca8e381d2eac6705cd50e1"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:e7e8694530ca21a6d7a19da3fab687a3e9d79e9c/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "5717ef211e7c694471d41cd8626d46bbaf647c66",
    "name": "Doc",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/5717ef211e7c694471d41cd8626d46bbaf647c66/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "5dd6fc9dc09374506491247872c868eca111e256",
    "name": "Grammar",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/5dd6fc9dc09374506491247872c868eca111e256/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "cd8b4bafaede5f4f9167489b08f32be2c9cc1cdd",
    "name": "Include",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/cd8b4bafaede5f4f9167489b08f32be2c9cc1cdd/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "f5d0b39a0cdddb91a31a537052b7d8d31a4aa79f",
    "name": "LICENSE",
    "perms": 33188,
    "status": "visible",
    "length": 12773,
    "checksums": {
      "sha1_git": "f5d0b39a0cdddb91a31a537052b7d8d31a4aa79f",
      "sha1": "4dfc4478b1d5f7388b298fdfc06802485bdeae0c",
      "sha256": "efc0c7f2994f95be25f7ce63a865f870e278ff0aa4f5d52d84d4d6d2a98128bd"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:f5d0b39a0cdddb91a31a537052b7d8d31a4aa79f/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "175167e77c40211e439202598830c68aa9d45f86",
    "name": "Lib",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/175167e77c40211e439202598830c68aa9d45f86/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "a8eb01b6b98ad20fa8f80d5c19c36ad27e36a6f1",
    "name": "Mac",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/a8eb01b6b98ad20fa8f80d5c19c36ad27e36a6f1/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "12593a2b63ac937f37f78f6046d5d84fc78af171",
    "name": "Makefile.pre.in",
    "perms": 33188,
    "status": "visible",
    "length": 59057,
    "checksums": {
      "sha1_git": "12593a2b63ac937f37f78f6046d5d84fc78af171",
      "sha1": "8c40fb0448a05a3e22cd760f8761ddd08f6093fd",
      "sha256": "acb089f45ca086407c1cf2fad7ea0603788871f9653c8e4fc9a1909808775752"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:12593a2b63ac937f37f78f6046d5d84fc78af171/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "6d8ce40edd85daec3653450f3b9220c135acdaca",
    "name": "Misc",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/6d8ce40edd85daec3653450f3b9220c135acdaca/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "64ab9e8e8b0edcec311a6fd4da7a8c6f290cfdd7",
    "name": "Modules",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/64ab9e8e8b0edcec311a6fd4da7a8c6f290cfdd7/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "06e40820601a63a04fc9f9baa54c442dfbeb1a6c",
    "name": "Objects",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/06e40820601a63a04fc9f9baa54c442dfbeb1a6c/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "b4df9f86f76a7e10a7d98259980d4f5d5d1223a1",
    "name": "PC",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/b4df9f86f76a7e10a7d98259980d4f5d5d1223a1/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "1e837a7d204fd247fb9a105adaa758a6d63486d8",
    "name": "PCbuild",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/1e837a7d204fd247fb9a105adaa758a6d63486d8/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "75771c7c20fe7a121d596299c5440aef10c6f884",
    "name": "Parser",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/75771c7c20fe7a121d596299c5440aef10c6f884/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "3efbcc80237ab7c3d4eb5bf31c893ca6de88e747",
    "name": "Programs",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/3efbcc80237ab7c3d4eb5bf31c893ca6de88e747/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "8f0f7237dc78c4344badeb84c7a65e5b106faf26",
    "name": "Python",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/8f0f7237dc78c4344badeb84c7a65e5b106faf26/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "9c95815d9e9d91b8dae8e05d8bbc696fe19f796b",
    "name": "README.rst",
    "perms": 33188,
    "status": "visible",
    "length": 9325,
    "checksums": {
      "sha1_git": "9c95815d9e9d91b8dae8e05d8bbc696fe19f796b",
      "sha1": "1b80225a4c9c14af02eb89ad3ad5ba6a22c27c0b",
      "sha256": "4d7c3907715129e40bbea30608d922bb39864f232638af016082465fb9a384f3"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:9c95815d9e9d91b8dae8e05d8bbc696fe19f796b/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "dir",
    "target": "01cd5744570f4be57d9ece0e62b324ef5d0d29fb",
    "name": "Tools",
    "perms": 16384,
    "length": null,
    "target_url": "https://archive.softwareheritage.org/api/1/directory/01cd5744570f4be57d9ece0e62b324ef5d0d29fb/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "2a745e57466cae1af014f239e5b91a8fc1fd64d8",
    "name": "aclocal.m4",
    "perms": 33188,
    "status": "visible",
    "length": 10914,
    "checksums": {
      "sha1_git": "2a745e57466cae1af014f239e5b91a8fc1fd64d8",
      "sha1": "a9fb2499c15f934a064bf3e9a59ce0516db01676",
      "sha256": "b062dd22e0803d82b381149d4ce30e58c7a2d2037a03e062c3e51160f1334fd8"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:2a745e57466cae1af014f239e5b91a8fc1fd64d8/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "1f5c50c0d1529d50b94dc3533ca72a47f0fa5849",
    "name": "config.guess",
    "perms": 33261,
    "status": "visible",
    "length": 42856,
    "checksums": {
      "sha1_git": "1f5c50c0d1529d50b94dc3533ca72a47f0fa5849",
      "sha1": "3646b015d9a91423feaf2bac86fd2df7ab3564f6",
      "sha256": "47da33933b3825049bbc04871747a9598ce90fd45a438b6a8a58b74bf6d73a4d"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:1f5c50c0d1529d50b94dc3533ca72a47f0fa5849/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "d654d03cdcd2226a5d7584890717e674a8122f4f",
    "name": "config.sub",
    "perms": 33261,
    "status": "visible",
    "length": 35740,
    "checksums": {
      "sha1_git": "d654d03cdcd2226a5d7584890717e674a8122f4f",
      "sha1": "0a7de3085aaa4ea3655c4221574dc9e503035428",
      "sha256": "78e9582b4b87ebcd2d3e95db04b22469cd2a132fa78f9d336afdf851fe4c1bf9"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:d654d03cdcd2226a5d7584890717e674a8122f4f/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "c9340c68605483dd0dada2fb2c6ea315b7c74019",
    "name": "configure",
    "perms": 33261,
    "status": "visible",
    "length": 484897,
    "checksums": {
      "sha1_git": "c9340c68605483dd0dada2fb2c6ea315b7c74019",
      "sha1": "a1fd0fd26de1fa61d8479ee668133d2900f6f58b",
      "sha256": "da1c3a1b44c761b12e846de5f7a724891cf4f3e4dd3751f3ef0a78ea2d616a60"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:c9340c68605483dd0dada2fb2c6ea315b7c74019/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "e065ce554b4ec92b33fc8225aa45b0d3225bb98f",
    "name": "configure.ac",
    "perms": 33188,
    "status": "visible",
    "length": 160603,
    "checksums": {
      "sha1_git": "e065ce554b4ec92b33fc8225aa45b0d3225bb98f",
      "sha1": "0b3c337c28885b966d7be1b1d3d6f7c18bddea42",
      "sha256": "dacf2d5dc82ebb2452301de9637d09bdc467a4e774eb9062d2c0d4b74c57cee7"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:e065ce554b4ec92b33fc8225aa45b0d3225bb98f/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "0ec27bcd488da5cad6ead13d70accbdbc40d31ef",
    "name": "install-sh",
    "perms": 33261,
    "status": "visible",
    "length": 7122,
    "checksums": {
      "sha1_git": "0ec27bcd488da5cad6ead13d70accbdbc40d31ef",
      "sha1": "260ef3ac7ec99c50e8c45c82be49465d38f7bb89",
      "sha256": "92adf3cf516747616fbe142306be804c8a7a6da49b85e0e1763d0aa4465f9fb6"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:0ec27bcd488da5cad6ead13d70accbdbc40d31ef/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "0a3d59ef9ae602fa02405c7b7068efaf7b55b21b",
    "name": "pyconfig.h.in",
    "perms": 33188,
    "status": "visible",
    "length": 41198,
    "checksums": {
      "sha1_git": "0a3d59ef9ae602fa02405c7b7068efaf7b55b21b",
      "sha1": "f65fb4a417b2c3de4c38c8f5c036af9cf2727322",
      "sha256": "5a7daffa2605b620500924f62964df590f7ff562d12997937ed5305d979f9e3e"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:0a3d59ef9ae602fa02405c7b7068efaf7b55b21b/"
  },
  {
    "dir_id": "977fc4b98c0e85816348cebd3b12026407c368b6",
    "type": "file",
    "target": "6a05643838bb4ef64ae093f7a0f816a80d6f09f4",
    "name": "setup.py",
    "perms": 33188,
    "status": "visible",
    "length": 98894,
    "checksums": {
      "sha1_git": "6a05643838bb4ef64ae093f7a0f816a80d6f09f4",
      "sha1": "de26417bb7aa146bb6a9abfcfe820466e80f08fe",
      "sha256": "c8d512dfe55b7d9503ee84c750e10724c7528ed1beb0b84c1b9ae82e82344ad8"
    },
    "target_url": "https://archive.softwareheritage.org/api/1/content/sha1_git:6a05643838bb4ef64ae093f7a0f816a80d6f09f4/"
  }
]
    """,  # NoQA: B950  # NoQA: B950
    "release/b9db10d00835e9a43e2eebef2db1d04d4ae82342/": r"""
{
  "author": {
    "fullname": "Paul Tagliamonte <tag@pault.ag>",
    "name": "Paul Tagliamonte",
    "email": "tag@pault.ag"
  },
  "date": "2013-07-06T19:34:11-04:00",
  "id": "b9db10d00835e9a43e2eebef2db1d04d4ae82342",
  "name": "0.9.9",
  "message": "Tagging 0.9.9\n",
  "synthetic": false,
  "target": "e005cb773c769436709ca6a1d625dc784dbc1636",
  "target_type": "revision",
  "target_url": "https://archive.softwareheritage.org/api/1/revision/e005cb773c769436709ca6a1d625dc784dbc1636/"
}
    """,  # NoQA: B950  # NoQA: B950
    "revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/": r"""
{
  "id": "aafb16d69fd30ff58afdd69036a26047f3aebdc6",
  "author": {
    "fullname": "Nicolas Dandrimont <nicolas.dandrimont@crans.org>",
    "name": "Nicolas Dandrimont",
    "email": "nicolas.dandrimont@crans.org"
  },
  "date": "2014-08-18T18:18:25+02:00",
  "committer": {
    "fullname": "Nicolas Dandrimont <nicolas.dandrimont@crans.org>",
    "name": "Nicolas Dandrimont",
    "email": "nicolas.dandrimont@crans.org"
  },
  "committer_date": "2014-08-18T18:18:25+02:00",
  "type": "git",
  "directory": "9f2e5898e00a66e6ac11033959d7e05b1593353b",
  "message": "Merge branch 'master' into pr/584\n",
  "metadata": {},
  "synthetic": false,
  "parents": [
    {
      "id": "26307d261279861c2d9c9eca3bb38519f951bea4",
      "url": "https://archive.softwareheritage.org/api/1/revision/26307d261279861c2d9c9eca3bb38519f951bea4/"
    },
    {
      "id": "37fc9e08d0c4b71807a4f1ecb06112e78d91c283",
      "url": "https://archive.softwareheritage.org/api/1/revision/37fc9e08d0c4b71807a4f1ecb06112e78d91c283/"
    }
  ],
  "merge": true,
  "url": "https://archive.softwareheritage.org/api/1/revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/",
  "history_url": "https://archive.softwareheritage.org/api/1/revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/log/",
  "directory_url": "https://archive.softwareheritage.org/api/1/directory/9f2e5898e00a66e6ac11033959d7e05b1593353b/"
}
    """,  # NoQA: B950  # NoQA: B950
    "snapshot/6a3a2cf0b2b90ce7ae1cf0a221ed68035b686f5a/": r"""
{
  "id": "6a3a2cf0b2b90ce7ae1cf0a221ed68035b686f5a",
  "branches": {
    "refs/heads/master": {
      "target": "83c20a6a63a7ebc1a549d367bc07a61b926cecf3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/83c20a6a63a7ebc1a549d367bc07a61b926cecf3/"
    },
    "refs/heads/wiki": {
      "target": "71f667aeb5d02562f2fa0941ad91df69c474ff3b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/71f667aeb5d02562f2fa0941ad91df69c474ff3b/"
    },
    "refs/tags/dpkt-1.6": {
      "target": "7fc0fd582812af36064d1c85fe51e33227920479",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7fc0fd582812af36064d1c85fe51e33227920479/"
    },
    "refs/tags/dpkt-1.7": {
      "target": "0c9dbfbc0974ec8ac1d8253aa1092366a03633a8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0c9dbfbc0974ec8ac1d8253aa1092366a03633a8/"
    }
  },
  "next_branch": null
}
    """,  # NoQA: B950  # NoQA: B950
    "snapshot/cabcc7d7bf639bbe1cc3b41989e1806618dd5764/": r"""
{
  "id": "cabcc7d7bf639bbe1cc3b41989e1806618dd5764",
  "branches": {
    "HEAD": {
      "target": "refs/heads/master",
      "target_type": "alias",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d96d875ef5dd372f533059a44f98e92de9cf0d42/"
    },
    "refs/heads/master": {
      "target": "d96d875ef5dd372f533059a44f98e92de9cf0d42",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d96d875ef5dd372f533059a44f98e92de9cf0d42/"
    },
    "refs/pull/10/head": {
      "target": "51b8b4fb32271d39fbdd760397406177b2b0fd36",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/51b8b4fb32271d39fbdd760397406177b2b0fd36/"
    },
    "refs/pull/100/head": {
      "target": "02b5a6031ba7a8cbfde5d65ff9e13ecdbc4a92ca",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/02b5a6031ba7a8cbfde5d65ff9e13ecdbc4a92ca/"
    },
    "refs/pull/102/head": {
      "target": "c0cae3784a6535760fce053bb1df01caf6ea69cb",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c0cae3784a6535760fce053bb1df01caf6ea69cb/"
    },
    "refs/pull/103/head": {
      "target": "0131d367fa05d8bf72f2c34140ce64a9b98b5278",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0131d367fa05d8bf72f2c34140ce64a9b98b5278/"
    },
    "refs/pull/105/head": {
      "target": "f3dff45d89668f4bd81abe8cacc2f9d023eaf429",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f3dff45d89668f4bd81abe8cacc2f9d023eaf429/"
    },
    "refs/pull/106/head": {
      "target": "07dd7fb6d88b6399e657ea1173ab5fdcc7fa4eb0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/07dd7fb6d88b6399e657ea1173ab5fdcc7fa4eb0/"
    },
    "refs/pull/107/head": {
      "target": "f5f68e734eeeda2287905018f083114266c37ad8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f5f68e734eeeda2287905018f083114266c37ad8/"
    },
    "refs/pull/108/head": {
      "target": "24070a449a6302370e2f13a67a11b65a1f2927cd",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/24070a449a6302370e2f13a67a11b65a1f2927cd/"
    },
    "refs/pull/109/head": {
      "target": "13fc2374a648a23303b24d1bf31ff4f71f8cdffe",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/13fc2374a648a23303b24d1bf31ff4f71f8cdffe/"
    },
    "refs/pull/11/head": {
      "target": "bdb25f1ee3bae85433d94e9192f8a18a880b28ac",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bdb25f1ee3bae85433d94e9192f8a18a880b28ac/"
    },
    "refs/pull/110/head": {
      "target": "8f827efed6fb6c91dc5befee5470ce3743bfab9d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8f827efed6fb6c91dc5befee5470ce3743bfab9d/"
    },
    "refs/pull/111/head": {
      "target": "7060dbc9139b58ee0e64ee8f7991c374fb249d1d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7060dbc9139b58ee0e64ee8f7991c374fb249d1d/"
    },
    "refs/pull/112/head": {
      "target": "39399a7d467990e994517e9137dd74e0ecc06c37",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/39399a7d467990e994517e9137dd74e0ecc06c37/"
    },
    "refs/pull/113/head": {
      "target": "f2c7cb774e3e0a8152b1134015c245ba0abddfc6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f2c7cb774e3e0a8152b1134015c245ba0abddfc6/"
    },
    "refs/pull/114/head": {
      "target": "bba606880aa30ba4b74ec092ce4fc9ae1a596ce4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bba606880aa30ba4b74ec092ce4fc9ae1a596ce4/"
    },
    "refs/pull/115/head": {
      "target": "1cd7a834347a9dcb326d14cf00cb9921f235968a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1cd7a834347a9dcb326d14cf00cb9921f235968a/"
    },
    "refs/pull/118/head": {
      "target": "bedc5a601906dd207429737670d64dc0b2637c0f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bedc5a601906dd207429737670d64dc0b2637c0f/"
    },
    "refs/pull/12/head": {
      "target": "4a4dd4719381642ad07daa1bb7c596ae0144a90e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4a4dd4719381642ad07daa1bb7c596ae0144a90e/"
    },
    "refs/pull/120/head": {
      "target": "73911eb24265aa56d2172b5d4c8f5b8948a6a646",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/73911eb24265aa56d2172b5d4c8f5b8948a6a646/"
    },
    "refs/pull/121/head": {
      "target": "c82ab37950a2332e6ffd29ac017a6c20941909d2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c82ab37950a2332e6ffd29ac017a6c20941909d2/"
    },
    "refs/pull/122/head": {
      "target": "aeb978c975d7af8a83a9cee20384636f1a628684",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/aeb978c975d7af8a83a9cee20384636f1a628684/"
    },
    "refs/pull/123/head": {
      "target": "aeb978c975d7af8a83a9cee20384636f1a628684",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/aeb978c975d7af8a83a9cee20384636f1a628684/"
    },
    "refs/pull/124/head": {
      "target": "aeb978c975d7af8a83a9cee20384636f1a628684",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/aeb978c975d7af8a83a9cee20384636f1a628684/"
    },
    "refs/pull/125/head": {
      "target": "71db4931a2aaeed5a639e260669be0de8aee0353",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/71db4931a2aaeed5a639e260669be0de8aee0353/"
    },
    "refs/pull/126/head": {
      "target": "881c4c07dd99f1b0a8b9f96e326cc755597ea522",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/881c4c07dd99f1b0a8b9f96e326cc755597ea522/"
    },
    "refs/pull/127/head": {
      "target": "5ebb68c104192d4341c3241e0b0ff6868c05c62b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5ebb68c104192d4341c3241e0b0ff6868c05c62b/"
    },
    "refs/pull/128/head": {
      "target": "a3928246651b92dec93a47b99c49ff3151b08bae",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a3928246651b92dec93a47b99c49ff3151b08bae/"
    },
    "refs/pull/129/head": {
      "target": "592b1552af72edaf9c587cbbc81b86c418aa85db",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/592b1552af72edaf9c587cbbc81b86c418aa85db/"
    },
    "refs/pull/13/head": {
      "target": "9493ade38be89aff3edab2a17af733b7239af6a9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9493ade38be89aff3edab2a17af733b7239af6a9/"
    },
    "refs/pull/130/head": {
      "target": "a1d0b61186b1c2cab766a7cdb24f672a6523bbca",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a1d0b61186b1c2cab766a7cdb24f672a6523bbca/"
    },
    "refs/pull/131/head": {
      "target": "a1d0b61186b1c2cab766a7cdb24f672a6523bbca",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a1d0b61186b1c2cab766a7cdb24f672a6523bbca/"
    },
    "refs/pull/133/head": {
      "target": "a0c20a1cc664509ce90bba1c22770b82627d69d7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a0c20a1cc664509ce90bba1c22770b82627d69d7/"
    },
    "refs/pull/134/head": {
      "target": "a1d0b61186b1c2cab766a7cdb24f672a6523bbca",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a1d0b61186b1c2cab766a7cdb24f672a6523bbca/"
    },
    "refs/pull/135/head": {
      "target": "e2f5be8b3fe16c0689e754c63c1d01833ee7eb12",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e2f5be8b3fe16c0689e754c63c1d01833ee7eb12/"
    },
    "refs/pull/136/head": {
      "target": "60604ffc58d716f0d1ff8d908f9695f96acc9011",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/60604ffc58d716f0d1ff8d908f9695f96acc9011/"
    },
    "refs/pull/137/head": {
      "target": "44a265a9e675120570815a324639cc3d2a1de628",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/44a265a9e675120570815a324639cc3d2a1de628/"
    },
    "refs/pull/138/head": {
      "target": "664f7e64edd3d1ecbcf0c4739c15968bdb9a1c96",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/664f7e64edd3d1ecbcf0c4739c15968bdb9a1c96/"
    },
    "refs/pull/139/head": {
      "target": "fadbf30c8386d6644e59a083d39059770f318ca8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fadbf30c8386d6644e59a083d39059770f318ca8/"
    },
    "refs/pull/14/head": {
      "target": "8fdfc03177fd082d46254371794ff7d1fe0c49ab",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8fdfc03177fd082d46254371794ff7d1fe0c49ab/"
    },
    "refs/pull/140/head": {
      "target": "150dcb6b6679f47710a5139d5732c2ea5cdb8d93",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/150dcb6b6679f47710a5139d5732c2ea5cdb8d93/"
    },
    "refs/pull/141/head": {
      "target": "b53f1de56782d17581da5aea61d1f15a51b4c6ee",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b53f1de56782d17581da5aea61d1f15a51b4c6ee/"
    },
    "refs/pull/142/head": {
      "target": "d9eb0a4b5f2dd99442549868eed450c4d6d5c4ba",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d9eb0a4b5f2dd99442549868eed450c4d6d5c4ba/"
    },
    "refs/pull/143/head": {
      "target": "a64a847fcba291d1d31d88037a11895915b1e6f4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a64a847fcba291d1d31d88037a11895915b1e6f4/"
    },
    "refs/pull/144/head": {
      "target": "f28e39a8f514fbd9a5ea78bda6f3d9a928b648c1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f28e39a8f514fbd9a5ea78bda6f3d9a928b648c1/"
    },
    "refs/pull/145/head": {
      "target": "1079ac58f26ffaad462ef62913a4033f5942d80b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1079ac58f26ffaad462ef62913a4033f5942d80b/"
    },
    "refs/pull/146/head": {
      "target": "04d9765bd22ed96c8f05949d02fcdd60ce30b717",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/04d9765bd22ed96c8f05949d02fcdd60ce30b717/"
    },
    "refs/pull/147/head": {
      "target": "6b4e10d9bd2cac919b7ee4af5e7e5895fe06a3de",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6b4e10d9bd2cac919b7ee4af5e7e5895fe06a3de/"
    },
    "refs/pull/148/head": {
      "target": "05de0a7fb8ff5cf0808298d8df1aa047371baecc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/05de0a7fb8ff5cf0808298d8df1aa047371baecc/"
    },
    "refs/pull/149/head": {
      "target": "6409caa0329ea015b2de5a7fdd4b5ee040024aea",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6409caa0329ea015b2de5a7fdd4b5ee040024aea/"
    },
    "refs/pull/150/head": {
      "target": "3a16c8ec95ab69828e7865c630d9b58d84388fd7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3a16c8ec95ab69828e7865c630d9b58d84388fd7/"
    },
    "refs/pull/151/head": {
      "target": "d851313125143b696e546c77899cc7f2450875f4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d851313125143b696e546c77899cc7f2450875f4/"
    },
    "refs/pull/152/head": {
      "target": "d851313125143b696e546c77899cc7f2450875f4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d851313125143b696e546c77899cc7f2450875f4/"
    },
    "refs/pull/153/head": {
      "target": "56731b16ce5a9f11cb9496d76e4ffac11987fc85",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/56731b16ce5a9f11cb9496d76e4ffac11987fc85/"
    },
    "refs/pull/154/head": {
      "target": "51e0b71cd22ca9063c79197fcd7a1f46a95802e0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/51e0b71cd22ca9063c79197fcd7a1f46a95802e0/"
    },
    "refs/pull/155/head": {
      "target": "52d50ca4efe09f5bbbdf48f65da637cdb6585497",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/52d50ca4efe09f5bbbdf48f65da637cdb6585497/"
    },
    "refs/pull/156/head": {
      "target": "69f9f9a9375d410d258fb8757a40dcc5f877c01f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/69f9f9a9375d410d258fb8757a40dcc5f877c01f/"
    },
    "refs/pull/157/head": {
      "target": "2eb19958989e185726cb12d862249c2b0a56a507",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2eb19958989e185726cb12d862249c2b0a56a507/"
    },
    "refs/pull/158/head": {
      "target": "9e5dd3cb71aa61ab172fca48f7a0132efc16a991",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9e5dd3cb71aa61ab172fca48f7a0132efc16a991/"
    },
    "refs/pull/160/head": {
      "target": "684c5f18e43b200d07f8aa347ae0836d8d1e4076",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/684c5f18e43b200d07f8aa347ae0836d8d1e4076/"
    },
    "refs/pull/161/head": {
      "target": "3de03ff63cc9babac6e50e587f04501f7b68459e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3de03ff63cc9babac6e50e587f04501f7b68459e/"
    },
    "refs/pull/162/head": {
      "target": "3eb5c27ade53a726d3de3de932e21077f85f3d22",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3eb5c27ade53a726d3de3de932e21077f85f3d22/"
    },
    "refs/pull/163/head": {
      "target": "a05af7b0837f131faf86618dbab07bb7dd6e2c3d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a05af7b0837f131faf86618dbab07bb7dd6e2c3d/"
    },
    "refs/pull/164/head": {
      "target": "4a1e0249887c3f85d82ad26c9e06fd276675e5ff",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4a1e0249887c3f85d82ad26c9e06fd276675e5ff/"
    },
    "refs/pull/165/head": {
      "target": "2d98265865ac4e41c496a6e0cde6a1d7db710445",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2d98265865ac4e41c496a6e0cde6a1d7db710445/"
    },
    "refs/pull/166/head": {
      "target": "2f1b2be645fd3d46760086af4085811215eb696a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2f1b2be645fd3d46760086af4085811215eb696a/"
    },
    "refs/pull/167/head": {
      "target": "24c260dba7fcaecd695898010df87bdf4ff210ec",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/24c260dba7fcaecd695898010df87bdf4ff210ec/"
    },
    "refs/pull/168/head": {
      "target": "c5dec4b63f38d2ea78b66f2db2081cfa12642a51",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c5dec4b63f38d2ea78b66f2db2081cfa12642a51/"
    },
    "refs/pull/169/head": {
      "target": "0b1553a8e85d4a925ea24e46337c451a42ee1b53",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0b1553a8e85d4a925ea24e46337c451a42ee1b53/"
    },
    "refs/pull/17/head": {
      "target": "757a6e01c6756040543e431147b7e2e30f70cd84",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/757a6e01c6756040543e431147b7e2e30f70cd84/"
    },
    "refs/pull/170/head": {
      "target": "ee4821efc296f901b94427c989308edf52c333c1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ee4821efc296f901b94427c989308edf52c333c1/"
    },
    "refs/pull/171/head": {
      "target": "7663a33d8dce0e0e6947e6bcf5bf596454a15e58",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7663a33d8dce0e0e6947e6bcf5bf596454a15e58/"
    },
    "refs/pull/172/head": {
      "target": "e9afb23f7e96b93a90dd6775811ec3fb5eafc62f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e9afb23f7e96b93a90dd6775811ec3fb5eafc62f/"
    },
    "refs/pull/173/head": {
      "target": "fbed5b96e0d0bd09dcb569c65d6b2c5289c526cf",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fbed5b96e0d0bd09dcb569c65d6b2c5289c526cf/"
    },
    "refs/pull/174/head": {
      "target": "d66f25668a4aab37a38fcdcee684edc59e4e102d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d66f25668a4aab37a38fcdcee684edc59e4e102d/"
    },
    "refs/pull/175/head": {
      "target": "fdb0ef4a3d227e186109eda5798ecc48250fb8d3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fdb0ef4a3d227e186109eda5798ecc48250fb8d3/"
    },
    "refs/pull/176/head": {
      "target": "30688a33314fbfc3c3c315a9772fe55eb0121bb6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/30688a33314fbfc3c3c315a9772fe55eb0121bb6/"
    },
    "refs/pull/177/head": {
      "target": "78914a82c618a08138766b7434579e47a6f2b3fd",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/78914a82c618a08138766b7434579e47a6f2b3fd/"
    },
    "refs/pull/178/head": {
      "target": "78914a82c618a08138766b7434579e47a6f2b3fd",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/78914a82c618a08138766b7434579e47a6f2b3fd/"
    },
    "refs/pull/179/head": {
      "target": "91404d21fe8fafb2d68aa10ceb2c7d94ec53ebc0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/91404d21fe8fafb2d68aa10ceb2c7d94ec53ebc0/"
    },
    "refs/pull/18/head": {
      "target": "bdd8b8788c10911daa44aa4f499b1ac3228b998d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bdd8b8788c10911daa44aa4f499b1ac3228b998d/"
    },
    "refs/pull/180/head": {
      "target": "24475d4ab37d8db3ad36e416d199989ba53fbf3d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/24475d4ab37d8db3ad36e416d199989ba53fbf3d/"
    },
    "refs/pull/181/head": {
      "target": "997a78b4ee73263b0c9ad1c6cddc6de5f879cfc9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/997a78b4ee73263b0c9ad1c6cddc6de5f879cfc9/"
    },
    "refs/pull/182/head": {
      "target": "c3902fe5fe7f50de16edaa69b93ea08a333e63af",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c3902fe5fe7f50de16edaa69b93ea08a333e63af/"
    },
    "refs/pull/183/head": {
      "target": "e46811c308d7175ae1ff384b363b582d1bf5092c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e46811c308d7175ae1ff384b363b582d1bf5092c/"
    },
    "refs/pull/184/head": {
      "target": "02acfc45baac94f0e99101feb1d24a792325039f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/02acfc45baac94f0e99101feb1d24a792325039f/"
    },
    "refs/pull/185/head": {
      "target": "ab7937e354e778aa82a7a70685ff3d013da7d2c6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ab7937e354e778aa82a7a70685ff3d013da7d2c6/"
    },
    "refs/pull/186/head": {
      "target": "f266d6b3c848680787cf588f2dde61e1b29e8c03",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f266d6b3c848680787cf588f2dde61e1b29e8c03/"
    },
    "refs/pull/187/head": {
      "target": "b959c8d043bfb39fc0b686b3ca3d7319f78b417d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b959c8d043bfb39fc0b686b3ca3d7319f78b417d/"
    },
    "refs/pull/188/head": {
      "target": "bb67532fc8107d5322bc546e2d6c9c63a8043e5f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bb67532fc8107d5322bc546e2d6c9c63a8043e5f/"
    },
    "refs/pull/189/head": {
      "target": "cbeb22aa359aeb2114e83d3edf03f89e36ff2a41",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cbeb22aa359aeb2114e83d3edf03f89e36ff2a41/"
    },
    "refs/pull/19/head": {
      "target": "abf6eb7690b40a2d9f3fd09ec6e55ca994ff198a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/abf6eb7690b40a2d9f3fd09ec6e55ca994ff198a/"
    },
    "refs/pull/190/head": {
      "target": "0d14441e809b01cc4433d63af5b96a1412eb6c87",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0d14441e809b01cc4433d63af5b96a1412eb6c87/"
    },
    "refs/pull/191/head": {
      "target": "0fb23e45b83d71aca5be8d6ffb98dfcf4743c057",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0fb23e45b83d71aca5be8d6ffb98dfcf4743c057/"
    },
    "refs/pull/192/head": {
      "target": "a72670d6d5116b6ebe7aa4669c000299dbb9105e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a72670d6d5116b6ebe7aa4669c000299dbb9105e/"
    },
    "refs/pull/193/head": {
      "target": "4c7813d63a1e2d007a8a4b022684552955fa6c9f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4c7813d63a1e2d007a8a4b022684552955fa6c9f/"
    },
    "refs/pull/194/head": {
      "target": "99390cbe3fab46e2f3cd0d37db26aff7601b10aa",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/99390cbe3fab46e2f3cd0d37db26aff7601b10aa/"
    },
    "refs/pull/195/head": {
      "target": "e32495dae328ed92d802b2297c96941c2a782020",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e32495dae328ed92d802b2297c96941c2a782020/"
    },
    "refs/pull/196/head": {
      "target": "f528095730dccbe8c220c2f4dbc982a8c7d0977a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f528095730dccbe8c220c2f4dbc982a8c7d0977a/"
    },
    "refs/pull/197/head": {
      "target": "67944aac56642c37ade09c103f5247f62c997c65",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/67944aac56642c37ade09c103f5247f62c997c65/"
    },
    "refs/pull/198/head": {
      "target": "10779f9d2c10af545d2c11cd42df26d4a43e6100",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/10779f9d2c10af545d2c11cd42df26d4a43e6100/"
    },
    "refs/pull/199/head": {
      "target": "ff2803c3e6aa05cdc03a00c70a61a3652732b39b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ff2803c3e6aa05cdc03a00c70a61a3652732b39b/"
    },
    "refs/pull/20/head": {
      "target": "62cf68120f647cfd17773bc40f02c7b5f7f6f8a1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/62cf68120f647cfd17773bc40f02c7b5f7f6f8a1/"
    },
    "refs/pull/200/head": {
      "target": "c4e500ecb8eaf811a2fb2a3e77156763e1c0d4f2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c4e500ecb8eaf811a2fb2a3e77156763e1c0d4f2/"
    },
    "refs/pull/201/head": {
      "target": "bb187e2c8ae7dac4a261064843b28c83eceef13f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bb187e2c8ae7dac4a261064843b28c83eceef13f/"
    },
    "refs/pull/202/head": {
      "target": "af7f07c46cb21258940e1a3395fec858ff7e0e08",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/af7f07c46cb21258940e1a3395fec858ff7e0e08/"
    },
    "refs/pull/203/head": {
      "target": "018d0d25754a3ea80f4c9471f82fe6ba0b040605",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/018d0d25754a3ea80f4c9471f82fe6ba0b040605/"
    },
    "refs/pull/204/head": {
      "target": "7158f37c74295cce67337c3363142feed0ce1228",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7158f37c74295cce67337c3363142feed0ce1228/"
    },
    "refs/pull/205/head": {
      "target": "e4163e11d84befbfe9483516c8af9581b822bfc3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e4163e11d84befbfe9483516c8af9581b822bfc3/"
    },
    "refs/pull/206/head": {
      "target": "ed71e8a2fd8bd3116685d8e42f9582d138ee56f0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ed71e8a2fd8bd3116685d8e42f9582d138ee56f0/"
    },
    "refs/pull/207/head": {
      "target": "e71c1b9261d627677590f61273dddc0e87eb6a7e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e71c1b9261d627677590f61273dddc0e87eb6a7e/"
    },
    "refs/pull/208/head": {
      "target": "dbd279c51af4b1702aad7cdc13fcd08d486e0267",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/dbd279c51af4b1702aad7cdc13fcd08d486e0267/"
    },
    "refs/pull/209/head": {
      "target": "d73de37ebc4bc3372805ae0f6226b580d98f9f41",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d73de37ebc4bc3372805ae0f6226b580d98f9f41/"
    },
    "refs/pull/210/head": {
      "target": "500fed2454f56863f992feac4ba012608e0454ee",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/500fed2454f56863f992feac4ba012608e0454ee/"
    },
    "refs/pull/211/head": {
      "target": "1fd2a09fe7aa63e5f7993ccef5ccb6cf71dde189",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1fd2a09fe7aa63e5f7993ccef5ccb6cf71dde189/"
    },
    "refs/pull/212/head": {
      "target": "0416cd1e2fadffec525f29373cacd037d1beb6dd",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0416cd1e2fadffec525f29373cacd037d1beb6dd/"
    },
    "refs/pull/213/head": {
      "target": "ec60bf33d95ea879346d0e5b6919958f9ca45cd4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ec60bf33d95ea879346d0e5b6919958f9ca45cd4/"
    },
    "refs/pull/214/head": {
      "target": "8cd99fd908d9e6586a8e49560f505778e65bbc79",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8cd99fd908d9e6586a8e49560f505778e65bbc79/"
    },
    "refs/pull/215/head": {
      "target": "b307947c5c192112d0f5ed1ddda91236128d072d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b307947c5c192112d0f5ed1ddda91236128d072d/"
    },
    "refs/pull/216/head": {
      "target": "08026d71d52063920048e3e64921eed37b069663",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/08026d71d52063920048e3e64921eed37b069663/"
    },
    "refs/pull/217/head": {
      "target": "31a343aaad2e2cbda6cd5900552c267e5402c4e4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/31a343aaad2e2cbda6cd5900552c267e5402c4e4/"
    },
    "refs/pull/218/head": {
      "target": "e734cf01a2b76edf5bef7e84e60c2ab5c50baeb0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e734cf01a2b76edf5bef7e84e60c2ab5c50baeb0/"
    },
    "refs/pull/219/head": {
      "target": "7387ca73a084b1654d19d6d28983887b392d78b3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7387ca73a084b1654d19d6d28983887b392d78b3/"
    },
    "refs/pull/22/head": {
      "target": "9045893e49f9944c9683e13e4aa3554c522d1e4c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9045893e49f9944c9683e13e4aa3554c522d1e4c/"
    },
    "refs/pull/220/head": {
      "target": "12f7c7375d8c212590edfb42acdf821f3e507169",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/12f7c7375d8c212590edfb42acdf821f3e507169/"
    },
    "refs/pull/221/head": {
      "target": "185b564ab788e84bb7b07482f7fdbb816b209b46",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/185b564ab788e84bb7b07482f7fdbb816b209b46/"
    },
    "refs/pull/222/head": {
      "target": "bfb074d24e95205a59238d5ee9f005534d5556ed",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bfb074d24e95205a59238d5ee9f005534d5556ed/"
    },
    "refs/pull/223/head": {
      "target": "d056931baa28b4b4190e7f0fc09cc8cc9cd53751",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d056931baa28b4b4190e7f0fc09cc8cc9cd53751/"
    },
    "refs/pull/224/head": {
      "target": "e1cccb1030a475467992c42b5018e38fa829b327",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e1cccb1030a475467992c42b5018e38fa829b327/"
    },
    "refs/pull/225/head": {
      "target": "86eaf54d07f239243f674d6256e5c4be6578348a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/86eaf54d07f239243f674d6256e5c4be6578348a/"
    },
    "refs/pull/226/head": {
      "target": "59b8925b8431b0e76a2c6a2e263f9776328b719b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/59b8925b8431b0e76a2c6a2e263f9776328b719b/"
    },
    "refs/pull/227/head": {
      "target": "fcd57629536553e78a452900a6e135daeced6b64",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fcd57629536553e78a452900a6e135daeced6b64/"
    },
    "refs/pull/228/head": {
      "target": "6ffeba9607343f15303a399bc402a538800d89d9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6ffeba9607343f15303a399bc402a538800d89d9/"
    },
    "refs/pull/229/head": {
      "target": "fda8cf4dc8fc20edbdf07b7e46f31efaa41ab4a8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fda8cf4dc8fc20edbdf07b7e46f31efaa41ab4a8/"
    },
    "refs/pull/23/head": {
      "target": "131a80b81d4da2af6b5cad2221d6219ccb7378ac",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/131a80b81d4da2af6b5cad2221d6219ccb7378ac/"
    },
    "refs/pull/230/head": {
      "target": "54a91f7975ac30f38d7152f1ac6541c4a08b8319",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/54a91f7975ac30f38d7152f1ac6541c4a08b8319/"
    },
    "refs/pull/231/head": {
      "target": "327ef99473999967101c6ce9f4797d3ebd63fcb9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/327ef99473999967101c6ce9f4797d3ebd63fcb9/"
    },
    "refs/pull/232/head": {
      "target": "e1d3c76814c839a835286843f0bca33b0c5d0dd8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e1d3c76814c839a835286843f0bca33b0c5d0dd8/"
    },
    "refs/pull/233/head": {
      "target": "296ab4986347fdd64d510a224ecc1ac4935b2ea0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/296ab4986347fdd64d510a224ecc1ac4935b2ea0/"
    },
    "refs/pull/234/head": {
      "target": "03844599e65c44a70c932e3e314ee89e23cb6dd3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/03844599e65c44a70c932e3e314ee89e23cb6dd3/"
    },
    "refs/pull/235/head": {
      "target": "b5b648fa234268b795c9aec206715f9f972217f9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b5b648fa234268b795c9aec206715f9f972217f9/"
    },
    "refs/pull/236/head": {
      "target": "874e66cba6c70865c714b5075f1c88e3788e76ae",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/874e66cba6c70865c714b5075f1c88e3788e76ae/"
    },
    "refs/pull/237/head": {
      "target": "8ec7e7ddc1c60a93deac54ef377004348fd155c5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8ec7e7ddc1c60a93deac54ef377004348fd155c5/"
    },
    "refs/pull/238/head": {
      "target": "afe27eb758b5b56631c89afc64925f4e360cd068",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/afe27eb758b5b56631c89afc64925f4e360cd068/"
    },
    "refs/pull/239/head": {
      "target": "38c14a94cd0e08cc328f9a4e30e0b57ce7d2a75b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/38c14a94cd0e08cc328f9a4e30e0b57ce7d2a75b/"
    },
    "refs/pull/24/head": {
      "target": "94e68603c9bfe49189a49df35ff22706a68db458",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/94e68603c9bfe49189a49df35ff22706a68db458/"
    },
    "refs/pull/240/head": {
      "target": "514d8784c5e3353790a3bec007f9de56eb7fde70",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/514d8784c5e3353790a3bec007f9de56eb7fde70/"
    },
    "refs/pull/241/head": {
      "target": "3d35defb6737559404b343b211ca6b57081ddf67",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3d35defb6737559404b343b211ca6b57081ddf67/"
    },
    "refs/pull/242/head": {
      "target": "2f150ef213f4728b97a6bd94e52350362c7384a5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2f150ef213f4728b97a6bd94e52350362c7384a5/"
    },
    "refs/pull/243/head": {
      "target": "cb693f10ea4e55b6e74da212149ff40387cd1d74",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cb693f10ea4e55b6e74da212149ff40387cd1d74/"
    },
    "refs/pull/244/head": {
      "target": "edc5284b08147dd182a1a793dd341dde5af21433",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/edc5284b08147dd182a1a793dd341dde5af21433/"
    },
    "refs/pull/245/head": {
      "target": "aaae6004749a6648bf45cee160c85557c6f453c4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/aaae6004749a6648bf45cee160c85557c6f453c4/"
    },
    "refs/pull/246/head": {
      "target": "74e12c52fad6bff3f93630ff21caf72c68bfd458",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/74e12c52fad6bff3f93630ff21caf72c68bfd458/"
    },
    "refs/pull/247/head": {
      "target": "da6c038ea785aa57880eae4e9cf12d80572d8d6f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/da6c038ea785aa57880eae4e9cf12d80572d8d6f/"
    },
    "refs/pull/248/head": {
      "target": "18e073a82e767fe0564ce5f5887335634015d16b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/18e073a82e767fe0564ce5f5887335634015d16b/"
    },
    "refs/pull/249/head": {
      "target": "18e073a82e767fe0564ce5f5887335634015d16b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/18e073a82e767fe0564ce5f5887335634015d16b/"
    },
    "refs/pull/25/head": {
      "target": "732b1575d1de4c85b31c7376a4ca6bd901b5e2e1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/732b1575d1de4c85b31c7376a4ca6bd901b5e2e1/"
    },
    "refs/pull/250/head": {
      "target": "3b0073aefc5f32f595fa3372aef99eb4a1bd5ac4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3b0073aefc5f32f595fa3372aef99eb4a1bd5ac4/"
    },
    "refs/pull/251/head": {
      "target": "4ed9ed9cc6bcd89bd801f3f2cbee2b5d2aaf3573",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4ed9ed9cc6bcd89bd801f3f2cbee2b5d2aaf3573/"
    },
    "refs/pull/252/head": {
      "target": "f41de3b5d604e56ba81d641b9a4a38a27a6dd1dd",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f41de3b5d604e56ba81d641b9a4a38a27a6dd1dd/"
    },
    "refs/pull/253/head": {
      "target": "5d33f9d0a23f6b6c1e47dfd3d56a9567d088cdc0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5d33f9d0a23f6b6c1e47dfd3d56a9567d088cdc0/"
    },
    "refs/pull/254/head": {
      "target": "615b2aba9503d7646069de48ed1927a93af0d25e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/615b2aba9503d7646069de48ed1927a93af0d25e/"
    },
    "refs/pull/255/head": {
      "target": "0db4f35dbef25971707f0c153698e977e1d62c32",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0db4f35dbef25971707f0c153698e977e1d62c32/"
    },
    "refs/pull/256/head": {
      "target": "0abe6a8345fae002a1750806b369a67f9f1aadb6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0abe6a8345fae002a1750806b369a67f9f1aadb6/"
    },
    "refs/pull/257/head": {
      "target": "4f9a10bce7df56decbc3d2e80efa1e11a8980daf",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4f9a10bce7df56decbc3d2e80efa1e11a8980daf/"
    },
    "refs/pull/258/head": {
      "target": "9742a7da0dab6c35a95a8b6f3c54097acf756d67",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9742a7da0dab6c35a95a8b6f3c54097acf756d67/"
    },
    "refs/pull/259/head": {
      "target": "5a46d4b6983db7039c71c244dd214dda034be550",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5a46d4b6983db7039c71c244dd214dda034be550/"
    },
    "refs/pull/26/head": {
      "target": "cf827bb4057fdcdaeac20b95b01c29c068109185",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cf827bb4057fdcdaeac20b95b01c29c068109185/"
    },
    "refs/pull/260/head": {
      "target": "b3c9e2b802f9bea33bc9f1c4775c645604ec8afc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b3c9e2b802f9bea33bc9f1c4775c645604ec8afc/"
    },
    "refs/pull/261/head": {
      "target": "40c09e1797cb4f7aa138dba44be41190d6225487",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/40c09e1797cb4f7aa138dba44be41190d6225487/"
    },
    "refs/pull/262/head": {
      "target": "2d98265865ac4e41c496a6e0cde6a1d7db710445",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2d98265865ac4e41c496a6e0cde6a1d7db710445/"
    },
    "refs/pull/263/head": {
      "target": "56731b16ce5a9f11cb9496d76e4ffac11987fc85",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/56731b16ce5a9f11cb9496d76e4ffac11987fc85/"
    },
    "refs/pull/264/head": {
      "target": "25f9a4d886dc11ae59b6327c1f9adf33da6d4c94",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/25f9a4d886dc11ae59b6327c1f9adf33da6d4c94/"
    },
    "refs/pull/265/head": {
      "target": "f630dbf5b53467319c21652545aa1e5c53a848a4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f630dbf5b53467319c21652545aa1e5c53a848a4/"
    },
    "refs/pull/266/head": {
      "target": "13572c6216fcd28d018d6a5614ec0f36ceeda55a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/13572c6216fcd28d018d6a5614ec0f36ceeda55a/"
    },
    "refs/pull/267/head": {
      "target": "611c9dd331ed99bc4f1db31c9b4ca8e5e60bc95d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/611c9dd331ed99bc4f1db31c9b4ca8e5e60bc95d/"
    },
    "refs/pull/268/head": {
      "target": "3c28ef4f5e7f51cd8b621df7ab3dc4a8c0530da3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3c28ef4f5e7f51cd8b621df7ab3dc4a8c0530da3/"
    },
    "refs/pull/269/head": {
      "target": "8e574c79bc4215e5ff8ec517421e53c42a8c35e5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8e574c79bc4215e5ff8ec517421e53c42a8c35e5/"
    },
    "refs/pull/27/head": {
      "target": "189fc7cc3159abdec65288ac7d41c54a6564662d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/189fc7cc3159abdec65288ac7d41c54a6564662d/"
    },
    "refs/pull/270/head": {
      "target": "50167000e1f0c7ccf9179ae5f545a5723323b855",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/50167000e1f0c7ccf9179ae5f545a5723323b855/"
    },
    "refs/pull/271/head": {
      "target": "a77ad3d320b723c38de004ead0381aa4f88a4444",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a77ad3d320b723c38de004ead0381aa4f88a4444/"
    },
    "refs/pull/272/head": {
      "target": "758e95157712b79d851847b696765056007fabea",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/758e95157712b79d851847b696765056007fabea/"
    },
    "refs/pull/273/head": {
      "target": "da6c038ea785aa57880eae4e9cf12d80572d8d6f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/da6c038ea785aa57880eae4e9cf12d80572d8d6f/"
    },
    "refs/pull/274/head": {
      "target": "da6c038ea785aa57880eae4e9cf12d80572d8d6f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/da6c038ea785aa57880eae4e9cf12d80572d8d6f/"
    },
    "refs/pull/275/head": {
      "target": "550e249d8fbe4888be3ed830bdde0b33f8d57477",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/550e249d8fbe4888be3ed830bdde0b33f8d57477/"
    },
    "refs/pull/276/head": {
      "target": "ac1e07bb68fe6f996a7723a50acb43bb0f86687a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ac1e07bb68fe6f996a7723a50acb43bb0f86687a/"
    },
    "refs/pull/277/head": {
      "target": "46e8281ad84affe18b22f5f278809248fd6270fa",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/46e8281ad84affe18b22f5f278809248fd6270fa/"
    },
    "refs/pull/278/head": {
      "target": "b562e44f507e863c6792946e4e1b1449fbbac85d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b562e44f507e863c6792946e4e1b1449fbbac85d/"
    },
    "refs/pull/279/head": {
      "target": "944dea41bf4dff0cfc4ce985884c6b2c8fa6b54b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/944dea41bf4dff0cfc4ce985884c6b2c8fa6b54b/"
    },
    "refs/pull/28/head": {
      "target": "4c309de0fab4c373c57ddf2a207ba88381552e33",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4c309de0fab4c373c57ddf2a207ba88381552e33/"
    },
    "refs/pull/280/head": {
      "target": "58ad70ca7fa582a06afa4ee34398b6076acb5991",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/58ad70ca7fa582a06afa4ee34398b6076acb5991/"
    },
    "refs/pull/281/head": {
      "target": "74aabf328e5a4c1047eb21fa31514cb28d81510c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/74aabf328e5a4c1047eb21fa31514cb28d81510c/"
    },
    "refs/pull/282/head": {
      "target": "04395066d181874904395276027fad49f3f932c8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/04395066d181874904395276027fad49f3f932c8/"
    },
    "refs/pull/283/head": {
      "target": "e8d612606e462d10c80c47eb680bcff4f34199a3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e8d612606e462d10c80c47eb680bcff4f34199a3/"
    },
    "refs/pull/284/head": {
      "target": "361b689a45bfc2bcd8576ba5c9d404d0ae996f7c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/361b689a45bfc2bcd8576ba5c9d404d0ae996f7c/"
    },
    "refs/pull/285/head": {
      "target": "0fa75782806ae9d938dda083aa60932b7796f652",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0fa75782806ae9d938dda083aa60932b7796f652/"
    },
    "refs/pull/286/head": {
      "target": "b5bdf8c33fe9965b5d138613e0592409c29d79df",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b5bdf8c33fe9965b5d138613e0592409c29d79df/"
    },
    "refs/pull/287/head": {
      "target": "49d9e361f11035deaeaa9272c07bea53c0b67215",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/49d9e361f11035deaeaa9272c07bea53c0b67215/"
    },
    "refs/pull/288/head": {
      "target": "39c6ec13dced7ad2d3dec3cea5442ddb3af6de80",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/39c6ec13dced7ad2d3dec3cea5442ddb3af6de80/"
    },
    "refs/pull/289/head": {
      "target": "b20a2a26f262c83913d0748e356a94db12fdd37e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b20a2a26f262c83913d0748e356a94db12fdd37e/"
    },
    "refs/pull/290/head": {
      "target": "50ff39c4b1e12eea4ba487e3b0dd4ac8ba495b9e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/50ff39c4b1e12eea4ba487e3b0dd4ac8ba495b9e/"
    },
    "refs/pull/291/head": {
      "target": "74559c69e8cdc3ad60bdb0fa88d77acc22342568",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/74559c69e8cdc3ad60bdb0fa88d77acc22342568/"
    },
    "refs/pull/292/head": {
      "target": "da6843394289dcb5268f0ba7b30dab9d62db639c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/da6843394289dcb5268f0ba7b30dab9d62db639c/"
    },
    "refs/pull/293/head": {
      "target": "63ba67c85bb8466e6ed4e3be9739977623d4fc26",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/63ba67c85bb8466e6ed4e3be9739977623d4fc26/"
    },
    "refs/pull/294/head": {
      "target": "e35a40ee43109f2adc08636d4451350b5ef4565f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e35a40ee43109f2adc08636d4451350b5ef4565f/"
    },
    "refs/pull/295/head": {
      "target": "c8b275208fe8d58ab9cbdf8d78562ed4c2409b90",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c8b275208fe8d58ab9cbdf8d78562ed4c2409b90/"
    },
    "refs/pull/296/head": {
      "target": "e32d8d64fc1f78613cf5c946e405738c794d066d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e32d8d64fc1f78613cf5c946e405738c794d066d/"
    },
    "refs/pull/297/head": {
      "target": "78fcc50d9d221dc7a1846889c5768b350e9a10a7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/78fcc50d9d221dc7a1846889c5768b350e9a10a7/"
    },
    "refs/pull/298/head": {
      "target": "32a332dece1fe21d28c7c428fb24231e959de4cd",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/32a332dece1fe21d28c7c428fb24231e959de4cd/"
    },
    "refs/pull/299/head": {
      "target": "35905f6b6870cec60843c2b0102ae0e89c6ba317",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/35905f6b6870cec60843c2b0102ae0e89c6ba317/"
    },
    "refs/pull/30/head": {
      "target": "d9572ca38bcb66a71673a6dd1a234891965ef268",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d9572ca38bcb66a71673a6dd1a234891965ef268/"
    },
    "refs/pull/300/head": {
      "target": "f4c2e8622ffde470c356f045af9d8741dce60b91",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f4c2e8622ffde470c356f045af9d8741dce60b91/"
    },
    "refs/pull/301/head": {
      "target": "2809820ccbf1fb7a7bbff61142e153637e8849d9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2809820ccbf1fb7a7bbff61142e153637e8849d9/"
    },
    "refs/pull/302/head": {
      "target": "598ebd1c63307e2a4fac5ad54fcaeb9eaee8a010",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/598ebd1c63307e2a4fac5ad54fcaeb9eaee8a010/"
    },
    "refs/pull/303/head": {
      "target": "b29a31082939eaf1d71e2c009d7007fda0500d3e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b29a31082939eaf1d71e2c009d7007fda0500d3e/"
    },
    "refs/pull/304/head": {
      "target": "e43d617acfada2fa4559143288b8a65d94213f29",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e43d617acfada2fa4559143288b8a65d94213f29/"
    },
    "refs/pull/305/head": {
      "target": "dce53a26afe31c63b88bc5b5c75b41ec86579291",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/dce53a26afe31c63b88bc5b5c75b41ec86579291/"
    },
    "refs/pull/306/head": {
      "target": "ea6079c93ed8edc027566183a035d27f57ca3d39",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ea6079c93ed8edc027566183a035d27f57ca3d39/"
    },
    "refs/pull/307/head": {
      "target": "5bfb021a963b529f7477b2d038749444f6911791",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5bfb021a963b529f7477b2d038749444f6911791/"
    },
    "refs/pull/308/head": {
      "target": "adf7b859143aeffacbadfdc961b9b1fe0a383584",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/adf7b859143aeffacbadfdc961b9b1fe0a383584/"
    },
    "refs/pull/309/head": {
      "target": "f0c8e44f6a51954b41e7437a9fb7de2283254243",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f0c8e44f6a51954b41e7437a9fb7de2283254243/"
    },
    "refs/pull/31/head": {
      "target": "59bc74e3e7e818215a52dd4e36faf133324f00a5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/59bc74e3e7e818215a52dd4e36faf133324f00a5/"
    },
    "refs/pull/310/head": {
      "target": "6fc7bf1f90f408a206a9c4963f1e9beb4763f0dc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6fc7bf1f90f408a206a9c4963f1e9beb4763f0dc/"
    },
    "refs/pull/311/head": {
      "target": "b31878ec488c2d489fa59727a3c8eb6953592fbd",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b31878ec488c2d489fa59727a3c8eb6953592fbd/"
    },
    "refs/pull/312/head": {
      "target": "f05a5c071a67ae7bab386613dab3aef51f5ce830",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f05a5c071a67ae7bab386613dab3aef51f5ce830/"
    },
    "refs/pull/313/head": {
      "target": "d7bd21469e5f666feffa6183702450201972f39e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d7bd21469e5f666feffa6183702450201972f39e/"
    },
    "refs/pull/314/head": {
      "target": "6cd0ca142f03232a8d30ed939381125914c031cb",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6cd0ca142f03232a8d30ed939381125914c031cb/"
    },
    "refs/pull/315/head": {
      "target": "ee5c85a10a11492065acdf789188747d6e61221c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ee5c85a10a11492065acdf789188747d6e61221c/"
    },
    "refs/pull/316/head": {
      "target": "289297fe68fcb19f50945bf71bc50d6722e6fcdc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/289297fe68fcb19f50945bf71bc50d6722e6fcdc/"
    },
    "refs/pull/317/head": {
      "target": "af73400dd8c9ec1034e75ad907729f015f9d20fd",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/af73400dd8c9ec1034e75ad907729f015f9d20fd/"
    },
    "refs/pull/318/head": {
      "target": "2d63596c10062ada68a418172135b99afa143c83",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2d63596c10062ada68a418172135b99afa143c83/"
    },
    "refs/pull/319/head": {
      "target": "8d7a38ab8eeca68c4f7174eec48f5f135f0d26c5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8d7a38ab8eeca68c4f7174eec48f5f135f0d26c5/"
    },
    "refs/pull/32/head": {
      "target": "385b8c7586d74a991ab2a57cf20c43eb19f8bf0b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/385b8c7586d74a991ab2a57cf20c43eb19f8bf0b/"
    },
    "refs/pull/320/head": {
      "target": "8b3bad56fbb3ecc77c5b1354bb40fe76680d127c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8b3bad56fbb3ecc77c5b1354bb40fe76680d127c/"
    },
    "refs/pull/321/head": {
      "target": "e4034a8b451213a3db7f88402a34343fecb3f6e3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e4034a8b451213a3db7f88402a34343fecb3f6e3/"
    },
    "refs/pull/322/head": {
      "target": "b5aaac93b3540cfeb7faebb7c7c69608d0fcc717",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b5aaac93b3540cfeb7faebb7c7c69608d0fcc717/"
    },
    "refs/pull/323/head": {
      "target": "249d773c8d77e8dcb622de9d4f951653c4897d5a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/249d773c8d77e8dcb622de9d4f951653c4897d5a/"
    },
    "refs/pull/324/head": {
      "target": "652d8f3a7c567cecd695f4599a759db038f1bcaf",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/652d8f3a7c567cecd695f4599a759db038f1bcaf/"
    },
    "refs/pull/325/head": {
      "target": "7c46f7e141e3bc1d3c48323655158fca1db984cc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7c46f7e141e3bc1d3c48323655158fca1db984cc/"
    },
    "refs/pull/326/head": {
      "target": "b19af90d59f80631ee07762f1d49965b3506f98d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b19af90d59f80631ee07762f1d49965b3506f98d/"
    },
    "refs/pull/327/head": {
      "target": "44373cf907353bb8d5646f8258c58efa09044424",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/44373cf907353bb8d5646f8258c58efa09044424/"
    },
    "refs/pull/328/head": {
      "target": "0f93ce620c1fed7a43ce5fbd9522a21669ac2fa2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0f93ce620c1fed7a43ce5fbd9522a21669ac2fa2/"
    },
    "refs/pull/329/head": {
      "target": "faec44105bb5ced27b3f354ab94caaea3ca8f82b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/faec44105bb5ced27b3f354ab94caaea3ca8f82b/"
    },
    "refs/pull/33/head": {
      "target": "e6b679fc3132a3a740260242cc76c21eac1e4580",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e6b679fc3132a3a740260242cc76c21eac1e4580/"
    },
    "refs/pull/330/head": {
      "target": "4b9d3609fb7c273eec0d57b19f714d0db90b781a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4b9d3609fb7c273eec0d57b19f714d0db90b781a/"
    },
    "refs/pull/331/head": {
      "target": "0d0031e73f3b51fd0fe41ad8843e6c368f92c121",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0d0031e73f3b51fd0fe41ad8843e6c368f92c121/"
    },
    "refs/pull/332/head": {
      "target": "fc64c8d9fcda9545ae6d2c890d458d727442ed30",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fc64c8d9fcda9545ae6d2c890d458d727442ed30/"
    },
    "refs/pull/333/head": {
      "target": "5142c05a4036394dc1a4c9a603665c707bb7bf1b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5142c05a4036394dc1a4c9a603665c707bb7bf1b/"
    },
    "refs/pull/334/head": {
      "target": "c8a15c0bcf0602072288498ab906f0b23f9c1e38",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c8a15c0bcf0602072288498ab906f0b23f9c1e38/"
    },
    "refs/pull/335/head": {
      "target": "166b7c6e4b3a15984dd738852787bd1451809bac",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/166b7c6e4b3a15984dd738852787bd1451809bac/"
    },
    "refs/pull/336/head": {
      "target": "75f10f028f79b12e1d40634edc54c505a9ee14e4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/75f10f028f79b12e1d40634edc54c505a9ee14e4/"
    },
    "refs/pull/337/head": {
      "target": "52ad41dc5b0ca1f967bc0d8246ece660feffa6ba",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/52ad41dc5b0ca1f967bc0d8246ece660feffa6ba/"
    },
    "refs/pull/338/head": {
      "target": "b4ef9a1b49e5eaf7bb352b0d55288a6cb468cf7a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b4ef9a1b49e5eaf7bb352b0d55288a6cb468cf7a/"
    },
    "refs/pull/339/head": {
      "target": "1a1891d762d6e64daf07b5be4817e3fbb29e3c59",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1a1891d762d6e64daf07b5be4817e3fbb29e3c59/"
    },
    "refs/pull/34/head": {
      "target": "4811ed66220715abc105cf421626c948c4f5c982",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4811ed66220715abc105cf421626c948c4f5c982/"
    },
    "refs/pull/340/head": {
      "target": "60777af38a5886a8baae829d20bc375626e27cb6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/60777af38a5886a8baae829d20bc375626e27cb6/"
    },
    "refs/pull/341/head": {
      "target": "bba77b62571c87d07e6b667dec31b112903f9287",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bba77b62571c87d07e6b667dec31b112903f9287/"
    },
    "refs/pull/342/head": {
      "target": "afa00fcd0b83e2dc7897746c3fc7be55dd77fdb9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/afa00fcd0b83e2dc7897746c3fc7be55dd77fdb9/"
    },
    "refs/pull/343/head": {
      "target": "a22eb213add1316725dd271145a3bcc35dcac4ff",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a22eb213add1316725dd271145a3bcc35dcac4ff/"
    },
    "refs/pull/344/head": {
      "target": "a22eb213add1316725dd271145a3bcc35dcac4ff",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a22eb213add1316725dd271145a3bcc35dcac4ff/"
    },
    "refs/pull/345/head": {
      "target": "3d77b6c484eb48d5be7125f22c9dd4d6f40d580c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3d77b6c484eb48d5be7125f22c9dd4d6f40d580c/"
    },
    "refs/pull/346/head": {
      "target": "6d325b06b6cbec6d6526ed2e70d2c8e6002eae86",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6d325b06b6cbec6d6526ed2e70d2c8e6002eae86/"
    },
    "refs/pull/347/head": {
      "target": "56731b16ce5a9f11cb9496d76e4ffac11987fc85",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/56731b16ce5a9f11cb9496d76e4ffac11987fc85/"
    },
    "refs/pull/348/head": {
      "target": "6a2646ab3b9c163f4ded416682a60f740c2979e2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6a2646ab3b9c163f4ded416682a60f740c2979e2/"
    },
    "refs/pull/349/head": {
      "target": "4ed9ed9cc6bcd89bd801f3f2cbee2b5d2aaf3573",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4ed9ed9cc6bcd89bd801f3f2cbee2b5d2aaf3573/"
    },
    "refs/pull/350/head": {
      "target": "c3877c41736a89b259c013cb411b9235dcea9978",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c3877c41736a89b259c013cb411b9235dcea9978/"
    },
    "refs/pull/351/head": {
      "target": "a6d0408ba316929156d000b340ec80162ebf1b69",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a6d0408ba316929156d000b340ec80162ebf1b69/"
    },
    "refs/pull/352/head": {
      "target": "7014d74dcf3ee97617ba97b2cf43d17f7af2897c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7014d74dcf3ee97617ba97b2cf43d17f7af2897c/"
    },
    "refs/pull/353/head": {
      "target": "7014d74dcf3ee97617ba97b2cf43d17f7af2897c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7014d74dcf3ee97617ba97b2cf43d17f7af2897c/"
    },
    "refs/pull/354/head": {
      "target": "7014d74dcf3ee97617ba97b2cf43d17f7af2897c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7014d74dcf3ee97617ba97b2cf43d17f7af2897c/"
    },
    "refs/pull/355/head": {
      "target": "7014d74dcf3ee97617ba97b2cf43d17f7af2897c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7014d74dcf3ee97617ba97b2cf43d17f7af2897c/"
    },
    "refs/pull/356/head": {
      "target": "3a388c950e82005e8ed0345223195f3f442d97db",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3a388c950e82005e8ed0345223195f3f442d97db/"
    },
    "refs/pull/357/head": {
      "target": "d6fb280aeb684071ca49950de4f111ebcf560c76",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d6fb280aeb684071ca49950de4f111ebcf560c76/"
    },
    "refs/pull/358/head": {
      "target": "171925806c6e522ee49a0ff939f5be58b4c6663a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/171925806c6e522ee49a0ff939f5be58b4c6663a/"
    },
    "refs/pull/359/head": {
      "target": "cb682a51bc88a8b3ffb72cc6ddca7d3ec0de36d6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cb682a51bc88a8b3ffb72cc6ddca7d3ec0de36d6/"
    },
    "refs/pull/36/head": {
      "target": "cd6d103e51ad63a32b3447e5d3629c8bd64ca0eb",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cd6d103e51ad63a32b3447e5d3629c8bd64ca0eb/"
    },
    "refs/pull/360/head": {
      "target": "8f45ebd868c1edf164ef05b5e9bc5f0da4e42168",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8f45ebd868c1edf164ef05b5e9bc5f0da4e42168/"
    },
    "refs/pull/361/head": {
      "target": "897504a9f47dbb50a9e1017035a18bc8f7ee2d2a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/897504a9f47dbb50a9e1017035a18bc8f7ee2d2a/"
    },
    "refs/pull/362/head": {
      "target": "d16932a050c3e46ce2ab1d49f4564bfb8163050f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d16932a050c3e46ce2ab1d49f4564bfb8163050f/"
    },
    "refs/pull/363/head": {
      "target": "9a7d0a9e51b582c8a97b9cfc3f20535aad26f862",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9a7d0a9e51b582c8a97b9cfc3f20535aad26f862/"
    },
    "refs/pull/364/head": {
      "target": "6c92d518525e1233c7c7cb94bac6fc15ce0b881d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6c92d518525e1233c7c7cb94bac6fc15ce0b881d/"
    },
    "refs/pull/365/head": {
      "target": "98342311d169dd0937f5d153db1cd7cdc4c2f175",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/98342311d169dd0937f5d153db1cd7cdc4c2f175/"
    },
    "refs/pull/366/head": {
      "target": "5cc60aeedf315a7513f92e98314e86d515b986d1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5cc60aeedf315a7513f92e98314e86d515b986d1/"
    },
    "refs/pull/367/head": {
      "target": "9126e5cb039244b915816225b9b34f39950a8951",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9126e5cb039244b915816225b9b34f39950a8951/"
    },
    "refs/pull/368/head": {
      "target": "f15aca3397d999482d8ceef70ed9f9fa6ca3622c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f15aca3397d999482d8ceef70ed9f9fa6ca3622c/"
    },
    "refs/pull/369/head": {
      "target": "12b2cf3ee652c8189ccef46a2f5601321136d1a8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/12b2cf3ee652c8189ccef46a2f5601321136d1a8/"
    },
    "refs/pull/37/head": {
      "target": "236bcd33d76cb1a39630bb732dec367ea13cdb05",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/236bcd33d76cb1a39630bb732dec367ea13cdb05/"
    },
    "refs/pull/370/head": {
      "target": "01c672b1a1bc78e5cf75c1fabeda4f1ab0650b2a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/01c672b1a1bc78e5cf75c1fabeda4f1ab0650b2a/"
    },
    "refs/pull/371/head": {
      "target": "06b4b8e0eab798b6aac383507a6146f4b626ff5c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/06b4b8e0eab798b6aac383507a6146f4b626ff5c/"
    },
    "refs/pull/372/head": {
      "target": "08101e32ecca4080a86335ea2f0cc00c7acffbb4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/08101e32ecca4080a86335ea2f0cc00c7acffbb4/"
    },
    "refs/pull/373/head": {
      "target": "9a7d0a9e51b582c8a97b9cfc3f20535aad26f862",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9a7d0a9e51b582c8a97b9cfc3f20535aad26f862/"
    },
    "refs/pull/374/head": {
      "target": "e066b5f672a7100338154a1b692885fe6e912e81",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e066b5f672a7100338154a1b692885fe6e912e81/"
    },
    "refs/pull/375/head": {
      "target": "ce1cd8975587aa8e809cdb01ab1ee1b9de1b9259",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ce1cd8975587aa8e809cdb01ab1ee1b9de1b9259/"
    },
    "refs/pull/376/head": {
      "target": "76aa120e8f5c994f0ca1d17952983983d0be2b11",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/76aa120e8f5c994f0ca1d17952983983d0be2b11/"
    },
    "refs/pull/377/head": {
      "target": "d5bd03ddaaea60da686a9a47fb637127452bfe89",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d5bd03ddaaea60da686a9a47fb637127452bfe89/"
    },
    "refs/pull/378/head": {
      "target": "2b7592be509df803a9753d251f11274b5df29d23",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2b7592be509df803a9753d251f11274b5df29d23/"
    },
    "refs/pull/379/head": {
      "target": "c8e67d5f10dda4f7dfeeea3c1688b8dc3341b40c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c8e67d5f10dda4f7dfeeea3c1688b8dc3341b40c/"
    },
    "refs/pull/38/head": {
      "target": "da7115edcd79581c692785dc959920ce4605916d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/da7115edcd79581c692785dc959920ce4605916d/"
    },
    "refs/pull/380/head": {
      "target": "216d14b289ca996c588fcb60780833725827246b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/216d14b289ca996c588fcb60780833725827246b/"
    },
    "refs/pull/381/head": {
      "target": "9fb8a76343c08acb6c0b4e5992adacaa9b280a86",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9fb8a76343c08acb6c0b4e5992adacaa9b280a86/"
    },
    "refs/pull/382/head": {
      "target": "b7e78624284d057c531061c79a9bf41acbe30c6b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b7e78624284d057c531061c79a9bf41acbe30c6b/"
    },
    "refs/pull/383/head": {
      "target": "0585733a81fcf61e4f442c064c4628fcee3c1802",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0585733a81fcf61e4f442c064c4628fcee3c1802/"
    },
    "refs/pull/384/head": {
      "target": "164efa15f83f7565d29adbdd403113e3a441e534",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/164efa15f83f7565d29adbdd403113e3a441e534/"
    },
    "refs/pull/385/head": {
      "target": "6c6eae943abe90bfb683d6d61d2b277c9d1c916f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6c6eae943abe90bfb683d6d61d2b277c9d1c916f/"
    },
    "refs/pull/386/head": {
      "target": "b4b7ab1430a3b40aa41f1a30a5e59d2bc5600f58",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b4b7ab1430a3b40aa41f1a30a5e59d2bc5600f58/"
    },
    "refs/pull/387/head": {
      "target": "be08675850745b55b7509fc0328417fa10221594",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/be08675850745b55b7509fc0328417fa10221594/"
    },
    "refs/pull/388/head": {
      "target": "4e2e05daac77f0ce6b4a80039a44051ac5381f5a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4e2e05daac77f0ce6b4a80039a44051ac5381f5a/"
    },
    "refs/pull/389/head": {
      "target": "53a1299e2c97f51f51c3f039eaa6a20b35a6a592",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/53a1299e2c97f51f51c3f039eaa6a20b35a6a592/"
    },
    "refs/pull/390/head": {
      "target": "30af8aca827ad299045ac576424f693aacf1f7a4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/30af8aca827ad299045ac576424f693aacf1f7a4/"
    },
    "refs/pull/391/head": {
      "target": "f0c6a571809ef76c383d508d17a63d84234096ef",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f0c6a571809ef76c383d508d17a63d84234096ef/"
    },
    "refs/pull/392/head": {
      "target": "83914c627967c6ef18b7f97ed08cdaf8cbe19812",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/83914c627967c6ef18b7f97ed08cdaf8cbe19812/"
    },
    "refs/pull/393/head": {
      "target": "61b32d6af970ab85b4f6c20b7fb9d6c2976250d5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/61b32d6af970ab85b4f6c20b7fb9d6c2976250d5/"
    },
    "refs/pull/394/head": {
      "target": "94023f53640bfb699771e2a94968e5faf4b93c6b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/94023f53640bfb699771e2a94968e5faf4b93c6b/"
    },
    "refs/pull/395/head": {
      "target": "8250804c1ee9f61ff834fca0fcadeec3c275b338",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8250804c1ee9f61ff834fca0fcadeec3c275b338/"
    },
    "refs/pull/396/head": {
      "target": "6ddf680f1af28e77b05e053b8266feb54bdf04e3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6ddf680f1af28e77b05e053b8266feb54bdf04e3/"
    },
    "refs/pull/397/head": {
      "target": "5dd1f027e72186dc6c8f66b718100d64423a86e4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5dd1f027e72186dc6c8f66b718100d64423a86e4/"
    },
    "refs/pull/398/head": {
      "target": "f91bac5b02798879cb32613326bf929ad70b6e76",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f91bac5b02798879cb32613326bf929ad70b6e76/"
    },
    "refs/pull/399/head": {
      "target": "d35fb555ea8cb6fcd1283b61ab9550e42da42891",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d35fb555ea8cb6fcd1283b61ab9550e42da42891/"
    },
    "refs/pull/40/head": {
      "target": "8647e99beb12cccdb997766322f0f5c6992900f2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8647e99beb12cccdb997766322f0f5c6992900f2/"
    },
    "refs/pull/400/head": {
      "target": "7a275245db618e6714823d18d0328b7cef06f894",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7a275245db618e6714823d18d0328b7cef06f894/"
    },
    "refs/pull/401/head": {
      "target": "786fb1bb524f43ea10c9e8d502f716fd29a3154c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/786fb1bb524f43ea10c9e8d502f716fd29a3154c/"
    },
    "refs/pull/402/head": {
      "target": "5cef316f086cfbf7bd07cd7517c2b2191cd94df6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5cef316f086cfbf7bd07cd7517c2b2191cd94df6/"
    },
    "refs/pull/403/head": {
      "target": "2e24a09e2155286fd631aef007c5366a74dcdba1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2e24a09e2155286fd631aef007c5366a74dcdba1/"
    },
    "refs/pull/404/head": {
      "target": "7a65df9ee7d606461ddbd3ad4eb4eee926e3c53b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7a65df9ee7d606461ddbd3ad4eb4eee926e3c53b/"
    },
    "refs/pull/405/head": {
      "target": "b6eac1be8e3627db6496d47dbdefe7a8efa18de1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b6eac1be8e3627db6496d47dbdefe7a8efa18de1/"
    },
    "refs/pull/406/head": {
      "target": "2a5751977750419c616313665b6b0b888d9a57c4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2a5751977750419c616313665b6b0b888d9a57c4/"
    },
    "refs/pull/407/head": {
      "target": "250168ad3c58a0482dc57320c9ed1bd26858bd7f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/250168ad3c58a0482dc57320c9ed1bd26858bd7f/"
    },
    "refs/pull/408/head": {
      "target": "2bfbf1b4477d1d5df68b13050a42000e0cec1211",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2bfbf1b4477d1d5df68b13050a42000e0cec1211/"
    },
    "refs/pull/409/head": {
      "target": "f0f2b60df2ee87b223a2b0dce95e9bed244cc488",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f0f2b60df2ee87b223a2b0dce95e9bed244cc488/"
    },
    "refs/pull/41/head": {
      "target": "14f309b5fdbb2dbf4542e486d9cf3d3fd6f7e315",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/14f309b5fdbb2dbf4542e486d9cf3d3fd6f7e315/"
    },
    "refs/pull/410/head": {
      "target": "190bcee8302b4ceef3d7efc5f88ed06395da363d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/190bcee8302b4ceef3d7efc5f88ed06395da363d/"
    },
    "refs/pull/411/head": {
      "target": "0fbf5f7a23372cc8052732a6b23ca333e9a7d10d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0fbf5f7a23372cc8052732a6b23ca333e9a7d10d/"
    },
    "refs/pull/412/head": {
      "target": "72c03333f35f7eddd707680d8e5de3527b5ef674",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/72c03333f35f7eddd707680d8e5de3527b5ef674/"
    },
    "refs/pull/413/head": {
      "target": "4a6869acbc5eb0b23107c525d40aff92a111c3d2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4a6869acbc5eb0b23107c525d40aff92a111c3d2/"
    },
    "refs/pull/414/head": {
      "target": "2ef9b2fda9a2ef3dbf4d39ec1b4e8a65d00c1e38",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2ef9b2fda9a2ef3dbf4d39ec1b4e8a65d00c1e38/"
    },
    "refs/pull/415/head": {
      "target": "217ff764007fcbca3186f22935454bae52a7f344",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/217ff764007fcbca3186f22935454bae52a7f344/"
    },
    "refs/pull/416/head": {
      "target": "03d22f936420fb14a3ce41b848c1221a95bcd78c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/03d22f936420fb14a3ce41b848c1221a95bcd78c/"
    },
    "refs/pull/417/head": {
      "target": "fad7e29c77a21303915db259c557db6cdc8b4f7e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fad7e29c77a21303915db259c557db6cdc8b4f7e/"
    },
    "refs/pull/418/head": {
      "target": "62558026ff2d073d42e1c332e6279c4f6331076e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/62558026ff2d073d42e1c332e6279c4f6331076e/"
    },
    "refs/pull/419/head": {
      "target": "6609687047dee2cffde7065f1ce0305b8ec182c6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6609687047dee2cffde7065f1ce0305b8ec182c6/"
    },
    "refs/pull/42/head": {
      "target": "ae2c32f562bcd2ed91a5087674ea4467eb0bf9eb",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ae2c32f562bcd2ed91a5087674ea4467eb0bf9eb/"
    },
    "refs/pull/420/head": {
      "target": "482458f0d7d6f0e99d3468fecaa0996b772f351d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/482458f0d7d6f0e99d3468fecaa0996b772f351d/"
    },
    "refs/pull/421/head": {
      "target": "693046a8f6dc773b4faf9f25a4293ce78f6ffd18",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/693046a8f6dc773b4faf9f25a4293ce78f6ffd18/"
    },
    "refs/pull/422/head": {
      "target": "24b670a6d5192b7530f0022eb2597a0c509718b2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/24b670a6d5192b7530f0022eb2597a0c509718b2/"
    },
    "refs/pull/423/head": {
      "target": "24b670a6d5192b7530f0022eb2597a0c509718b2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/24b670a6d5192b7530f0022eb2597a0c509718b2/"
    },
    "refs/pull/424/head": {
      "target": "9669d37610fdf13baa743b0240d33cbe62d91b5f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9669d37610fdf13baa743b0240d33cbe62d91b5f/"
    },
    "refs/pull/425/head": {
      "target": "4ea7937d42170370d1f39e17889b38ec573bc25a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4ea7937d42170370d1f39e17889b38ec573bc25a/"
    },
    "refs/pull/426/head": {
      "target": "e6067070c792f9bf61ad7874d63cc13d93b55e70",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e6067070c792f9bf61ad7874d63cc13d93b55e70/"
    },
    "refs/pull/427/head": {
      "target": "6bd51edbd06cb2051e0801b87ef9304ffcb3de77",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6bd51edbd06cb2051e0801b87ef9304ffcb3de77/"
    },
    "refs/pull/428/head": {
      "target": "d4880e208d8415c237816ea23f550f7aff82f1e0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d4880e208d8415c237816ea23f550f7aff82f1e0/"
    },
    "refs/pull/429/head": {
      "target": "8335cb5c1fe8d000a94fcb6fe4947aa6080f57ed",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8335cb5c1fe8d000a94fcb6fe4947aa6080f57ed/"
    },
    "refs/pull/43/head": {
      "target": "fed54e2603259d749f7545c37f916f3705f4d7a3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fed54e2603259d749f7545c37f916f3705f4d7a3/"
    },
    "refs/pull/430/head": {
      "target": "731d6453ff7888af06559689b28621316a8ec632",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/731d6453ff7888af06559689b28621316a8ec632/"
    },
    "refs/pull/431/head": {
      "target": "75ad8906db2692a7fe37b776b043240f957fd77d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/75ad8906db2692a7fe37b776b043240f957fd77d/"
    },
    "refs/pull/432/head": {
      "target": "33a0550e5b6d8b8539ea5793d7b3c44b8c56af6f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/33a0550e5b6d8b8539ea5793d7b3c44b8c56af6f/"
    },
    "refs/pull/433/head": {
      "target": "f25f5ea4720341732519a10ce46629525caf455e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f25f5ea4720341732519a10ce46629525caf455e/"
    },
    "refs/pull/434/head": {
      "target": "b6053feb6b8dbc36528f1ce77c836daf4e95c295",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b6053feb6b8dbc36528f1ce77c836daf4e95c295/"
    },
    "refs/pull/435/head": {
      "target": "a79852218b602641937ef8e569f71f4b8f7194a0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a79852218b602641937ef8e569f71f4b8f7194a0/"
    },
    "refs/pull/436/head": {
      "target": "bc0eefc372331d5dc65f23a2ba41c95e7ba6b588",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bc0eefc372331d5dc65f23a2ba41c95e7ba6b588/"
    },
    "refs/pull/437/head": {
      "target": "c78ffb46efd3561a0fef60cc0b7c6c1050bfb891",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c78ffb46efd3561a0fef60cc0b7c6c1050bfb891/"
    },
    "refs/pull/438/head": {
      "target": "87e9c8e7cc2b6a99917b36cbfcec7fe3a0c77f4e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/87e9c8e7cc2b6a99917b36cbfcec7fe3a0c77f4e/"
    },
    "refs/pull/439/head": {
      "target": "0a24672b5ee7fe4270362f80d1e8f6f8613c3776",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0a24672b5ee7fe4270362f80d1e8f6f8613c3776/"
    },
    "refs/pull/44/head": {
      "target": "c1a172978efc50d33cc9ff508253b722afe62439",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c1a172978efc50d33cc9ff508253b722afe62439/"
    },
    "refs/pull/440/head": {
      "target": "760a84d4c1b24bad30f527ff2f4682173fc20f5b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/760a84d4c1b24bad30f527ff2f4682173fc20f5b/"
    },
    "refs/pull/441/head": {
      "target": "4523622d274e0bf0b2b054445d570a3215c3c250",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4523622d274e0bf0b2b054445d570a3215c3c250/"
    },
    "refs/pull/442/head": {
      "target": "4f182950bb4757737d47ead3c6c7fb94d79cb455",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4f182950bb4757737d47ead3c6c7fb94d79cb455/"
    },
    "refs/pull/443/head": {
      "target": "e449d32cf47bebf079417e5cb41b6d3fc52b66e6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e449d32cf47bebf079417e5cb41b6d3fc52b66e6/"
    },
    "refs/pull/444/head": {
      "target": "5b009bbab149a3dedf48a1da10ea6ed105391d32",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5b009bbab149a3dedf48a1da10ea6ed105391d32/"
    },
    "refs/pull/445/head": {
      "target": "336f8adb0292a25ea41f0515a80850031f814b9b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/336f8adb0292a25ea41f0515a80850031f814b9b/"
    },
    "refs/pull/446/head": {
      "target": "82c1388f1d80463287f6d2e4a1afbdbe4e3b536f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/82c1388f1d80463287f6d2e4a1afbdbe4e3b536f/"
    },
    "refs/pull/447/head": {
      "target": "df0a8c2ff32c570152a52485a0411e455c5db324",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/df0a8c2ff32c570152a52485a0411e455c5db324/"
    },
    "refs/pull/448/head": {
      "target": "63e1b83b6547a6527307f857d246185e82554852",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/63e1b83b6547a6527307f857d246185e82554852/"
    },
    "refs/pull/449/head": {
      "target": "c797bc88f77b51c4ba0b5d782b8284c7ed0233bc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c797bc88f77b51c4ba0b5d782b8284c7ed0233bc/"
    },
    "refs/pull/45/head": {
      "target": "a20189cefb9c5b3b2d63e5bc44bd19e0cff92bbb",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a20189cefb9c5b3b2d63e5bc44bd19e0cff92bbb/"
    },
    "refs/pull/450/head": {
      "target": "589d3f748326d11d5215f7ac9d79d8f8eb63a027",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/589d3f748326d11d5215f7ac9d79d8f8eb63a027/"
    },
    "refs/pull/451/head": {
      "target": "b7f548a4326007ebb0e4430d7d474f4216eb24e4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b7f548a4326007ebb0e4430d7d474f4216eb24e4/"
    },
    "refs/pull/452/head": {
      "target": "37fa970c315e291d6357b3c5fd2adfc174877daf",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/37fa970c315e291d6357b3c5fd2adfc174877daf/"
    },
    "refs/pull/453/head": {
      "target": "81ef89a4d0bdec56d5761eda40c67f499dfac59e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/81ef89a4d0bdec56d5761eda40c67f499dfac59e/"
    },
    "refs/pull/454/head": {
      "target": "95ddcb49ed5df35196633ffd077106090aeca622",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/95ddcb49ed5df35196633ffd077106090aeca622/"
    },
    "refs/pull/455/head": {
      "target": "e508c12752cb5dcf08978f59d3f1ea5d7d1eded9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e508c12752cb5dcf08978f59d3f1ea5d7d1eded9/"
    },
    "refs/pull/456/head": {
      "target": "0fa3aa711f3b893ad791be394f5cae33e7f6a445",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0fa3aa711f3b893ad791be394f5cae33e7f6a445/"
    },
    "refs/pull/457/head": {
      "target": "365974bb47d4f2c9942ca51dc817bc56688c90c4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/365974bb47d4f2c9942ca51dc817bc56688c90c4/"
    },
    "refs/pull/458/head": {
      "target": "a9c43121e937ca3fedb00e709d824b18e6e7e529",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a9c43121e937ca3fedb00e709d824b18e6e7e529/"
    },
    "refs/pull/459/head": {
      "target": "f341f1122f922de82b377837791d85b8679c0876",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f341f1122f922de82b377837791d85b8679c0876/"
    },
    "refs/pull/46/head": {
      "target": "03df0e88460cbecb44b7e254d9dd386b562d3345",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/03df0e88460cbecb44b7e254d9dd386b562d3345/"
    },
    "refs/pull/460/head": {
      "target": "030c2255da38ebbf2ac8b3b212277530ef04ed86",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/030c2255da38ebbf2ac8b3b212277530ef04ed86/"
    },
    "refs/pull/461/head": {
      "target": "0d9bef62d2372a6289d89de5a7f8c818a2493505",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0d9bef62d2372a6289d89de5a7f8c818a2493505/"
    },
    "refs/pull/462/head": {
      "target": "dd75eb5de7e7d8edea434ee00020e7936b95ac2d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/dd75eb5de7e7d8edea434ee00020e7936b95ac2d/"
    },
    "refs/pull/463/head": {
      "target": "71caf6325b7276732aec4f048bebea8149093c7a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/71caf6325b7276732aec4f048bebea8149093c7a/"
    },
    "refs/pull/464/head": {
      "target": "71caf6325b7276732aec4f048bebea8149093c7a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/71caf6325b7276732aec4f048bebea8149093c7a/"
    },
    "refs/pull/465/head": {
      "target": "0a74b619581a73d3138f2aa3ef17f8c66c609381",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0a74b619581a73d3138f2aa3ef17f8c66c609381/"
    },
    "refs/pull/466/head": {
      "target": "0a74b619581a73d3138f2aa3ef17f8c66c609381",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0a74b619581a73d3138f2aa3ef17f8c66c609381/"
    },
    "refs/pull/467/head": {
      "target": "5713045910fae41895bf134cc8bf68bd8f7e058a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5713045910fae41895bf134cc8bf68bd8f7e058a/"
    },
    "refs/pull/468/head": {
      "target": "4c49fe7c3bbf5e1be09a6780ed585dac2a71bede",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4c49fe7c3bbf5e1be09a6780ed585dac2a71bede/"
    },
    "refs/pull/469/head": {
      "target": "0f82123c346bc90e1de89604e25b01889e5f5272",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0f82123c346bc90e1de89604e25b01889e5f5272/"
    },
    "refs/pull/47/head": {
      "target": "45151edaf3ac28abd009a03c0ff9502cd508d50f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/45151edaf3ac28abd009a03c0ff9502cd508d50f/"
    },
    "refs/pull/470/head": {
      "target": "17e0146ce93eb4a2ab77aeab85e7757f14794ab0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/17e0146ce93eb4a2ab77aeab85e7757f14794ab0/"
    },
    "refs/pull/471/head": {
      "target": "297e16f6ad1cc967a278d08d6f9fec3bf5e91c7f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/297e16f6ad1cc967a278d08d6f9fec3bf5e91c7f/"
    },
    "refs/pull/472/head": {
      "target": "dad6cea8b76588dd7c193c034a0f446e2ff2edf4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/dad6cea8b76588dd7c193c034a0f446e2ff2edf4/"
    },
    "refs/pull/473/head": {
      "target": "e094fd835917f7df676835dc5f48552e8134b241",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e094fd835917f7df676835dc5f48552e8134b241/"
    },
    "refs/pull/474/head": {
      "target": "6398f5ce546e8059dc8fb248df085a8492a13cf8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6398f5ce546e8059dc8fb248df085a8492a13cf8/"
    },
    "refs/pull/475/head": {
      "target": "201005f2e7785df6b4b703a98dd77ff8c8bc9b69",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/201005f2e7785df6b4b703a98dd77ff8c8bc9b69/"
    },
    "refs/pull/476/head": {
      "target": "2c3a8037a7749186dd2be2208fd9b738fb0994b0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2c3a8037a7749186dd2be2208fd9b738fb0994b0/"
    },
    "refs/pull/477/head": {
      "target": "89e6f9710b96add5cc9d2ae529419d1ec0fc98a1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/89e6f9710b96add5cc9d2ae529419d1ec0fc98a1/"
    },
    "refs/pull/478/head": {
      "target": "240d02ef98ba52afd0d4a4c39911005177c13d01",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/240d02ef98ba52afd0d4a4c39911005177c13d01/"
    },
    "refs/pull/479/head": {
      "target": "f19cdb2bce127e50b6a74cb18905b906c014ab15",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f19cdb2bce127e50b6a74cb18905b906c014ab15/"
    },
    "refs/pull/48/head": {
      "target": "5a704fa565ddf2b76337ce1f37591ab9430b6b51",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5a704fa565ddf2b76337ce1f37591ab9430b6b51/"
    },
    "refs/pull/480/head": {
      "target": "fded58cb1edcdbb3b5b97428a7dac39d65e0833f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fded58cb1edcdbb3b5b97428a7dac39d65e0833f/"
    },
    "refs/pull/481/head": {
      "target": "f523ba29b0442f77d78bdfa6f31648b237b54357",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f523ba29b0442f77d78bdfa6f31648b237b54357/"
    },
    "refs/pull/482/head": {
      "target": "edff8c8b50ce1d42a041f08fc8f7d02cbfddf9c1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/edff8c8b50ce1d42a041f08fc8f7d02cbfddf9c1/"
    },
    "refs/pull/483/head": {
      "target": "b577ca73c54e39de924231ff09c99034dda4c4c6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b577ca73c54e39de924231ff09c99034dda4c4c6/"
    },
    "refs/pull/484/head": {
      "target": "bac437eeee544bd1db8dc8c7101f12acba1abbd5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bac437eeee544bd1db8dc8c7101f12acba1abbd5/"
    },
    "refs/pull/485/head": {
      "target": "14feee48ee6a7be52b6a936c1e4fde7474a8ca1e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/14feee48ee6a7be52b6a936c1e4fde7474a8ca1e/"
    },
    "refs/pull/486/head": {
      "target": "70365d4a9614b4b70f9b40b690e698e2684671a6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/70365d4a9614b4b70f9b40b690e698e2684671a6/"
    },
    "refs/pull/487/head": {
      "target": "777364e4c241fbcc29dab21b5587cb582eaff9ba",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/777364e4c241fbcc29dab21b5587cb582eaff9ba/"
    },
    "refs/pull/488/head": {
      "target": "4cc4895087ca050abc30ac6ad7f2d3dce0195d9f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4cc4895087ca050abc30ac6ad7f2d3dce0195d9f/"
    },
    "refs/pull/489/head": {
      "target": "695e7ce35a0687e4257b177dfaf9333d6b3d06a0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/695e7ce35a0687e4257b177dfaf9333d6b3d06a0/"
    },
    "refs/pull/490/head": {
      "target": "ebaba108de36be5834f7ea6de720c616cb94dcd1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ebaba108de36be5834f7ea6de720c616cb94dcd1/"
    },
    "refs/pull/491/head": {
      "target": "b679ccccb768420c69170cbcad9d4c83e29f0c53",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b679ccccb768420c69170cbcad9d4c83e29f0c53/"
    },
    "refs/pull/492/head": {
      "target": "b679ccccb768420c69170cbcad9d4c83e29f0c53",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b679ccccb768420c69170cbcad9d4c83e29f0c53/"
    },
    "refs/pull/493/head": {
      "target": "8e7fcb0af563742f4cde9540192751c68c22b506",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8e7fcb0af563742f4cde9540192751c68c22b506/"
    },
    "refs/pull/494/head": {
      "target": "e2a1d1730c3244943a8763d04e740ad93f901944",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e2a1d1730c3244943a8763d04e740ad93f901944/"
    },
    "refs/pull/495/head": {
      "target": "4117d05b5b2df91849f443a83ddf8b7a3c096936",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4117d05b5b2df91849f443a83ddf8b7a3c096936/"
    },
    "refs/pull/496/head": {
      "target": "916c7ff67ac58f71214fb46fde00ab95574761ec",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/916c7ff67ac58f71214fb46fde00ab95574761ec/"
    },
    "refs/pull/497/head": {
      "target": "f210a44c2f884fae982a1168f87f8a74e0abb6aa",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f210a44c2f884fae982a1168f87f8a74e0abb6aa/"
    },
    "refs/pull/498/head": {
      "target": "32ea3f53f1ff3fb61946f18cb5734b50d8cbe107",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/32ea3f53f1ff3fb61946f18cb5734b50d8cbe107/"
    },
    "refs/pull/499/head": {
      "target": "a98bc1a9f28939bd945647a81e7e0467223d0bdc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a98bc1a9f28939bd945647a81e7e0467223d0bdc/"
    },
    "refs/pull/50/head": {
      "target": "c34a39af8f334b2c4614aec52f450274672a5611",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c34a39af8f334b2c4614aec52f450274672a5611/"
    },
    "refs/pull/500/head": {
      "target": "5cc3afeee6098886d6569733182c6108d7258383",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5cc3afeee6098886d6569733182c6108d7258383/"
    },
    "refs/pull/501/head": {
      "target": "fbf8629b99e2778114d17a0c8e6f2052c8e311fc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fbf8629b99e2778114d17a0c8e6f2052c8e311fc/"
    },
    "refs/pull/502/head": {
      "target": "de07441bf8aba3e61eb4d78a3626d5ba9d5c22ef",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/de07441bf8aba3e61eb4d78a3626d5ba9d5c22ef/"
    },
    "refs/pull/503/head": {
      "target": "698ac3bae6f91af880efa91b1567bc9527e92275",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/698ac3bae6f91af880efa91b1567bc9527e92275/"
    },
    "refs/pull/504/head": {
      "target": "ca3864cfbfcc3145cd618639adf34136cc1b5f66",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ca3864cfbfcc3145cd618639adf34136cc1b5f66/"
    },
    "refs/pull/505/head": {
      "target": "653e4d10903b092be79c804ea0570691a2e4bebb",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/653e4d10903b092be79c804ea0570691a2e4bebb/"
    },
    "refs/pull/506/head": {
      "target": "5747fe9dd2b92bf8462c7ab839dc3eb095a49c7e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5747fe9dd2b92bf8462c7ab839dc3eb095a49c7e/"
    },
    "refs/pull/507/head": {
      "target": "3dbfb1776e535cdf0464e4b39e707173acd62c69",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3dbfb1776e535cdf0464e4b39e707173acd62c69/"
    },
    "refs/pull/508/head": {
      "target": "bfb0f0be5bacb831274a31f5a90f45a34642f2b0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bfb0f0be5bacb831274a31f5a90f45a34642f2b0/"
    },
    "refs/pull/509/head": {
      "target": "1f9f978449a779816fa57c0cac84aa49a8ce92d4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1f9f978449a779816fa57c0cac84aa49a8ce92d4/"
    },
    "refs/pull/51/head": {
      "target": "9ee9fc5f0988df5677f0f142b5b88a8988d283d7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9ee9fc5f0988df5677f0f142b5b88a8988d283d7/"
    },
    "refs/pull/510/head": {
      "target": "c24dede5742488cbfe8a19b03dd16ef3782e7821",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c24dede5742488cbfe8a19b03dd16ef3782e7821/"
    },
    "refs/pull/511/head": {
      "target": "cfa1a85b872ee8597783597ef57c4829ff250e94",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cfa1a85b872ee8597783597ef57c4829ff250e94/"
    },
    "refs/pull/512/head": {
      "target": "00b6d582180e21de1889b8076eb5be899dfdad23",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/00b6d582180e21de1889b8076eb5be899dfdad23/"
    },
    "refs/pull/513/head": {
      "target": "0141d3014d47a0a2af03179e0d3f90f3bad7004b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0141d3014d47a0a2af03179e0d3f90f3bad7004b/"
    },
    "refs/pull/514/head": {
      "target": "54505fe77d3623d01b3fb0b8fac599810d837f4a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/54505fe77d3623d01b3fb0b8fac599810d837f4a/"
    },
    "refs/pull/515/head": {
      "target": "dbad815dbf8c7705814460acb942ae75a3ebb05f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/dbad815dbf8c7705814460acb942ae75a3ebb05f/"
    },
    "refs/pull/516/head": {
      "target": "ac68ebb4407003615964a5c13b121a4ef093fe8e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ac68ebb4407003615964a5c13b121a4ef093fe8e/"
    },
    "refs/pull/517/head": {
      "target": "9c5866fdeccae63dc81d98e4707b90965b3d922d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9c5866fdeccae63dc81d98e4707b90965b3d922d/"
    },
    "refs/pull/518/head": {
      "target": "50355b9a86d1720631042d9487dad6fd93e1d064",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/50355b9a86d1720631042d9487dad6fd93e1d064/"
    },
    "refs/pull/519/head": {
      "target": "fff02cff7d4968014a585489c312bc7e6c2edfef",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fff02cff7d4968014a585489c312bc7e6c2edfef/"
    },
    "refs/pull/520/head": {
      "target": "03dc6f2d2a54e7b06363db98cedd578566f60eea",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/03dc6f2d2a54e7b06363db98cedd578566f60eea/"
    },
    "refs/pull/521/head": {
      "target": "26a81a994feef8d54f53ca15fbd2a527b285d56f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/26a81a994feef8d54f53ca15fbd2a527b285d56f/"
    },
    "refs/pull/522/head": {
      "target": "169f5eed50c12274196eef1d620ff4fcb486ab1a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/169f5eed50c12274196eef1d620ff4fcb486ab1a/"
    },
    "refs/pull/523/head": {
      "target": "7c36fffedf81239ebee07b2f564ee90729b2858d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7c36fffedf81239ebee07b2f564ee90729b2858d/"
    },
    "refs/pull/524/head": {
      "target": "95c701445052467ffec8a39d5c931836d52f936b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/95c701445052467ffec8a39d5c931836d52f936b/"
    },
    "refs/pull/525/head": {
      "target": "de25c71cf6b4d9bb49937f96485c8a35c577d567",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/de25c71cf6b4d9bb49937f96485c8a35c577d567/"
    },
    "refs/pull/526/head": {
      "target": "de25c71cf6b4d9bb49937f96485c8a35c577d567",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/de25c71cf6b4d9bb49937f96485c8a35c577d567/"
    },
    "refs/pull/527/head": {
      "target": "e23b879fa6f83b8d9b8a395c0297374917b19d1b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e23b879fa6f83b8d9b8a395c0297374917b19d1b/"
    },
    "refs/pull/528/head": {
      "target": "61867ab54695cb4b98cb988698c11e4a2ddf1e8a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/61867ab54695cb4b98cb988698c11e4a2ddf1e8a/"
    },
    "refs/pull/529/head": {
      "target": "eff4b7d38e9f9d37679a339177d1b9802d5896d5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/eff4b7d38e9f9d37679a339177d1b9802d5896d5/"
    },
    "refs/pull/530/head": {
      "target": "31789e1a93385f570926ee6f9d0a95ceca8ce034",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/31789e1a93385f570926ee6f9d0a95ceca8ce034/"
    },
    "refs/pull/531/head": {
      "target": "cdb06e9d8f520c969676e7d6778cffe5894f079f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cdb06e9d8f520c969676e7d6778cffe5894f079f/"
    },
    "refs/pull/532/head": {
      "target": "bd169a7c16ae38babd3ccada3c1c38174c13f852",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bd169a7c16ae38babd3ccada3c1c38174c13f852/"
    },
    "refs/pull/533/head": {
      "target": "3c8c2df5de61233a0e350a689a37f7c09a780cdb",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3c8c2df5de61233a0e350a689a37f7c09a780cdb/"
    },
    "refs/pull/534/head": {
      "target": "f605ba97fb80522656c7dce9825a908f1e765b57",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f605ba97fb80522656c7dce9825a908f1e765b57/"
    },
    "refs/pull/535/head": {
      "target": "0865b299fef18104e44f62e076624f265ea60894",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0865b299fef18104e44f62e076624f265ea60894/"
    },
    "refs/pull/536/head": {
      "target": "15bd388188d3bff61b5b74a3368fd3e8a95f1b37",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/15bd388188d3bff61b5b74a3368fd3e8a95f1b37/"
    },
    "refs/pull/537/head": {
      "target": "aba84eeef538d1b710b7be37e53bf507c7a647ac",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/aba84eeef538d1b710b7be37e53bf507c7a647ac/"
    },
    "refs/pull/538/head": {
      "target": "4125d6aebab009bf0168f5a4e59abdc14a1eff8e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4125d6aebab009bf0168f5a4e59abdc14a1eff8e/"
    },
    "refs/pull/539/head": {
      "target": "5758c763a77d84ff423d1b546ee82015a1128ae0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5758c763a77d84ff423d1b546ee82015a1128ae0/"
    },
    "refs/pull/54/head": {
      "target": "e36433fc62d3dc373aaa67041fc6773744289cef",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e36433fc62d3dc373aaa67041fc6773744289cef/"
    },
    "refs/pull/540/head": {
      "target": "d6525ea4e4b39100432e09f0a31e37af81454b9b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d6525ea4e4b39100432e09f0a31e37af81454b9b/"
    },
    "refs/pull/541/head": {
      "target": "9b4b5459207142cda725a44262696d0e7c45e5d9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9b4b5459207142cda725a44262696d0e7c45e5d9/"
    },
    "refs/pull/542/head": {
      "target": "7658ab9d01388961f608386e086f4c9bc8f9fe7c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7658ab9d01388961f608386e086f4c9bc8f9fe7c/"
    },
    "refs/pull/543/head": {
      "target": "ff4f576ab2ff383c07a4916d55e6b0049e57d415",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ff4f576ab2ff383c07a4916d55e6b0049e57d415/"
    },
    "refs/pull/544/head": {
      "target": "cedf3d01cb002602a4d366a07984f7506dce8f34",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cedf3d01cb002602a4d366a07984f7506dce8f34/"
    },
    "refs/pull/545/head": {
      "target": "b1ca4f281c2bc505f0535955eafc9730bc7ee851",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b1ca4f281c2bc505f0535955eafc9730bc7ee851/"
    },
    "refs/pull/546/head": {
      "target": "13f5b03007d1ae271f0c5cdd0a1b046dc5d752d8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/13f5b03007d1ae271f0c5cdd0a1b046dc5d752d8/"
    },
    "refs/pull/547/head": {
      "target": "e03ae7c6080a1326af979d60c9e758cd5c351d35",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e03ae7c6080a1326af979d60c9e758cd5c351d35/"
    },
    "refs/pull/548/head": {
      "target": "10462f9abafc40138362eb499face751ff392a7b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/10462f9abafc40138362eb499face751ff392a7b/"
    },
    "refs/pull/549/head": {
      "target": "607ca06d061f06a914373c40a5ca61761860cbd3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/607ca06d061f06a914373c40a5ca61761860cbd3/"
    },
    "refs/pull/55/head": {
      "target": "bfd411613cb73ef9c29b01ac1407d1205da9b9dc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bfd411613cb73ef9c29b01ac1407d1205da9b9dc/"
    },
    "refs/pull/550/head": {
      "target": "4885041d59e114c3fa2f900714b66c3c001a8550",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4885041d59e114c3fa2f900714b66c3c001a8550/"
    },
    "refs/pull/551/head": {
      "target": "757d15e21d9b304ee2c9fc0a8239458dcd011047",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/757d15e21d9b304ee2c9fc0a8239458dcd011047/"
    },
    "refs/pull/552/head": {
      "target": "19939bae48c9e5b46a6af6e39f51bf5c31250c03",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/19939bae48c9e5b46a6af6e39f51bf5c31250c03/"
    },
    "refs/pull/553/head": {
      "target": "d1d16c997e7c7901ea03a81420221dd1554bc392",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d1d16c997e7c7901ea03a81420221dd1554bc392/"
    },
    "refs/pull/554/head": {
      "target": "0f8d768071ab2cbf7dbfde671988746774ec7f6e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0f8d768071ab2cbf7dbfde671988746774ec7f6e/"
    },
    "refs/pull/555/head": {
      "target": "f94cdc90b37dfd7b4866b0d6da4a018b4708fc46",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f94cdc90b37dfd7b4866b0d6da4a018b4708fc46/"
    },
    "refs/pull/556/head": {
      "target": "de3ee929f97490fe4f1e84efd6c5882a3e8c0664",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/de3ee929f97490fe4f1e84efd6c5882a3e8c0664/"
    },
    "refs/pull/557/head": {
      "target": "2cfef8d6cbcacb1ab20150e262e7723ad58a1c8c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2cfef8d6cbcacb1ab20150e262e7723ad58a1c8c/"
    },
    "refs/pull/558/head": {
      "target": "2d664a27fadcc46a9f6d75408d7e69c0ce86654d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2d664a27fadcc46a9f6d75408d7e69c0ce86654d/"
    },
    "refs/pull/559/head": {
      "target": "e99b836d1e14adfda71bce6496c369873650e980",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e99b836d1e14adfda71bce6496c369873650e980/"
    },
    "refs/pull/56/head": {
      "target": "bfd411613cb73ef9c29b01ac1407d1205da9b9dc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bfd411613cb73ef9c29b01ac1407d1205da9b9dc/"
    },
    "refs/pull/560/head": {
      "target": "de3ee929f97490fe4f1e84efd6c5882a3e8c0664",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/de3ee929f97490fe4f1e84efd6c5882a3e8c0664/"
    },
    "refs/pull/561/head": {
      "target": "7244f9e014d951e66cc03bd4ef213acbc40f759c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7244f9e014d951e66cc03bd4ef213acbc40f759c/"
    },
    "refs/pull/562/head": {
      "target": "74c23ca9d3918d50175bf9da4ce396ca654fe85f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/74c23ca9d3918d50175bf9da4ce396ca654fe85f/"
    },
    "refs/pull/563/head": {
      "target": "e331a381c3287cbdb7013fdea9b7c4acbd5b2489",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e331a381c3287cbdb7013fdea9b7c4acbd5b2489/"
    },
    "refs/pull/564/head": {
      "target": "b3aa772d0b9f919aaca9ee44240500087f13fa20",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b3aa772d0b9f919aaca9ee44240500087f13fa20/"
    },
    "refs/pull/565/head": {
      "target": "f9a0577922804d579c2bfe8f7cbd63b7a44637c1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f9a0577922804d579c2bfe8f7cbd63b7a44637c1/"
    },
    "refs/pull/566/head": {
      "target": "8c39283522cf620806acc46926c8bc880d6c3603",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8c39283522cf620806acc46926c8bc880d6c3603/"
    },
    "refs/pull/567/head": {
      "target": "24ab0be682b16911efcbac271538dc7ce30fcfe0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/24ab0be682b16911efcbac271538dc7ce30fcfe0/"
    },
    "refs/pull/568/head": {
      "target": "4da7b54690665353f469799733d689e23d5b7407",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4da7b54690665353f469799733d689e23d5b7407/"
    },
    "refs/pull/569/head": {
      "target": "7c586fde771a6d34e63df0237a42f8e4bce66ae0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7c586fde771a6d34e63df0237a42f8e4bce66ae0/"
    },
    "refs/pull/57/head": {
      "target": "8b2ac8926cc592bd2654a7dbe824ba95378c33c1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8b2ac8926cc592bd2654a7dbe824ba95378c33c1/"
    },
    "refs/pull/570/head": {
      "target": "e96e1ac821118a042038ffab293557612c02e5c8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e96e1ac821118a042038ffab293557612c02e5c8/"
    },
    "refs/pull/571/head": {
      "target": "c105eccb3165feb7da0d083ebce29fb678b00921",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c105eccb3165feb7da0d083ebce29fb678b00921/"
    },
    "refs/pull/572/head": {
      "target": "bc3262dea7524136749ca263bd1f0e09ac9c0209",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bc3262dea7524136749ca263bd1f0e09ac9c0209/"
    },
    "refs/pull/573/head": {
      "target": "7697b18d2463c7cce5e1f501d005272910ba407a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7697b18d2463c7cce5e1f501d005272910ba407a/"
    },
    "refs/pull/574/head": {
      "target": "421273bb9ccda2d35c59c1460446924218784cae",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/421273bb9ccda2d35c59c1460446924218784cae/"
    },
    "refs/pull/575/head": {
      "target": "798a7d682e48f9238059aa36f48e675897f0e726",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/798a7d682e48f9238059aa36f48e675897f0e726/"
    },
    "refs/pull/576/head": {
      "target": "8133cec66f541ec96ae08c45ffd968cd09d13485",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8133cec66f541ec96ae08c45ffd968cd09d13485/"
    },
    "refs/pull/577/head": {
      "target": "9ad530a1c92b3981e740bb2415fa31b8b9639601",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9ad530a1c92b3981e740bb2415fa31b8b9639601/"
    },
    "refs/pull/578/head": {
      "target": "5bd3a65f04f2f7bc5edd4e4dc156eaf982bbd900",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5bd3a65f04f2f7bc5edd4e4dc156eaf982bbd900/"
    },
    "refs/pull/579/head": {
      "target": "86115efe331970564952816d2d54afdb7b04a08a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/86115efe331970564952816d2d54afdb7b04a08a/"
    },
    "refs/pull/580/head": {
      "target": "1552604219847fbe2b4e3c0288093d41dd604f6b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1552604219847fbe2b4e3c0288093d41dd604f6b/"
    },
    "refs/pull/581/head": {
      "target": "aa1efbcccb25df81037cef73f137593fb2c80c11",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/aa1efbcccb25df81037cef73f137593fb2c80c11/"
    },
    "refs/pull/582/head": {
      "target": "006d34a39c79f16b25322ace7a0ef8acc56fc896",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/006d34a39c79f16b25322ace7a0ef8acc56fc896/"
    },
    "refs/pull/583/head": {
      "target": "af0f7992c86ecf230943d21e98a702d9ec9eedc1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/af0f7992c86ecf230943d21e98a702d9ec9eedc1/"
    },
    "refs/pull/584/head": {
      "target": "765cf442a8afe8e5c8c6896b5072066df5129077",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/765cf442a8afe8e5c8c6896b5072066df5129077/"
    },
    "refs/pull/585/head": {
      "target": "77a715984586a7796c918760ecf58f02fee94fac",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/77a715984586a7796c918760ecf58f02fee94fac/"
    },
    "refs/pull/586/head": {
      "target": "06a45de8e7493737064192471db6c2cb4536226d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/06a45de8e7493737064192471db6c2cb4536226d/"
    },
    "refs/pull/587/head": {
      "target": "b11fd4b17c29bdad53de10d95439939ffa8f5103",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b11fd4b17c29bdad53de10d95439939ffa8f5103/"
    },
    "refs/pull/588/head": {
      "target": "3c44dafb98c4b5049d47f68b09c66fd20f3d7ed9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3c44dafb98c4b5049d47f68b09c66fd20f3d7ed9/"
    },
    "refs/pull/589/head": {
      "target": "d0953b44f5bcfc977f15c4341d685c2aad979fd2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d0953b44f5bcfc977f15c4341d685c2aad979fd2/"
    },
    "refs/pull/59/head": {
      "target": "d006810e3baa4f269189bd84c662656ba3b3baad",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d006810e3baa4f269189bd84c662656ba3b3baad/"
    },
    "refs/pull/590/head": {
      "target": "605f73bca199dfda5bb156ab69e4cf248d4ac540",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/605f73bca199dfda5bb156ab69e4cf248d4ac540/"
    },
    "refs/pull/591/head": {
      "target": "d9761ce16e3b4d5481cba3e8dbf66b915948396d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d9761ce16e3b4d5481cba3e8dbf66b915948396d/"
    },
    "refs/pull/592/head": {
      "target": "bdf7b044979b62d63f805f7aab9d7040e57ebb2c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bdf7b044979b62d63f805f7aab9d7040e57ebb2c/"
    },
    "refs/pull/593/head": {
      "target": "011eb0d31e3f92e0d84bac6fc496b4e35123df31",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/011eb0d31e3f92e0d84bac6fc496b4e35123df31/"
    },
    "refs/pull/594/head": {
      "target": "eaf686f96950582e4e5987544db1330e836892e8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/eaf686f96950582e4e5987544db1330e836892e8/"
    },
    "refs/pull/595/head": {
      "target": "4158c805653202da3fac1f4c338303ca47b1c4e1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4158c805653202da3fac1f4c338303ca47b1c4e1/"
    },
    "refs/pull/596/head": {
      "target": "9aa1f1a2e3e655aa6c8e9c0673a1b8a3dc2f7cb7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9aa1f1a2e3e655aa6c8e9c0673a1b8a3dc2f7cb7/"
    },
    "refs/pull/597/head": {
      "target": "5d75175c357a4e4b4f5f8dc989a4f9c76467aa21",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5d75175c357a4e4b4f5f8dc989a4f9c76467aa21/"
    },
    "refs/pull/598/head": {
      "target": "095564bf9188062bf43ade84a50e051f284c7e47",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/095564bf9188062bf43ade84a50e051f284c7e47/"
    },
    "refs/pull/599/head": {
      "target": "43577dea337ac9ab11f8aec00c71f073ad4a104a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/43577dea337ac9ab11f8aec00c71f073ad4a104a/"
    },
    "refs/pull/6/head": {
      "target": "11dfcd9567e920fd90738450944f9e3912090af8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/11dfcd9567e920fd90738450944f9e3912090af8/"
    },
    "refs/pull/60/head": {
      "target": "38736fb4291ceb83e5cdeed410ae72d9c7b871aa",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/38736fb4291ceb83e5cdeed410ae72d9c7b871aa/"
    },
    "refs/pull/600/head": {
      "target": "8c9c14e978336f9538824692fdc382398d6281d7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8c9c14e978336f9538824692fdc382398d6281d7/"
    },
    "refs/pull/601/head": {
      "target": "764aece7cfc876aaddf524e4f30b18e1a814e109",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/764aece7cfc876aaddf524e4f30b18e1a814e109/"
    },
    "refs/pull/602/head": {
      "target": "fd994bf7356439d3e8d90275c7418894919d82cc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fd994bf7356439d3e8d90275c7418894919d82cc/"
    },
    "refs/pull/603/head": {
      "target": "e2a3d75cd645ebeaa272e01a8bc4212abff30238",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e2a3d75cd645ebeaa272e01a8bc4212abff30238/"
    },
    "refs/pull/604/head": {
      "target": "2f866fdfc8825877291c58a5bb828a7cba91f57e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2f866fdfc8825877291c58a5bb828a7cba91f57e/"
    },
    "refs/pull/605/head": {
      "target": "f7256b0b65ed679cb99c5e69271bf998ca7ae45e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f7256b0b65ed679cb99c5e69271bf998ca7ae45e/"
    },
    "refs/pull/606/head": {
      "target": "e7dfcfd29c9b2788be488669597c5bb1283ca28e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e7dfcfd29c9b2788be488669597c5bb1283ca28e/"
    },
    "refs/pull/607/head": {
      "target": "859afbdc27af7882c66cc58a035ff738ab4e42f6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/859afbdc27af7882c66cc58a035ff738ab4e42f6/"
    },
    "refs/pull/608/head": {
      "target": "d1afe43d4a0ccb3dd98b3ad723a5d43aa0c87125",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d1afe43d4a0ccb3dd98b3ad723a5d43aa0c87125/"
    },
    "refs/pull/609/head": {
      "target": "e9c5282e609040393c697702cbddf0bc5847e0c8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e9c5282e609040393c697702cbddf0bc5847e0c8/"
    },
    "refs/pull/61/head": {
      "target": "d80339484c5e8eb8e8b2a78f9d273a677ef2ee39",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d80339484c5e8eb8e8b2a78f9d273a677ef2ee39/"
    },
    "refs/pull/610/head": {
      "target": "e49eb8ccf3f7ff15da4d2190762a58bc4e98137a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e49eb8ccf3f7ff15da4d2190762a58bc4e98137a/"
    },
    "refs/pull/611/head": {
      "target": "1f1741eee8c8aaebf3bb7103808fc5eb36405251",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1f1741eee8c8aaebf3bb7103808fc5eb36405251/"
    },
    "refs/pull/612/head": {
      "target": "65b3d643ce66b8ae199217ba3cf53f79bd4df413",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/65b3d643ce66b8ae199217ba3cf53f79bd4df413/"
    },
    "refs/pull/613/head": {
      "target": "a3b82f2a1bccc114f9d42de816085f5f5e09f8de",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a3b82f2a1bccc114f9d42de816085f5f5e09f8de/"
    },
    "refs/pull/614/head": {
      "target": "7dab9e5fc4b2a74c0186c53206b4d901e8b74251",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7dab9e5fc4b2a74c0186c53206b4d901e8b74251/"
    },
    "refs/pull/615/head": {
      "target": "e90d4a2dc4d16f354bff593e0711a504b536e428",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e90d4a2dc4d16f354bff593e0711a504b536e428/"
    },
    "refs/pull/616/head": {
      "target": "fc6b8bbf790b1d7b2452b4d7f5fb443d3750b740",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fc6b8bbf790b1d7b2452b4d7f5fb443d3750b740/"
    },
    "refs/pull/617/head": {
      "target": "9f057c7d35e5dd35d069677f8037ba3cdf558565",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9f057c7d35e5dd35d069677f8037ba3cdf558565/"
    },
    "refs/pull/618/head": {
      "target": "bfde66cd15efe02bc5ed1d69bb724b5a789b67e7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bfde66cd15efe02bc5ed1d69bb724b5a789b67e7/"
    },
    "refs/pull/619/head": {
      "target": "dd4587324ae900b881dcf411b44cb291c909fd19",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/dd4587324ae900b881dcf411b44cb291c909fd19/"
    },
    "refs/pull/62/head": {
      "target": "348bc78e57f91f04be4054600d618217da87c9fc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/348bc78e57f91f04be4054600d618217da87c9fc/"
    },
    "refs/pull/620/head": {
      "target": "5125e93e009ec86f8cbebbe445c27eb8c14ca98f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5125e93e009ec86f8cbebbe445c27eb8c14ca98f/"
    },
    "refs/pull/621/head": {
      "target": "5125e93e009ec86f8cbebbe445c27eb8c14ca98f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5125e93e009ec86f8cbebbe445c27eb8c14ca98f/"
    },
    "refs/pull/622/head": {
      "target": "798a7d682e48f9238059aa36f48e675897f0e726",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/798a7d682e48f9238059aa36f48e675897f0e726/"
    },
    "refs/pull/623/head": {
      "target": "6bc8c94c36c8f997e2c77f1c8840c6aa06112457",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6bc8c94c36c8f997e2c77f1c8840c6aa06112457/"
    },
    "refs/pull/624/head": {
      "target": "6cc7ebb36508db979ae4e00d6c0b273b7dfe7e91",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6cc7ebb36508db979ae4e00d6c0b273b7dfe7e91/"
    },
    "refs/pull/625/head": {
      "target": "4ec91a4ce5cdd43858bc48607129c542fb91a423",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4ec91a4ce5cdd43858bc48607129c542fb91a423/"
    },
    "refs/pull/626/head": {
      "target": "a4d64afb021400c22665c4caad3ed00694d0dca3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a4d64afb021400c22665c4caad3ed00694d0dca3/"
    },
    "refs/pull/627/head": {
      "target": "2c90908607119ff97b8adbb549d2ce0ca75db6f8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2c90908607119ff97b8adbb549d2ce0ca75db6f8/"
    },
    "refs/pull/628/head": {
      "target": "260044453c5361ced6457b960d0646443a4f2288",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/260044453c5361ced6457b960d0646443a4f2288/"
    },
    "refs/pull/629/head": {
      "target": "f8669906389916b67343487604ac867e35eeff2e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f8669906389916b67343487604ac867e35eeff2e/"
    },
    "refs/pull/63/head": {
      "target": "fe533806adeaa87e57e55e9ebd168ef906531a17",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fe533806adeaa87e57e55e9ebd168ef906531a17/"
    },
    "refs/pull/630/head": {
      "target": "f261c4e529dac5608a604d3dd3ae1cd2adf23c89",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f261c4e529dac5608a604d3dd3ae1cd2adf23c89/"
    },
    "refs/pull/631/head": {
      "target": "fb402d626b9efd6e575f5904baef5cc954d40a77",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fb402d626b9efd6e575f5904baef5cc954d40a77/"
    },
    "refs/pull/632/head": {
      "target": "e5eb09d9f3267625402a64da431563ccb2a2c285",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e5eb09d9f3267625402a64da431563ccb2a2c285/"
    },
    "refs/pull/633/head": {
      "target": "58c3522a972b14b7e6005c69e1d3dfe795997a1f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/58c3522a972b14b7e6005c69e1d3dfe795997a1f/"
    },
    "refs/pull/634/head": {
      "target": "3d5e229212df0b0d08322f34bdbb66c31670fbf2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3d5e229212df0b0d08322f34bdbb66c31670fbf2/"
    },
    "refs/pull/635/head": {
      "target": "d7004006648c23c808c61b4f49d0373038476b32",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d7004006648c23c808c61b4f49d0373038476b32/"
    },
    "refs/pull/636/head": {
      "target": "936c1c1e5d4a77c5cdd322f5c5a69e3376643722",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/936c1c1e5d4a77c5cdd322f5c5a69e3376643722/"
    },
    "refs/pull/637/head": {
      "target": "723fd469a7aff8f91817249de32bb10358cf8d6f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/723fd469a7aff8f91817249de32bb10358cf8d6f/"
    },
    "refs/pull/638/head": {
      "target": "d8f11a6236a7b9f57bdbfcf095e05d3bf59af41e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d8f11a6236a7b9f57bdbfcf095e05d3bf59af41e/"
    },
    "refs/pull/639/head": {
      "target": "a4aa3d620ee5ecae2377d4d1de18ff454a2b444f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a4aa3d620ee5ecae2377d4d1de18ff454a2b444f/"
    },
    "refs/pull/64/head": {
      "target": "849ee3d46a7334e950fb17d1ac2cbe2c6b088f65",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/849ee3d46a7334e950fb17d1ac2cbe2c6b088f65/"
    },
    "refs/pull/640/head": {
      "target": "6f2cb5eec3f420c553c495801ee67a4de1cdb6e2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6f2cb5eec3f420c553c495801ee67a4de1cdb6e2/"
    },
    "refs/pull/641/head": {
      "target": "0829a3b4196852fe1799627a4ceb8eefc9a4db41",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0829a3b4196852fe1799627a4ceb8eefc9a4db41/"
    },
    "refs/pull/642/head": {
      "target": "c1c96d250055fdd78e55c2d1faf58828a9656fce",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c1c96d250055fdd78e55c2d1faf58828a9656fce/"
    },
    "refs/pull/643/head": {
      "target": "65b6e3bfb9d7b3a20874d9e85a528a6cbf8bad7c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/65b6e3bfb9d7b3a20874d9e85a528a6cbf8bad7c/"
    },
    "refs/pull/644/head": {
      "target": "450313c5d1313e79059031e6185174616f7ea329",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/450313c5d1313e79059031e6185174616f7ea329/"
    },
    "refs/pull/645/head": {
      "target": "78cafdd17cc5839d760ae26409a5eed2569092f7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/78cafdd17cc5839d760ae26409a5eed2569092f7/"
    },
    "refs/pull/646/head": {
      "target": "6672c39177d208deb3574724ac5a007a6f5fa316",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6672c39177d208deb3574724ac5a007a6f5fa316/"
    },
    "refs/pull/647/head": {
      "target": "8f94a6d78fc6528d752be66e2c66b802b5ecda78",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8f94a6d78fc6528d752be66e2c66b802b5ecda78/"
    },
    "refs/pull/648/head": {
      "target": "d3904ac62f15359dbcef96f3e3152e5759de8d1b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d3904ac62f15359dbcef96f3e3152e5759de8d1b/"
    },
    "refs/pull/649/head": {
      "target": "d967f0457a922721d68e46610c3b5ed57263b3a9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d967f0457a922721d68e46610c3b5ed57263b3a9/"
    },
    "refs/pull/65/head": {
      "target": "87e6350a4e94fae11b9bce7d6d4fe24d8177edba",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/87e6350a4e94fae11b9bce7d6d4fe24d8177edba/"
    },
    "refs/pull/650/head": {
      "target": "6e5e9b46fcd0c297c6504414af1bcad71f1825a0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6e5e9b46fcd0c297c6504414af1bcad71f1825a0/"
    },
    "refs/pull/651/head": {
      "target": "1ef8bf0f7fb9dd139468c416203ff0cc93a35fb3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1ef8bf0f7fb9dd139468c416203ff0cc93a35fb3/"
    },
    "refs/pull/652/head": {
      "target": "10faad6b2448480e0879cec241097a4cea8da6ad",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/10faad6b2448480e0879cec241097a4cea8da6ad/"
    },
    "refs/pull/653/head": {
      "target": "7b6a3df54049fe4b28a33b875f8dd2258deea945",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7b6a3df54049fe4b28a33b875f8dd2258deea945/"
    },
    "refs/pull/654/head": {
      "target": "9135557a1919de4879d0dcfcc0017f3695b0fe9f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9135557a1919de4879d0dcfcc0017f3695b0fe9f/"
    },
    "refs/pull/655/head": {
      "target": "ae57ca7caeb78c5d6021dcf69cc19752dfc36f04",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ae57ca7caeb78c5d6021dcf69cc19752dfc36f04/"
    },
    "refs/pull/656/head": {
      "target": "4761e210373c7706131e2cce030413c101c942e0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4761e210373c7706131e2cce030413c101c942e0/"
    },
    "refs/pull/657/head": {
      "target": "a4e98c1f89535c35bdddfb0abe9c4278664e0140",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a4e98c1f89535c35bdddfb0abe9c4278664e0140/"
    },
    "refs/pull/658/head": {
      "target": "8ee14458ed6dbd7d4fd75757e79b8963885c1f06",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8ee14458ed6dbd7d4fd75757e79b8963885c1f06/"
    },
    "refs/pull/659/head": {
      "target": "f018cba9ba4c1a68115c0cfc95cffbff9359fc39",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f018cba9ba4c1a68115c0cfc95cffbff9359fc39/"
    },
    "refs/pull/66/head": {
      "target": "00395b6fcac3d8db53bb9f105c5f6fdeb9fb41ae",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/00395b6fcac3d8db53bb9f105c5f6fdeb9fb41ae/"
    },
    "refs/pull/660/head": {
      "target": "3867a24cc4082051d1068e6b1f0e8a188a65b902",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3867a24cc4082051d1068e6b1f0e8a188a65b902/"
    },
    "refs/pull/661/head": {
      "target": "ee9ca9c2e5aa887814ca8c4f333cb4f1efe2e203",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ee9ca9c2e5aa887814ca8c4f333cb4f1efe2e203/"
    },
    "refs/pull/662/head": {
      "target": "619ddae3eb9ba3b40e0e43fb24acf0db0cf97271",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/619ddae3eb9ba3b40e0e43fb24acf0db0cf97271/"
    },
    "refs/pull/663/head": {
      "target": "f7ee5a1d5eebb5a3d9e204a8824eef7d6a182daf",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f7ee5a1d5eebb5a3d9e204a8824eef7d6a182daf/"
    },
    "refs/pull/664/head": {
      "target": "973c0391d270ada21be1129808a7dcbc11a11faf",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/973c0391d270ada21be1129808a7dcbc11a11faf/"
    },
    "refs/pull/665/head": {
      "target": "0acba081d9e67e24deaf80852382b89e33eaa30f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0acba081d9e67e24deaf80852382b89e33eaa30f/"
    },
    "refs/pull/666/head": {
      "target": "f2b8ab05a6419f2fd0ddac9965f177553f4f0eee",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f2b8ab05a6419f2fd0ddac9965f177553f4f0eee/"
    },
    "refs/pull/667/head": {
      "target": "3a51bdb8ea5fca736f874defa20103926cd22d51",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3a51bdb8ea5fca736f874defa20103926cd22d51/"
    },
    "refs/pull/668/head": {
      "target": "0619770a02a308349c378cbe9c4bbe9b330d9cae",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0619770a02a308349c378cbe9c4bbe9b330d9cae/"
    },
    "refs/pull/669/head": {
      "target": "406524c6fa928f5dd1a54d6b4dfd2e40d45e379b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/406524c6fa928f5dd1a54d6b4dfd2e40d45e379b/"
    },
    "refs/pull/67/head": {
      "target": "d47d367036be38c5180632ec8a3ad169a4593a88",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d47d367036be38c5180632ec8a3ad169a4593a88/"
    },
    "refs/pull/670/head": {
      "target": "c29642fb6a07307e6ed5fad9502262cfc96a9e43",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c29642fb6a07307e6ed5fad9502262cfc96a9e43/"
    },
    "refs/pull/671/head": {
      "target": "62e29e4f4781b0ed876f7e3baabb629aa126d9a7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/62e29e4f4781b0ed876f7e3baabb629aa126d9a7/"
    },
    "refs/pull/672/head": {
      "target": "d18fd2ef0726ee7ad8bbbe3d91682495df819f71",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d18fd2ef0726ee7ad8bbbe3d91682495df819f71/"
    },
    "refs/pull/673/head": {
      "target": "08145a76a9c188662179f5e34425f97be6c0a8c1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/08145a76a9c188662179f5e34425f97be6c0a8c1/"
    },
    "refs/pull/674/head": {
      "target": "fb9ca2e23f14eb81009fbd9e8f71322704c80f94",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fb9ca2e23f14eb81009fbd9e8f71322704c80f94/"
    },
    "refs/pull/675/head": {
      "target": "0f09813a689fe808aa59d773687fcb7a3920cbc8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0f09813a689fe808aa59d773687fcb7a3920cbc8/"
    },
    "refs/pull/676/head": {
      "target": "ef83f1c67ae5a57ef6fe2d3d5a6b54a33d64f3ef",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ef83f1c67ae5a57ef6fe2d3d5a6b54a33d64f3ef/"
    },
    "refs/pull/677/head": {
      "target": "ef83f1c67ae5a57ef6fe2d3d5a6b54a33d64f3ef",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ef83f1c67ae5a57ef6fe2d3d5a6b54a33d64f3ef/"
    },
    "refs/pull/678/head": {
      "target": "e66bc92df8f3c906db5c365bec5d0bd470c17313",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e66bc92df8f3c906db5c365bec5d0bd470c17313/"
    },
    "refs/pull/679/head": {
      "target": "b3b4caaa49a1fc4cb2396c7420896618099dc1fc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b3b4caaa49a1fc4cb2396c7420896618099dc1fc/"
    },
    "refs/pull/68/head": {
      "target": "35e764ceb7374278e5828223721f3e0e5b004e2e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/35e764ceb7374278e5828223721f3e0e5b004e2e/"
    },
    "refs/pull/680/head": {
      "target": "1d158463492d51f95e2aa585ba15b5834f85f7e3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1d158463492d51f95e2aa585ba15b5834f85f7e3/"
    },
    "refs/pull/681/head": {
      "target": "851a9b706bc2620d87b787cb51f179193d64b0bf",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/851a9b706bc2620d87b787cb51f179193d64b0bf/"
    },
    "refs/pull/682/head": {
      "target": "1c22c46db12120bccf1990eb18d5181dd933a416",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1c22c46db12120bccf1990eb18d5181dd933a416/"
    },
    "refs/pull/683/head": {
      "target": "cfa17c3951c0dc5e2fa816b73504f84adcede8f5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cfa17c3951c0dc5e2fa816b73504f84adcede8f5/"
    },
    "refs/pull/684/head": {
      "target": "402f0fff292bd1f4376b4a9b260130494accfd0a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/402f0fff292bd1f4376b4a9b260130494accfd0a/"
    },
    "refs/pull/685/head": {
      "target": "76e3b5f81bf0e42c1a36cda1999fb47ee7cc556d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/76e3b5f81bf0e42c1a36cda1999fb47ee7cc556d/"
    },
    "refs/pull/686/head": {
      "target": "188060660302b38ded8b294636be74998ba66a0f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/188060660302b38ded8b294636be74998ba66a0f/"
    },
    "refs/pull/687/head": {
      "target": "e2f61d211f689a502d7180222f71afa0a1cb458d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e2f61d211f689a502d7180222f71afa0a1cb458d/"
    },
    "refs/pull/688/head": {
      "target": "ea7e02d2d4597d0f5f49277cbd4f5e1a7c244f03",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ea7e02d2d4597d0f5f49277cbd4f5e1a7c244f03/"
    },
    "refs/pull/689/head": {
      "target": "56b93ba6a92afce446631036473bad2b49cdbe48",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/56b93ba6a92afce446631036473bad2b49cdbe48/"
    },
    "refs/pull/69/head": {
      "target": "98429ca30d9d55e694ea38c3d54265f533f62c00",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/98429ca30d9d55e694ea38c3d54265f533f62c00/"
    },
    "refs/pull/690/head": {
      "target": "235a071589222ca21ff648534c7fc5c20ba5b6b7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/235a071589222ca21ff648534c7fc5c20ba5b6b7/"
    },
    "refs/pull/691/head": {
      "target": "acc658561a7c4bdc75dfa0c081b9f02972444f54",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/acc658561a7c4bdc75dfa0c081b9f02972444f54/"
    },
    "refs/pull/692/head": {
      "target": "0cd32a63ceb07504677b50270dc75a4c618b7c05",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0cd32a63ceb07504677b50270dc75a4c618b7c05/"
    },
    "refs/pull/693/head": {
      "target": "84ee8bc7b927ca9f88870a6a2f73906bb9492698",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/84ee8bc7b927ca9f88870a6a2f73906bb9492698/"
    },
    "refs/pull/694/head": {
      "target": "88b96aa5ac01aa5d7bd89bd38bc753407c3296bb",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/88b96aa5ac01aa5d7bd89bd38bc753407c3296bb/"
    },
    "refs/pull/695/head": {
      "target": "fc1659ea85f73462d773fb00561f06b21d69746f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fc1659ea85f73462d773fb00561f06b21d69746f/"
    },
    "refs/pull/696/head": {
      "target": "cfc401f3cc8793c450aa4fb18d34eab410ae23c8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cfc401f3cc8793c450aa4fb18d34eab410ae23c8/"
    },
    "refs/pull/697/head": {
      "target": "c342dec662b5242112dd399b7650d44a467c7165",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c342dec662b5242112dd399b7650d44a467c7165/"
    },
    "refs/pull/698/head": {
      "target": "74f3923577807d6d409748969c3320f65ac0db6c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/74f3923577807d6d409748969c3320f65ac0db6c/"
    },
    "refs/pull/699/head": {
      "target": "d4c74aa0b9ea812df95f4804aae3504468e16a50",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d4c74aa0b9ea812df95f4804aae3504468e16a50/"
    },
    "refs/pull/7/head": {
      "target": "6510b5e11b96b38b93b5ee448577e7cd4b62641d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6510b5e11b96b38b93b5ee448577e7cd4b62641d/"
    },
    "refs/pull/700/head": {
      "target": "4d8f8d9f93a92695c68e11c96ed55e6b3c6a2016",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4d8f8d9f93a92695c68e11c96ed55e6b3c6a2016/"
    },
    "refs/pull/701/head": {
      "target": "2f5881a04dcddd2af109575cb5f65872367f0e13",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2f5881a04dcddd2af109575cb5f65872367f0e13/"
    },
    "refs/pull/702/head": {
      "target": "f1c384dcef2cc1aa279fed7aed0c755c3dcd0414",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f1c384dcef2cc1aa279fed7aed0c755c3dcd0414/"
    },
    "refs/pull/703/head": {
      "target": "ff68dcbc91de94b4919f4862a5f0b2fa3048b8b8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ff68dcbc91de94b4919f4862a5f0b2fa3048b8b8/"
    },
    "refs/pull/704/head": {
      "target": "2f4be8b8d7722ff98aa0b100ac9fd81709f3b45b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2f4be8b8d7722ff98aa0b100ac9fd81709f3b45b/"
    },
    "refs/pull/705/head": {
      "target": "9b2ec6c03ad18c7e18e4621b7b3df30952812fd3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9b2ec6c03ad18c7e18e4621b7b3df30952812fd3/"
    },
    "refs/pull/706/head": {
      "target": "d00bd59850cc4b42f61c1b303904bdeb424332d2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d00bd59850cc4b42f61c1b303904bdeb424332d2/"
    },
    "refs/pull/707/head": {
      "target": "aeeae07814e78a078c20b53b18557799037e720a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/aeeae07814e78a078c20b53b18557799037e720a/"
    },
    "refs/pull/708/head": {
      "target": "0e8cc547aaa2a0d9a74d4bd604f697e7dc3c79d2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0e8cc547aaa2a0d9a74d4bd604f697e7dc3c79d2/"
    },
    "refs/pull/709/head": {
      "target": "e3cef850750fa6bd64d5b6b69e976d1dbb3bb97d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e3cef850750fa6bd64d5b6b69e976d1dbb3bb97d/"
    },
    "refs/pull/71/head": {
      "target": "4b833790795b4a3665a6b01872fbab1c258d0ea3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4b833790795b4a3665a6b01872fbab1c258d0ea3/"
    },
    "refs/pull/710/head": {
      "target": "9e458dba898039ad56e5a3fefeda630b776bf864",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9e458dba898039ad56e5a3fefeda630b776bf864/"
    },
    "refs/pull/711/head": {
      "target": "9bae6214611067d88ad2f48077fc7880ed29ec9d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9bae6214611067d88ad2f48077fc7880ed29ec9d/"
    },
    "refs/pull/712/head": {
      "target": "942ef16574e041c64d4fdfcd380d58d7e0134796",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/942ef16574e041c64d4fdfcd380d58d7e0134796/"
    },
    "refs/pull/713/head": {
      "target": "792db1bdda9bd720a49e049cc4dfd054abeac655",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/792db1bdda9bd720a49e049cc4dfd054abeac655/"
    },
    "refs/pull/714/head": {
      "target": "37c938c74389719666c56e326ba7a77efdd7d409",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/37c938c74389719666c56e326ba7a77efdd7d409/"
    },
    "refs/pull/715/head": {
      "target": "73ec31329b81b61ed27cb80e12d3568b4bd87785",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/73ec31329b81b61ed27cb80e12d3568b4bd87785/"
    },
    "refs/pull/716/head": {
      "target": "d013e7051b7e067b873de56f2d1eb51ba241845b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d013e7051b7e067b873de56f2d1eb51ba241845b/"
    },
    "refs/pull/717/head": {
      "target": "c116c2805e4c92776bfb2c854259ed3a7b109436",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c116c2805e4c92776bfb2c854259ed3a7b109436/"
    },
    "refs/pull/718/head": {
      "target": "e288985d502897e9b2753afd2451a1972150a697",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e288985d502897e9b2753afd2451a1972150a697/"
    },
    "refs/pull/72/head": {
      "target": "dd54c43a917ee84a5e8692cc660926c92352829a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/dd54c43a917ee84a5e8692cc660926c92352829a/"
    },
    "refs/pull/720/head": {
      "target": "dc3093520e0586799052851d37ad2d8f6093a39f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/dc3093520e0586799052851d37ad2d8f6093a39f/"
    },
    "refs/pull/721/head": {
      "target": "bf6751bfe60ae266d214ccb60edc6164391755f9",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/bf6751bfe60ae266d214ccb60edc6164391755f9/"
    },
    "refs/pull/722/head": {
      "target": "abeccc4666bf429326a4c5e174c5458ae3727316",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/abeccc4666bf429326a4c5e174c5458ae3727316/"
    },
    "refs/pull/723/head": {
      "target": "16192132ba66e8e626cc868639cbd0d74fdfb84e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/16192132ba66e8e626cc868639cbd0d74fdfb84e/"
    },
    "refs/pull/724/head": {
      "target": "72248ba02c23b63b332786cd5e55affb829f91b5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/72248ba02c23b63b332786cd5e55affb829f91b5/"
    },
    "refs/pull/725/head": {
      "target": "f7a726e28401f5b39aa4765935d95ff1f8da1cf0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f7a726e28401f5b39aa4765935d95ff1f8da1cf0/"
    },
    "refs/pull/726/head": {
      "target": "f1132ba3465d56622c286beef6b511fd55c6482b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f1132ba3465d56622c286beef6b511fd55c6482b/"
    },
    "refs/pull/727/head": {
      "target": "9d5dc645b19a2eaaaebfcfd3df4f8f8441021ebe",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9d5dc645b19a2eaaaebfcfd3df4f8f8441021ebe/"
    },
    "refs/pull/728/head": {
      "target": "81acaceb1af69f913030f99eee25def62eded32a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/81acaceb1af69f913030f99eee25def62eded32a/"
    },
    "refs/pull/729/head": {
      "target": "5e31091595472c6694991eb0fb2aab33a50e00db",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5e31091595472c6694991eb0fb2aab33a50e00db/"
    },
    "refs/pull/73/head": {
      "target": "8f6bad2f9ac0fdaa8ae341763890d9012a183683",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8f6bad2f9ac0fdaa8ae341763890d9012a183683/"
    },
    "refs/pull/730/head": {
      "target": "55fc77d70b3887e146f04c474909083cc0d1bc58",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/55fc77d70b3887e146f04c474909083cc0d1bc58/"
    },
    "refs/pull/731/head": {
      "target": "c982707af6da8db446813dcfc4409a077c014a33",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c982707af6da8db446813dcfc4409a077c014a33/"
    },
    "refs/pull/732/head": {
      "target": "03aa248c5aa61f774deacea63018c46b845afcc7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/03aa248c5aa61f774deacea63018c46b845afcc7/"
    },
    "refs/pull/733/head": {
      "target": "c453f1c688de581e7f188f27a0d89a2a320129d6",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c453f1c688de581e7f188f27a0d89a2a320129d6/"
    },
    "refs/pull/734/head": {
      "target": "24270a05816eaa98dd8946725c87e0a21ce99905",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/24270a05816eaa98dd8946725c87e0a21ce99905/"
    },
    "refs/pull/735/head": {
      "target": "a7aa298f89805660cdc781a45750fbe23af2d657",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a7aa298f89805660cdc781a45750fbe23af2d657/"
    },
    "refs/pull/736/head": {
      "target": "7d7fc6abcb255100164824795631fadb920fad42",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7d7fc6abcb255100164824795631fadb920fad42/"
    },
    "refs/pull/737/head": {
      "target": "682b17156ecd2b91a8794762203de092eed1e991",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/682b17156ecd2b91a8794762203de092eed1e991/"
    },
    "refs/pull/738/head": {
      "target": "e77bccf6a1a3cf5790347af5fec38d0728af2e9d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e77bccf6a1a3cf5790347af5fec38d0728af2e9d/"
    },
    "refs/pull/739/head": {
      "target": "c7eecf1b10da6adb3518877808ff188b634061c2",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c7eecf1b10da6adb3518877808ff188b634061c2/"
    },
    "refs/pull/74/head": {
      "target": "3fc7f95d1c80ed60f98a0c7c09dcdb7d0ba90c79",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3fc7f95d1c80ed60f98a0c7c09dcdb7d0ba90c79/"
    },
    "refs/pull/740/head": {
      "target": "30c0923370175a01a806b23d797f1d9f2f3e4f40",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/30c0923370175a01a806b23d797f1d9f2f3e4f40/"
    },
    "refs/pull/741/head": {
      "target": "cbb52786e43509578ddbbda6ba8bcc7cd31567d8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cbb52786e43509578ddbbda6ba8bcc7cd31567d8/"
    },
    "refs/pull/742/head": {
      "target": "ff793b282529146762d314ded3e6836b3c3499a1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ff793b282529146762d314ded3e6836b3c3499a1/"
    },
    "refs/pull/743/head": {
      "target": "7195fdd2ba6c2d28caaf34cfd704d83d46b16d89",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7195fdd2ba6c2d28caaf34cfd704d83d46b16d89/"
    },
    "refs/pull/744/head": {
      "target": "6e2a364fdfc702b060e39c0be43b0dffdf6afb9e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6e2a364fdfc702b060e39c0be43b0dffdf6afb9e/"
    },
    "refs/pull/745/head": {
      "target": "cd5fa057c43f49343e747d2e19422ed77b2e8147",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cd5fa057c43f49343e747d2e19422ed77b2e8147/"
    },
    "refs/pull/746/head": {
      "target": "7ba12ca084f57a62e3caeb356ff77f2247ef7a6d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7ba12ca084f57a62e3caeb356ff77f2247ef7a6d/"
    },
    "refs/pull/747/head": {
      "target": "87ef8d523762af361c1fffcfce54f19151cf54aa",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/87ef8d523762af361c1fffcfce54f19151cf54aa/"
    },
    "refs/pull/748/head": {
      "target": "394aeaf6d35dd9202c9d12875fafcdeeb6ea40e4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/394aeaf6d35dd9202c9d12875fafcdeeb6ea40e4/"
    },
    "refs/pull/749/head": {
      "target": "ce59980af5ba0a673da38b2ff9efe89554cf373b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ce59980af5ba0a673da38b2ff9efe89554cf373b/"
    },
    "refs/pull/75/head": {
      "target": "a7bf7eb25d8e479050417f0c0e7187ef26c514ed",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a7bf7eb25d8e479050417f0c0e7187ef26c514ed/"
    },
    "refs/pull/750/head": {
      "target": "0b77b97066628deef03a6eac7b29e5c3756f4ab0",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0b77b97066628deef03a6eac7b29e5c3756f4ab0/"
    },
    "refs/pull/751/head": {
      "target": "44286992afd2e960a290f79ebeeb6d3957f4698b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/44286992afd2e960a290f79ebeeb6d3957f4698b/"
    },
    "refs/pull/752/head": {
      "target": "7a06c1dba37b8cb8b4f4bf663c4579bc14b3e8b4",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7a06c1dba37b8cb8b4f4bf663c4579bc14b3e8b4/"
    },
    "refs/pull/753/head": {
      "target": "63909ee865fe5a52538ed6357eac4874d02c081c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/63909ee865fe5a52538ed6357eac4874d02c081c/"
    },
    "refs/pull/754/head": {
      "target": "a9c946ee0a58877655eb1f6500aa83431207ff48",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a9c946ee0a58877655eb1f6500aa83431207ff48/"
    },
    "refs/pull/755/head": {
      "target": "51e588c515594386bd52f05c050ee07b182f8284",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/51e588c515594386bd52f05c050ee07b182f8284/"
    },
    "refs/pull/756/head": {
      "target": "35420ae6947d85fe02e172f73c4db6f824a1625b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/35420ae6947d85fe02e172f73c4db6f824a1625b/"
    },
    "refs/pull/757/head": {
      "target": "c2ccca7bfc122a892a1560b7b290406e09aabf9e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c2ccca7bfc122a892a1560b7b290406e09aabf9e/"
    },
    "refs/pull/758/head": {
      "target": "e1d54a9a405576e68820a019706b88fa291104ca",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e1d54a9a405576e68820a019706b88fa291104ca/"
    },
    "refs/pull/759/head": {
      "target": "6a7ef73882393e23e966e7b53e7b03397ad18e63",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/6a7ef73882393e23e966e7b53e7b03397ad18e63/"
    },
    "refs/pull/76/head": {
      "target": "a4a39c784d9829a2c199de27663571869cbd3342",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a4a39c784d9829a2c199de27663571869cbd3342/"
    },
    "refs/pull/760/head": {
      "target": "842740808c3b85945a472a181c2b1467dfd1734a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/842740808c3b85945a472a181c2b1467dfd1734a/"
    },
    "refs/pull/761/head": {
      "target": "840c05ab4e2b71ad6eecb1bc9332e6c75477259e",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/840c05ab4e2b71ad6eecb1bc9332e6c75477259e/"
    },
    "refs/pull/762/head": {
      "target": "de3c4c71a9c51d3cdbb0c7598d4ac18412ad7ecc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/de3c4c71a9c51d3cdbb0c7598d4ac18412ad7ecc/"
    },
    "refs/pull/763/head": {
      "target": "dd70e93a100c5f49455e5936a6d855b87c038642",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/dd70e93a100c5f49455e5936a6d855b87c038642/"
    },
    "refs/pull/764/head": {
      "target": "e4d9a314a248d39778b6295761604f58455ead3a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e4d9a314a248d39778b6295761604f58455ead3a/"
    },
    "refs/pull/765/head": {
      "target": "1ab5b111f6f7173f4af96d795613ec61207ef797",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1ab5b111f6f7173f4af96d795613ec61207ef797/"
    },
    "refs/pull/766/head": {
      "target": "7e93a673fc3e5a9ea3b01468d7dfd183e6ba0877",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/7e93a673fc3e5a9ea3b01468d7dfd183e6ba0877/"
    },
    "refs/pull/767/head": {
      "target": "87756b070112a67d6a19d7ce410e01106069e403",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/87756b070112a67d6a19d7ce410e01106069e403/"
    },
    "refs/pull/768/head": {
      "target": "d0de9fd69463461087a0abcf561eb308035a870a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/d0de9fd69463461087a0abcf561eb308035a870a/"
    },
    "refs/pull/769/head": {
      "target": "0dd44502893fe3e510d47ed057429df29b3c1507",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/0dd44502893fe3e510d47ed057429df29b3c1507/"
    },
    "refs/pull/77/head": {
      "target": "b35078aef12640df0ca99a3ad8c57ff56d43154b",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/b35078aef12640df0ca99a3ad8c57ff56d43154b/"
    },
    "refs/pull/770/head": {
      "target": "9521e379f9bda7270f674ca4a0f144df7abedfbc",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/9521e379f9bda7270f674ca4a0f144df7abedfbc/"
    },
    "refs/pull/771/head": {
      "target": "754168e682b4f8556c9aae1c8f427e667a1bea0a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/754168e682b4f8556c9aae1c8f427e667a1bea0a/"
    },
    "refs/pull/772/head": {
      "target": "ba77ea0776f17add36327be5843c94af488dca7d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ba77ea0776f17add36327be5843c94af488dca7d/"
    },
    "refs/pull/773/head": {
      "target": "3442d5c2171ddc3e765d7c1e5e463cb412478779",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/3442d5c2171ddc3e765d7c1e5e463cb412478779/"
    },
    "refs/pull/774/head": {
      "target": "f52501d02b209af0c7928253470271663e2b4839",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f52501d02b209af0c7928253470271663e2b4839/"
    },
    "refs/pull/775/head": {
      "target": "ef216f43df89569ebbbe6e7922f95b0ab6fd3846",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/ef216f43df89569ebbbe6e7922f95b0ab6fd3846/"
    },
    "refs/pull/776/head": {
      "target": "75c167afefd5aae1358fb15205d161f801339352",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/75c167afefd5aae1358fb15205d161f801339352/"
    },
    "refs/pull/777/head": {
      "target": "accb6466e9ef917f5e74b2a3cae9ae2832673f99",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/accb6466e9ef917f5e74b2a3cae9ae2832673f99/"
    },
    "refs/pull/778/head": {
      "target": "12db14909e0cbd1b2d5db816e7c8abd04d27f1c5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/12db14909e0cbd1b2d5db816e7c8abd04d27f1c5/"
    },
    "refs/pull/779/head": {
      "target": "8042783592c5393256b1009ee35a721c0e03571a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8042783592c5393256b1009ee35a721c0e03571a/"
    },
    "refs/pull/78/head": {
      "target": "cf484c9c362d9f79ac3c7ced2ede6d326099bf74",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cf484c9c362d9f79ac3c7ced2ede6d326099bf74/"
    },
    "refs/pull/8/head": {
      "target": "cc0f90ce6e6a729f63ecf34862960b62c5bd22f5",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/cc0f90ce6e6a729f63ecf34862960b62c5bd22f5/"
    },
    "refs/pull/80/head": {
      "target": "a6bff87fa6cb5c9b8fb67fc7c517b75deafc6792",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a6bff87fa6cb5c9b8fb67fc7c517b75deafc6792/"
    },
    "refs/pull/81/head": {
      "target": "4651d70f4af9134396a14a28bb9b664fccef8368",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/4651d70f4af9134396a14a28bb9b664fccef8368/"
    },
    "refs/pull/82/head": {
      "target": "8cae4b5037591437905d46ffc8e0a5e82e8b8090",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8cae4b5037591437905d46ffc8e0a5e82e8b8090/"
    },
    "refs/pull/83/head": {
      "target": "1434a7e347c3e9450e7062fe5e097b132d5191c8",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/1434a7e347c3e9450e7062fe5e097b132d5191c8/"
    },
    "refs/pull/84/head": {
      "target": "70695c26e9f04d76634f08704da43859ab328240",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/70695c26e9f04d76634f08704da43859ab328240/"
    },
    "refs/pull/85/head": {
      "target": "e28d69f4c575f615198bc7b46138bdf1ab41f817",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/e28d69f4c575f615198bc7b46138bdf1ab41f817/"
    },
    "refs/pull/86/head": {
      "target": "5eee762fddae39e50a3a2c8b49223f9c87bd9175",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/5eee762fddae39e50a3a2c8b49223f9c87bd9175/"
    },
    "refs/pull/87/head": {
      "target": "fc9046a105265355bec57972498ea74572cff4b3",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fc9046a105265355bec57972498ea74572cff4b3/"
    },
    "refs/pull/88/head": {
      "target": "a798c10faf62a505d24e5f6213fbaf904a39623f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/a798c10faf62a505d24e5f6213fbaf904a39623f/"
    },
    "refs/pull/89/head": {
      "target": "8f7fab2980fee3eab1dc5e782dfbbc3fa5b83b5d",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/8f7fab2980fee3eab1dc5e782dfbbc3fa5b83b5d/"
    },
    "refs/pull/9/head": {
      "target": "81a5f50f6b5356607dd8bb3bebdca24e6d13ef9c",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/81a5f50f6b5356607dd8bb3bebdca24e6d13ef9c/"
    },
    "refs/pull/90/head": {
      "target": "dffcfb0469b72afa6443bf2952e04bc7e0655170",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/dffcfb0469b72afa6443bf2952e04bc7e0655170/"
    },
    "refs/pull/92/head": {
      "target": "399f4591bf9eaec1c2dffaaf67cc48a45bcb5ae1",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/399f4591bf9eaec1c2dffaaf67cc48a45bcb5ae1/"
    },
    "refs/pull/93/head": {
      "target": "479a57f4a1d4dd37f01db1d49730e034d786d6b7",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/479a57f4a1d4dd37f01db1d49730e034d786d6b7/"
    },
    "refs/pull/94/head": {
      "target": "afd28540c537772552a40d949f88730a1c9f8766",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/afd28540c537772552a40d949f88730a1c9f8766/"
    },
    "refs/pull/95/head": {
      "target": "f5a3a81bb154aaa370daef42f2b1cae253525b62",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/f5a3a81bb154aaa370daef42f2b1cae253525b62/"
    },
    "refs/pull/96/head": {
      "target": "67860fdf9d18481668edabfafec8173f33de0e2a",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/67860fdf9d18481668edabfafec8173f33de0e2a/"
    },
    "refs/pull/97/head": {
      "target": "2920af181a9a345aaaa29e8296f677956afd6c6f",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/2920af181a9a345aaaa29e8296f677956afd6c6f/"
    },
    "refs/pull/98/head": {
      "target": "fdb9c7fdd5bbcc62b53508ced9b2203edb27acca",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/fdb9c7fdd5bbcc62b53508ced9b2203edb27acca/"
    },
    "refs/pull/99/head": {
      "target": "c195df80e7aa5c290b232155ad78d919a2e29cdd",
      "target_type": "revision",
      "target_url": "https://archive.softwareheritage.org/api/1/revision/c195df80e7aa5c290b232155ad78d919a2e29cdd/"
    },
    "refs/tags/v2.6.11": {
      "target": "5dc01c595e6c6ec9ccda4f6f69c131c0dd945f8c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5dc01c595e6c6ec9ccda4f6f69c131c0dd945f8c/"
    },
    "refs/tags/v2.6.11-tree": {
      "target": "5dc01c595e6c6ec9ccda4f6f69c131c0dd945f8c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5dc01c595e6c6ec9ccda4f6f69c131c0dd945f8c/"
    },
    "refs/tags/v2.6.12": {
      "target": "26791a8bcf0e6d33f43aef7682bdb555236d56de",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/26791a8bcf0e6d33f43aef7682bdb555236d56de/"
    },
    "refs/tags/v2.6.12-rc2": {
      "target": "9e734775f7c22d2f89943ad6c745571f1930105f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9e734775f7c22d2f89943ad6c745571f1930105f/"
    },
    "refs/tags/v2.6.12-rc3": {
      "target": "0397236d43e48e821cce5bbe6a80a1a56bb7cc3a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0397236d43e48e821cce5bbe6a80a1a56bb7cc3a/"
    },
    "refs/tags/v2.6.12-rc4": {
      "target": "ebb5573ea8beaf000d4833735f3e53acb9af844c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ebb5573ea8beaf000d4833735f3e53acb9af844c/"
    },
    "refs/tags/v2.6.12-rc5": {
      "target": "06f6d9e2f140466eeb41e494e14167f90210f89d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/06f6d9e2f140466eeb41e494e14167f90210f89d/"
    },
    "refs/tags/v2.6.12-rc6": {
      "target": "701d7ecec3e0c6b4ab9bb824fd2b34be4da63b7e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/701d7ecec3e0c6b4ab9bb824fd2b34be4da63b7e/"
    },
    "refs/tags/v2.6.13": {
      "target": "0da688d20078783b23f99b232b272b027d6c3f59",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0da688d20078783b23f99b232b272b027d6c3f59/"
    },
    "refs/tags/v2.6.13-rc1": {
      "target": "733ad933f62e82ebc92fed988c7f0795e64dea62",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/733ad933f62e82ebc92fed988c7f0795e64dea62/"
    },
    "refs/tags/v2.6.13-rc2": {
      "target": "c521cb0f10ef2bf28a18e1cc8adf378ccbbe5a19",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c521cb0f10ef2bf28a18e1cc8adf378ccbbe5a19/"
    },
    "refs/tags/v2.6.13-rc3": {
      "target": "a339981ec18d304f9efeb9ccf01b1f04302edf32",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a339981ec18d304f9efeb9ccf01b1f04302edf32/"
    },
    "refs/tags/v2.6.13-rc4": {
      "target": "7eab951de91d95875ba34ec4c599f37e1208db93",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7eab951de91d95875ba34ec4c599f37e1208db93/"
    },
    "refs/tags/v2.6.13-rc5": {
      "target": "2ef68d3fdfacf92ac5ff84be969b4b8acb5c0723",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2ef68d3fdfacf92ac5ff84be969b4b8acb5c0723/"
    },
    "refs/tags/v2.6.13-rc6": {
      "target": "50f0e20a16496a9feb52cf1d79a0c33cce1a682d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/50f0e20a16496a9feb52cf1d79a0c33cce1a682d/"
    },
    "refs/tags/v2.6.13-rc7": {
      "target": "3e6f704488332c74e7defec13f22bf2e2bccb444",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3e6f704488332c74e7defec13f22bf2e2bccb444/"
    },
    "refs/tags/v2.6.14": {
      "target": "2b10839e32c4c476e9d94492756bb1a3e1ec4aa8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2b10839e32c4c476e9d94492756bb1a3e1ec4aa8/"
    },
    "refs/tags/v2.6.14-rc1": {
      "target": "1f9d1e3248d4eb96b229eecf0e5d9445d3529e85",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1f9d1e3248d4eb96b229eecf0e5d9445d3529e85/"
    },
    "refs/tags/v2.6.14-rc2": {
      "target": "c2bbf523f1d454649897b3e4bcd71778e4fa5913",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c2bbf523f1d454649897b3e4bcd71778e4fa5913/"
    },
    "refs/tags/v2.6.14-rc3": {
      "target": "f92737b18abac90af30ac26a050fda879c9b238b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f92737b18abac90af30ac26a050fda879c9b238b/"
    },
    "refs/tags/v2.6.14-rc4": {
      "target": "a1aeb6f9e4306b1a62629c56197eea65afd372b3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a1aeb6f9e4306b1a62629c56197eea65afd372b3/"
    },
    "refs/tags/v2.6.14-rc5": {
      "target": "c08359b174c3b3a175979f81cc786c791f74c852",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c08359b174c3b3a175979f81cc786c791f74c852/"
    },
    "refs/tags/v2.6.15": {
      "target": "dab47a31f42a23d2b374e1cd7d0b797e8e08b23d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/dab47a31f42a23d2b374e1cd7d0b797e8e08b23d/"
    },
    "refs/tags/v2.6.15-rc1": {
      "target": "1400e40b758e4099e4d758d8fc38981b33b71897",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1400e40b758e4099e4d758d8fc38981b33b71897/"
    },
    "refs/tags/v2.6.15-rc2": {
      "target": "7305b5cb045e2c71250b5b7472771ed2620bc514",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7305b5cb045e2c71250b5b7472771ed2620bc514/"
    },
    "refs/tags/v2.6.15-rc3": {
      "target": "00d7a7358e3f9f2575501674e604fe4c6700b365",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/00d7a7358e3f9f2575501674e604fe4c6700b365/"
    },
    "refs/tags/v2.6.15-rc4": {
      "target": "8c27e9cccbc96548646606a2eb52c27e72abf5e1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8c27e9cccbc96548646606a2eb52c27e72abf5e1/"
    },
    "refs/tags/v2.6.15-rc5": {
      "target": "b58cb01b0f7eff054d1246e63f0eebfbdf8e6686",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b58cb01b0f7eff054d1246e63f0eebfbdf8e6686/"
    },
    "refs/tags/v2.6.15-rc6": {
      "target": "20f065bc620db2d6157c95b117910c4ed59acc46",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/20f065bc620db2d6157c95b117910c4ed59acc46/"
    },
    "refs/tags/v2.6.15-rc7": {
      "target": "373540886574ecf8b864fa7b8a5cf879b1d7b054",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/373540886574ecf8b864fa7b8a5cf879b1d7b054/"
    },
    "refs/tags/v2.6.16": {
      "target": "414ad0ded83f088608f7c0e774df8cccbba4e229",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/414ad0ded83f088608f7c0e774df8cccbba4e229/"
    },
    "refs/tags/v2.6.16-rc1": {
      "target": "f3bcf72eb85aba88a7bd0a6116dd0b5418590dbe",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f3bcf72eb85aba88a7bd0a6116dd0b5418590dbe/"
    },
    "refs/tags/v2.6.16-rc2": {
      "target": "e4f9aae0d74cb7d2fd5f0eb315cf9de1118fe260",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e4f9aae0d74cb7d2fd5f0eb315cf9de1118fe260/"
    },
    "refs/tags/v2.6.16-rc3": {
      "target": "a3adf278bcd5852d8c6345c9b0a44a0cc71c7454",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a3adf278bcd5852d8c6345c9b0a44a0cc71c7454/"
    },
    "refs/tags/v2.6.16-rc4": {
      "target": "d94ef3bc908905e8bc0f1560a90df76a4c4b60bb",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d94ef3bc908905e8bc0f1560a90df76a4c4b60bb/"
    },
    "refs/tags/v2.6.16-rc5": {
      "target": "afdf555c717a6ace9e5446dda4d169eaa5ca8062",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/afdf555c717a6ace9e5446dda4d169eaa5ca8062/"
    },
    "refs/tags/v2.6.16-rc6": {
      "target": "76dbe3f496eeaadab8977409b930e7d522dcf750",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/76dbe3f496eeaadab8977409b930e7d522dcf750/"
    },
    "refs/tags/v2.6.17": {
      "target": "8ba130df4b67fa40878ccf80d54615132d24bc68",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8ba130df4b67fa40878ccf80d54615132d24bc68/"
    },
    "refs/tags/v2.6.17-rc1": {
      "target": "d882e0c80e6e3c60640492b83395e6fbbae04276",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d882e0c80e6e3c60640492b83395e6fbbae04276/"
    },
    "refs/tags/v2.6.17-rc2": {
      "target": "f61c8059ffbc29bd8a1ffbd5a87e5135bc28a752",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f61c8059ffbc29bd8a1ffbd5a87e5135bc28a752/"
    },
    "refs/tags/v2.6.17-rc3": {
      "target": "6716c37ec2dbf78b85a55cc5605677b6cf2299a0",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6716c37ec2dbf78b85a55cc5605677b6cf2299a0/"
    },
    "refs/tags/v2.6.17-rc4": {
      "target": "90b92312eeebc70e61415394be3cc03b08a74945",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/90b92312eeebc70e61415394be3cc03b08a74945/"
    },
    "refs/tags/v2.6.17-rc5": {
      "target": "39beb382e4e11ed01cb5e73022f18bbed2aefd8b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/39beb382e4e11ed01cb5e73022f18bbed2aefd8b/"
    },
    "refs/tags/v2.6.17-rc6": {
      "target": "831695cbeb7a0e8f9ddb9c0203a22723da2c3f2f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/831695cbeb7a0e8f9ddb9c0203a22723da2c3f2f/"
    },
    "refs/tags/v2.6.18": {
      "target": "119248f4578ca60b09c20893724e10f19806e6f1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/119248f4578ca60b09c20893724e10f19806e6f1/"
    },
    "refs/tags/v2.6.18-rc1": {
      "target": "7df8ea909888d4856d3aff1c41192739d715a393",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7df8ea909888d4856d3aff1c41192739d715a393/"
    },
    "refs/tags/v2.6.18-rc2": {
      "target": "f4035c6284f7de0b47ccf68264f289e7e0b47da2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f4035c6284f7de0b47ccf68264f289e7e0b47da2/"
    },
    "refs/tags/v2.6.18-rc3": {
      "target": "5031a46c0d098346e4c9cd56f7858def243da096",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5031a46c0d098346e4c9cd56f7858def243da096/"
    },
    "refs/tags/v2.6.18-rc4": {
      "target": "06abe048193087f81ee67c225a1b685d937f220d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/06abe048193087f81ee67c225a1b685d937f220d/"
    },
    "refs/tags/v2.6.18-rc5": {
      "target": "69b7510036d06c6aabcdc0bb03f35cfc62160302",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/69b7510036d06c6aabcdc0bb03f35cfc62160302/"
    },
    "refs/tags/v2.6.18-rc6": {
      "target": "12c8b6ff70a17bfcb59bcca55fa8e79d809c1c70",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/12c8b6ff70a17bfcb59bcca55fa8e79d809c1c70/"
    },
    "refs/tags/v2.6.18-rc7": {
      "target": "3d3f47c98b3696f5bd677f89a755c239f4d52a4c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3d3f47c98b3696f5bd677f89a755c239f4d52a4c/"
    },
    "refs/tags/v2.6.19": {
      "target": "c3fe6924620fd733ffe8bc8a9da1e9cde08402b3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c3fe6924620fd733ffe8bc8a9da1e9cde08402b3/"
    },
    "refs/tags/v2.6.19-rc1": {
      "target": "008616c979c90b1a4ea2d2cebf4996743463b001",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/008616c979c90b1a4ea2d2cebf4996743463b001/"
    },
    "refs/tags/v2.6.19-rc2": {
      "target": "5e2c92804ed42d0be784e4dc4041dc3ca001810b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5e2c92804ed42d0be784e4dc4041dc3ca001810b/"
    },
    "refs/tags/v2.6.19-rc3": {
      "target": "aaa751e8aa7dd8b1c1f787728d279d9dadbb5938",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/aaa751e8aa7dd8b1c1f787728d279d9dadbb5938/"
    },
    "refs/tags/v2.6.19-rc4": {
      "target": "7a9d289b6650bf78df77ab463bedc2919df89833",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7a9d289b6650bf78df77ab463bedc2919df89833/"
    },
    "refs/tags/v2.6.19-rc5": {
      "target": "2f4871018917f13c7cd01254335561a152decc00",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2f4871018917f13c7cd01254335561a152decc00/"
    },
    "refs/tags/v2.6.19-rc6": {
      "target": "82471364173618c5a97b6c02bf6e72deddde9632",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/82471364173618c5a97b6c02bf6e72deddde9632/"
    },
    "refs/tags/v2.6.20": {
      "target": "fa285a3d7924a0e3782926e51f16865c5129a2f7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fa285a3d7924a0e3782926e51f16865c5129a2f7/"
    },
    "refs/tags/v2.6.20-rc1": {
      "target": "a7ecdd29e85a7d51179c6a4507a4c25a87ab5c0e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a7ecdd29e85a7d51179c6a4507a4c25a87ab5c0e/"
    },
    "refs/tags/v2.6.20-rc2": {
      "target": "8a2d17a56a71c5c796b0a5378ee76a105f21fdd9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8a2d17a56a71c5c796b0a5378ee76a105f21fdd9/"
    },
    "refs/tags/v2.6.20-rc3": {
      "target": "9d37e643a198d4232d2a12f1107358eb02b6cece",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9d37e643a198d4232d2a12f1107358eb02b6cece/"
    },
    "refs/tags/v2.6.20-rc4": {
      "target": "cb15cca892ac09f7cb99618a09304c5f5620e6ee",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cb15cca892ac09f7cb99618a09304c5f5620e6ee/"
    },
    "refs/tags/v2.6.20-rc5": {
      "target": "c81c70d32732b7abc644ec7aca24f3597aa921ea",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c81c70d32732b7abc644ec7aca24f3597aa921ea/"
    },
    "refs/tags/v2.6.20-rc6": {
      "target": "0909eebced6eadf15ea6aadce0d79c08676de3fe",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0909eebced6eadf15ea6aadce0d79c08676de3fe/"
    },
    "refs/tags/v2.6.20-rc7": {
      "target": "97c617f72b58983017f32da97fc228597c0dae8f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/97c617f72b58983017f32da97fc228597c0dae8f/"
    },
    "refs/tags/v2.6.21": {
      "target": "d1be341dba5521506d9e6dccfd66179080705bea",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d1be341dba5521506d9e6dccfd66179080705bea/"
    },
    "refs/tags/v2.6.21-rc1": {
      "target": "2eb1ae149a28c1b8ade687c5fbab3c37da4c0fba",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2eb1ae149a28c1b8ade687c5fbab3c37da4c0fba/"
    },
    "refs/tags/v2.6.21-rc2": {
      "target": "fa18364691754673df97824d81bfe8524a3a0595",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fa18364691754673df97824d81bfe8524a3a0595/"
    },
    "refs/tags/v2.6.21-rc3": {
      "target": "44e05067b4b767fa3137335818c399b9b198c925",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/44e05067b4b767fa3137335818c399b9b198c925/"
    },
    "refs/tags/v2.6.21-rc4": {
      "target": "bac6eefe96204d0ad67d144f2511a6fc487aa594",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/bac6eefe96204d0ad67d144f2511a6fc487aa594/"
    },
    "refs/tags/v2.6.21-rc5": {
      "target": "6fb04ccf5c5e054c4107090bed6e866489f1089f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6fb04ccf5c5e054c4107090bed6e866489f1089f/"
    },
    "refs/tags/v2.6.21-rc6": {
      "target": "a5d693edb558469a6f72bfda1253c7ba2278d657",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a5d693edb558469a6f72bfda1253c7ba2278d657/"
    },
    "refs/tags/v2.6.21-rc7": {
      "target": "60afa917fea11f7cc93281b2dfd64b19b6521a93",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/60afa917fea11f7cc93281b2dfd64b19b6521a93/"
    },
    "refs/tags/v2.6.22": {
      "target": "098fd16f00005f665d3baa7e682d8cb3d7c0fe6f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/098fd16f00005f665d3baa7e682d8cb3d7c0fe6f/"
    },
    "refs/tags/v2.6.22-rc1": {
      "target": "cb22632a37a5d797da988453924206d1638e4e6c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cb22632a37a5d797da988453924206d1638e4e6c/"
    },
    "refs/tags/v2.6.22-rc2": {
      "target": "5a1b8597cd250efd5bda1cba08417c95b6b314d7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5a1b8597cd250efd5bda1cba08417c95b6b314d7/"
    },
    "refs/tags/v2.6.22-rc3": {
      "target": "0119a8416e0827bf2b937a1cf21d4909db3fa111",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0119a8416e0827bf2b937a1cf21d4909db3fa111/"
    },
    "refs/tags/v2.6.22-rc4": {
      "target": "5b78c77092a64e253fe1fde9fbbe818b49330ffc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5b78c77092a64e253fe1fde9fbbe818b49330ffc/"
    },
    "refs/tags/v2.6.22-rc5": {
      "target": "aec07c7abc280bd5d0ca33b7cda3eb7b9b6e89c1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/aec07c7abc280bd5d0ca33b7cda3eb7b9b6e89c1/"
    },
    "refs/tags/v2.6.22-rc6": {
      "target": "953e420db7c599f7db00548243f2afddc8440329",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/953e420db7c599f7db00548243f2afddc8440329/"
    },
    "refs/tags/v2.6.22-rc7": {
      "target": "087ea061253277de2b27e82d8572a386835a1b7e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/087ea061253277de2b27e82d8572a386835a1b7e/"
    },
    "refs/tags/v2.6.23": {
      "target": "0b8bc8b91cf6befea20fe78b90367ca7b61cfa0d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0b8bc8b91cf6befea20fe78b90367ca7b61cfa0d/"
    },
    "refs/tags/v2.6.23-rc1": {
      "target": "7d57c74238cdf570bca20b711b2c0b31a553c1e5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7d57c74238cdf570bca20b711b2c0b31a553c1e5/"
    },
    "refs/tags/v2.6.23-rc2": {
      "target": "2c7522b19c386ed601d27b2aed3e7b84ac7852f0",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2c7522b19c386ed601d27b2aed3e7b84ac7852f0/"
    },
    "refs/tags/v2.6.23-rc3": {
      "target": "39e7e2ec80646a62a9d61871bee8d2736088a86f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/39e7e2ec80646a62a9d61871bee8d2736088a86f/"
    },
    "refs/tags/v2.6.23-rc4": {
      "target": "e7afef45b41f5f5e2211322f083247e42ba13a78",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e7afef45b41f5f5e2211322f083247e42ba13a78/"
    },
    "refs/tags/v2.6.23-rc5": {
      "target": "f3cfc7abf10379d926dd04008059d2b04eaf4499",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f3cfc7abf10379d926dd04008059d2b04eaf4499/"
    },
    "refs/tags/v2.6.23-rc6": {
      "target": "a33969a68b624d98356398af0a59856cc52f47a5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a33969a68b624d98356398af0a59856cc52f47a5/"
    },
    "refs/tags/v2.6.23-rc7": {
      "target": "c3e1edc6a6b420f81a6bc1ea47c5b3dd157e76aa",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c3e1edc6a6b420f81a6bc1ea47c5b3dd157e76aa/"
    },
    "refs/tags/v2.6.23-rc8": {
      "target": "fc4a2ad046f06bed41eda33142c5767149a72fe7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fc4a2ad046f06bed41eda33142c5767149a72fe7/"
    },
    "refs/tags/v2.6.23-rc9": {
      "target": "da0a81e98c06aa0d1e05b9012c2b2facb1807e12",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/da0a81e98c06aa0d1e05b9012c2b2facb1807e12/"
    },
    "refs/tags/v2.6.24": {
      "target": "0d733ddb2026683da26c1722847b99911c43ccb5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0d733ddb2026683da26c1722847b99911c43ccb5/"
    },
    "refs/tags/v2.6.24-rc1": {
      "target": "cebdeed27b068dcc3e7c311d7ec0d9c33b5138c2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cebdeed27b068dcc3e7c311d7ec0d9c33b5138c2/"
    },
    "refs/tags/v2.6.24-rc2": {
      "target": "9aae299f7fd1888ea3a195cfe0edef17bb647415",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9aae299f7fd1888ea3a195cfe0edef17bb647415/"
    },
    "refs/tags/v2.6.24-rc3": {
      "target": "f05092637dc0d9a3f2249c9b283b973e6e96b7d2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f05092637dc0d9a3f2249c9b283b973e6e96b7d2/"
    },
    "refs/tags/v2.6.24-rc4": {
      "target": "b6fa40f5916811c6aad6625c384d26fd01135014",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b6fa40f5916811c6aad6625c384d26fd01135014/"
    },
    "refs/tags/v2.6.24-rc5": {
      "target": "9f11d5919577129413e8389e43e5b6e8413dff53",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9f11d5919577129413e8389e43e5b6e8413dff53/"
    },
    "refs/tags/v2.6.24-rc6": {
      "target": "f49e4e249d57ddfa97e046bc5c994ef72c93e63b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f49e4e249d57ddfa97e046bc5c994ef72c93e63b/"
    },
    "refs/tags/v2.6.24-rc7": {
      "target": "fcb31af14662059db467201ec73dfbb6f3300342",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fcb31af14662059db467201ec73dfbb6f3300342/"
    },
    "refs/tags/v2.6.24-rc8": {
      "target": "c9ba0caa9650a1898c839a79f6ff96a8a982424c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c9ba0caa9650a1898c839a79f6ff96a8a982424c/"
    },
    "refs/tags/v2.6.25": {
      "target": "20b8df8e5501bac243e64c0c8c52907735a0041b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/20b8df8e5501bac243e64c0c8c52907735a0041b/"
    },
    "refs/tags/v2.6.25-rc1": {
      "target": "abf6976c818c553eb2209fe32028a4c5eecab0cb",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/abf6976c818c553eb2209fe32028a4c5eecab0cb/"
    },
    "refs/tags/v2.6.25-rc2": {
      "target": "b74415eac8d3f1fcb39ad4bcef0c829635a3bc9f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b74415eac8d3f1fcb39ad4bcef0c829635a3bc9f/"
    },
    "refs/tags/v2.6.25-rc3": {
      "target": "d622f5379e88a3bac4f8decfa49c0a04a8e209d3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d622f5379e88a3bac4f8decfa49c0a04a8e209d3/"
    },
    "refs/tags/v2.6.25-rc4": {
      "target": "c6c155e032361b0031943141b1a6f231e4f63817",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c6c155e032361b0031943141b1a6f231e4f63817/"
    },
    "refs/tags/v2.6.25-rc5": {
      "target": "cd81f35c48b7e0c2a871f88e1973f391f8330449",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cd81f35c48b7e0c2a871f88e1973f391f8330449/"
    },
    "refs/tags/v2.6.25-rc6": {
      "target": "b22f07f908a648c864b16d2ba71f03aba4b684c9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b22f07f908a648c864b16d2ba71f03aba4b684c9/"
    },
    "refs/tags/v2.6.25-rc7": {
      "target": "f4281310b609edd587922b7d4afa63e4b9a1ffd4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f4281310b609edd587922b7d4afa63e4b9a1ffd4/"
    },
    "refs/tags/v2.6.25-rc8": {
      "target": "e39586f39c2829d30f4ea6680a846dfe4aad2f2e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e39586f39c2829d30f4ea6680a846dfe4aad2f2e/"
    },
    "refs/tags/v2.6.25-rc9": {
      "target": "3df83da958163beeca00d1254f512fafd79a19ed",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3df83da958163beeca00d1254f512fafd79a19ed/"
    },
    "refs/tags/v2.6.26": {
      "target": "14650d6ec137e70b6c1918cdef235027c5156020",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/14650d6ec137e70b6c1918cdef235027c5156020/"
    },
    "refs/tags/v2.6.26-rc1": {
      "target": "d6b7f73ed134769c86966697e61b235b200cc4ae",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d6b7f73ed134769c86966697e61b235b200cc4ae/"
    },
    "refs/tags/v2.6.26-rc2": {
      "target": "b67fc588ce611ca847620bd1353bf2d68fc3027f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b67fc588ce611ca847620bd1353bf2d68fc3027f/"
    },
    "refs/tags/v2.6.26-rc3": {
      "target": "b041a30258df00c90ac1ed532cec3f25c00a3ce8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b041a30258df00c90ac1ed532cec3f25c00a3ce8/"
    },
    "refs/tags/v2.6.26-rc4": {
      "target": "e21868a4cdd93e5883ff61579d4cd799d1a3c244",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e21868a4cdd93e5883ff61579d4cd799d1a3c244/"
    },
    "refs/tags/v2.6.26-rc5": {
      "target": "9ab8267ac47ce50b932cc4b1cbd9b05e2faac8b7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9ab8267ac47ce50b932cc4b1cbd9b05e2faac8b7/"
    },
    "refs/tags/v2.6.26-rc6": {
      "target": "c047413a582cbf2c7e1b012458c1665d959703be",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c047413a582cbf2c7e1b012458c1665d959703be/"
    },
    "refs/tags/v2.6.26-rc7": {
      "target": "f40883d058ed196976285fc1fd5fd6c85dcb5bef",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f40883d058ed196976285fc1fd5fd6c85dcb5bef/"
    },
    "refs/tags/v2.6.26-rc8": {
      "target": "496a3db2bfb98f1e9c7b73514d8d25790f69f5fb",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/496a3db2bfb98f1e9c7b73514d8d25790f69f5fb/"
    },
    "refs/tags/v2.6.26-rc9": {
      "target": "c22689a6f45beff21b97df566e0da17b4fa9ec19",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c22689a6f45beff21b97df566e0da17b4fa9ec19/"
    },
    "refs/tags/v2.6.27": {
      "target": "4b5127df968616dee2f4775d795198878ef1638b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4b5127df968616dee2f4775d795198878ef1638b/"
    },
    "refs/tags/v2.6.27-rc1": {
      "target": "78e28361b194c98eaa987e368264c2209ca08976",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/78e28361b194c98eaa987e368264c2209ca08976/"
    },
    "refs/tags/v2.6.27-rc2": {
      "target": "b27c893faffea2a0c49cf6170d89c2a7aeba6598",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b27c893faffea2a0c49cf6170d89c2a7aeba6598/"
    },
    "refs/tags/v2.6.27-rc3": {
      "target": "4aca83081f7ddf04d74989cc9b5e149160096324",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4aca83081f7ddf04d74989cc9b5e149160096324/"
    },
    "refs/tags/v2.6.27-rc4": {
      "target": "0572569f90ba4175e0400f3be7b9f42e1c803e55",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0572569f90ba4175e0400f3be7b9f42e1c803e55/"
    },
    "refs/tags/v2.6.27-rc5": {
      "target": "d17d499628a7d4a34c11f54a203281773885a3e7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d17d499628a7d4a34c11f54a203281773885a3e7/"
    },
    "refs/tags/v2.6.27-rc6": {
      "target": "89c44b4a5ad50f8b85846b16af5f977f3861d197",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/89c44b4a5ad50f8b85846b16af5f977f3861d197/"
    },
    "refs/tags/v2.6.27-rc7": {
      "target": "cb1d73008533d7c4a949fc0867de2612dab23572",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cb1d73008533d7c4a949fc0867de2612dab23572/"
    },
    "refs/tags/v2.6.27-rc8": {
      "target": "7b4e06aa7c0952353096105f87be67615e4382f5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7b4e06aa7c0952353096105f87be67615e4382f5/"
    },
    "refs/tags/v2.6.27-rc9": {
      "target": "5a97794d66909dbe3282062d7637705bcd352815",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5a97794d66909dbe3282062d7637705bcd352815/"
    },
    "refs/tags/v2.6.28": {
      "target": "8a38e7fd7a30cd44be954f9a3b062e607cec5d41",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8a38e7fd7a30cd44be954f9a3b062e607cec5d41/"
    },
    "refs/tags/v2.6.28-rc1": {
      "target": "cb50773491b0066d0e55f31f8875d5678fa3f8ad",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cb50773491b0066d0e55f31f8875d5678fa3f8ad/"
    },
    "refs/tags/v2.6.28-rc2": {
      "target": "5eb14db1f80df4eb0ecb0976e47e8e287e3175fc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5eb14db1f80df4eb0ecb0976e47e8e287e3175fc/"
    },
    "refs/tags/v2.6.28-rc3": {
      "target": "31cb515c75388d457c2f318a0ee9606b3527852f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/31cb515c75388d457c2f318a0ee9606b3527852f/"
    },
    "refs/tags/v2.6.28-rc4": {
      "target": "b65a80a5ee7923355cbca669cead08e067fc7ada",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b65a80a5ee7923355cbca669cead08e067fc7ada/"
    },
    "refs/tags/v2.6.28-rc5": {
      "target": "68185b00cf91c1c4dcc761a2f3a1631562ed52f3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/68185b00cf91c1c4dcc761a2f3a1631562ed52f3/"
    },
    "refs/tags/v2.6.28-rc6": {
      "target": "b503092a16bdba0a418e155fe592521fc20855af",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b503092a16bdba0a418e155fe592521fc20855af/"
    },
    "refs/tags/v2.6.28-rc7": {
      "target": "1a0bff987b27da5181f112bcc60f34d6fbb7e67e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1a0bff987b27da5181f112bcc60f34d6fbb7e67e/"
    },
    "refs/tags/v2.6.28-rc8": {
      "target": "6fa7003fe34e9a8a31fb91754f3c289cc045564b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6fa7003fe34e9a8a31fb91754f3c289cc045564b/"
    },
    "refs/tags/v2.6.28-rc9": {
      "target": "7d4b1bcc5e7411fc9e63f610c16e5de8fe6dfde8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7d4b1bcc5e7411fc9e63f610c16e5de8fe6dfde8/"
    },
    "refs/tags/v2.6.29": {
      "target": "5dfd736f95b3d84a18b5bb8e50ac71f245438acf",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5dfd736f95b3d84a18b5bb8e50ac71f245438acf/"
    },
    "refs/tags/v2.6.29-rc1": {
      "target": "7a3862d6e9934ffe107fe7ddfbe2c63dba321793",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7a3862d6e9934ffe107fe7ddfbe2c63dba321793/"
    },
    "refs/tags/v2.6.29-rc2": {
      "target": "d31ce8060b0e875179ba5ca1d40475dc2a082cc7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d31ce8060b0e875179ba5ca1d40475dc2a082cc7/"
    },
    "refs/tags/v2.6.29-rc3": {
      "target": "8be00154b8e949bf4b89ac198aef9a247532ac2d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8be00154b8e949bf4b89ac198aef9a247532ac2d/"
    },
    "refs/tags/v2.6.29-rc4": {
      "target": "87c16e9e8bb74f14f4504305957e4346e7fc46ea",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/87c16e9e8bb74f14f4504305957e4346e7fc46ea/"
    },
    "refs/tags/v2.6.29-rc5": {
      "target": "1dcda2df87ba4ecc7988be7a45d01645e11c9f4c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1dcda2df87ba4ecc7988be7a45d01645e11c9f4c/"
    },
    "refs/tags/v2.6.29-rc6": {
      "target": "0715562512ca6cf14c1b8f08e09d5907118deaf0",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0715562512ca6cf14c1b8f08e09d5907118deaf0/"
    },
    "refs/tags/v2.6.29-rc7": {
      "target": "b21232ea962bbaf0e909365f4964f6cceb2ba8ce",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b21232ea962bbaf0e909365f4964f6cceb2ba8ce/"
    },
    "refs/tags/v2.6.29-rc8": {
      "target": "73e37758f6b500a67d918528204832cc8f256516",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/73e37758f6b500a67d918528204832cc8f256516/"
    },
    "refs/tags/v2.6.30": {
      "target": "fa9c4d0983f98945b32d6bd0dfc1ba1b02d3773c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fa9c4d0983f98945b32d6bd0dfc1ba1b02d3773c/"
    },
    "refs/tags/v2.6.30-rc1": {
      "target": "42ae7400074d449189d41fceb6d6f871490d7842",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/42ae7400074d449189d41fceb6d6f871490d7842/"
    },
    "refs/tags/v2.6.30-rc2": {
      "target": "7c941a7798a5169ee0dd69a9e8d5c40ceb702023",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7c941a7798a5169ee0dd69a9e8d5c40ceb702023/"
    },
    "refs/tags/v2.6.30-rc3": {
      "target": "b2fdb301af8f488952aaab7de3ff8d3294c3274f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b2fdb301af8f488952aaab7de3ff8d3294c3274f/"
    },
    "refs/tags/v2.6.30-rc4": {
      "target": "176c5e45fe4f1c83df9429b7c2668b41446baac2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/176c5e45fe4f1c83df9429b7c2668b41446baac2/"
    },
    "refs/tags/v2.6.30-rc5": {
      "target": "4266b00c9dc3e1e071bde0ebfeadc40bbc1e8316",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4266b00c9dc3e1e071bde0ebfeadc40bbc1e8316/"
    },
    "refs/tags/v2.6.30-rc6": {
      "target": "6492b02a211a5fd0fc92e68d171fc3644cda71a7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6492b02a211a5fd0fc92e68d171fc3644cda71a7/"
    },
    "refs/tags/v2.6.30-rc7": {
      "target": "efa3e68c670b745894255af9827b3902bbc9376e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/efa3e68c670b745894255af9827b3902bbc9376e/"
    },
    "refs/tags/v2.6.30-rc8": {
      "target": "e6c72abc9d239d788b0cdb20cb3d20ba04c33707",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e6c72abc9d239d788b0cdb20cb3d20ba04c33707/"
    },
    "refs/tags/v2.6.31": {
      "target": "a271b16ba5478acead8773ebe01ee9b6365154d8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a271b16ba5478acead8773ebe01ee9b6365154d8/"
    },
    "refs/tags/v2.6.31-rc1": {
      "target": "eeadf87f8411b42b9803312d2870aa424602a99c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/eeadf87f8411b42b9803312d2870aa424602a99c/"
    },
    "refs/tags/v2.6.31-rc2": {
      "target": "c5d511255186f0bba081f11cbe11c856cadaedf7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c5d511255186f0bba081f11cbe11c856cadaedf7/"
    },
    "refs/tags/v2.6.31-rc3": {
      "target": "c361304f4a17c05992b1cf68172a5ea8389649ee",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c361304f4a17c05992b1cf68172a5ea8389649ee/"
    },
    "refs/tags/v2.6.31-rc4": {
      "target": "12ae5c63b8a8ac621ddfc810f774d00cad44765a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/12ae5c63b8a8ac621ddfc810f774d00cad44765a/"
    },
    "refs/tags/v2.6.31-rc5": {
      "target": "948e7a0b0e000a8d646b72fcfd8ccaa047e046ab",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/948e7a0b0e000a8d646b72fcfd8ccaa047e046ab/"
    },
    "refs/tags/v2.6.31-rc6": {
      "target": "73e5ade3a45e2096db80f0e87d6d838d0499f0fe",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/73e5ade3a45e2096db80f0e87d6d838d0499f0fe/"
    },
    "refs/tags/v2.6.31-rc7": {
      "target": "dcd5e628b7dc7f393dd728999fb0fb73a96d5e1b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/dcd5e628b7dc7f393dd728999fb0fb73a96d5e1b/"
    },
    "refs/tags/v2.6.31-rc8": {
      "target": "da1ee8b71154887fe5298ae683c089050f2c49a5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/da1ee8b71154887fe5298ae683c089050f2c49a5/"
    },
    "refs/tags/v2.6.31-rc9": {
      "target": "0827eb5df631ed1756653faf74248a9524f202cd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0827eb5df631ed1756653faf74248a9524f202cd/"
    },
    "refs/tags/v2.6.32": {
      "target": "459b3d520991ec1b8e5ba68fbc4b206d602fee6e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/459b3d520991ec1b8e5ba68fbc4b206d602fee6e/"
    },
    "refs/tags/v2.6.32-rc1": {
      "target": "1016bf08944977a33d3a48edc15ee34b425f6d8a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1016bf08944977a33d3a48edc15ee34b425f6d8a/"
    },
    "refs/tags/v2.6.32-rc2": {
      "target": "1016bf08944977a33d3a48edc15ee34b425f6d8a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1016bf08944977a33d3a48edc15ee34b425f6d8a/"
    },
    "refs/tags/v2.6.32-rc3": {
      "target": "910eff4ec30f648f297700d43784b2159d35fb4f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/910eff4ec30f648f297700d43784b2159d35fb4f/"
    },
    "refs/tags/v2.6.32-rc4": {
      "target": "742a213497d587595f23674eafad1e520c5af6bd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/742a213497d587595f23674eafad1e520c5af6bd/"
    },
    "refs/tags/v2.6.32-rc5": {
      "target": "c6add0a844533aeaa7bf86dcd4f924dca085d287",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c6add0a844533aeaa7bf86dcd4f924dca085d287/"
    },
    "refs/tags/v2.6.32-rc6": {
      "target": "53a1963436f23ee8b6fa29a5ebcd925a9912594b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/53a1963436f23ee8b6fa29a5ebcd925a9912594b/"
    },
    "refs/tags/v2.6.32-rc7": {
      "target": "5946e9740318d61b3d1c3d7bfdc3b54fc3ac181c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5946e9740318d61b3d1c3d7bfdc3b54fc3ac181c/"
    },
    "refs/tags/v2.6.32-rc8": {
      "target": "de8a4af91ab6fa3cde2618f4021b5faabdcd95ea",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/de8a4af91ab6fa3cde2618f4021b5faabdcd95ea/"
    },
    "refs/tags/v2.6.33": {
      "target": "4ac8e07ee3f251ae32329a24e0b01a316b21ead9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4ac8e07ee3f251ae32329a24e0b01a316b21ead9/"
    },
    "refs/tags/v2.6.33-rc1": {
      "target": "d14e040a3592de665407269688d70296955c5f14",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d14e040a3592de665407269688d70296955c5f14/"
    },
    "refs/tags/v2.6.33-rc2": {
      "target": "331ce84170c8ebba5f0fadac64f66d6f00a438e4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/331ce84170c8ebba5f0fadac64f66d6f00a438e4/"
    },
    "refs/tags/v2.6.33-rc3": {
      "target": "31fdc15b99788540d0ee8b8b337242e38489f603",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/31fdc15b99788540d0ee8b8b337242e38489f603/"
    },
    "refs/tags/v2.6.33-rc4": {
      "target": "5ba7808eabc37cb2464096077d0df55f33148245",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5ba7808eabc37cb2464096077d0df55f33148245/"
    },
    "refs/tags/v2.6.33-rc5": {
      "target": "76efcb71c910774213480cdfe20b73e07c6a00aa",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/76efcb71c910774213480cdfe20b73e07c6a00aa/"
    },
    "refs/tags/v2.6.33-rc6": {
      "target": "b02c43040da3d3e4f56d34d443c4c2a0d41da367",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b02c43040da3d3e4f56d34d443c4c2a0d41da367/"
    },
    "refs/tags/v2.6.33-rc7": {
      "target": "600255d9e9ec0eecc49be78197c630504cf8c263",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/600255d9e9ec0eecc49be78197c630504cf8c263/"
    },
    "refs/tags/v2.6.33-rc8": {
      "target": "22ebef85968508e596dc60d970b716024cb9a34e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/22ebef85968508e596dc60d970b716024cb9a34e/"
    },
    "refs/tags/v2.6.34": {
      "target": "dba2e709efc365df385a762e763b51365403bc0f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/dba2e709efc365df385a762e763b51365403bc0f/"
    },
    "refs/tags/v2.6.34-rc1": {
      "target": "fe2098f5dff78881162f71ae028384435681a90a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fe2098f5dff78881162f71ae028384435681a90a/"
    },
    "refs/tags/v2.6.34-rc2": {
      "target": "2fc56a2a7aa32adeddf7efe074b38cbdbb41894a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2fc56a2a7aa32adeddf7efe074b38cbdbb41894a/"
    },
    "refs/tags/v2.6.34-rc3": {
      "target": "fc2199137e85558e1f8f2bf44e0d3fa2b5cc4371",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fc2199137e85558e1f8f2bf44e0d3fa2b5cc4371/"
    },
    "refs/tags/v2.6.34-rc4": {
      "target": "cb0cbeb16ff949783023da2270d5af36af416865",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cb0cbeb16ff949783023da2270d5af36af416865/"
    },
    "refs/tags/v2.6.34-rc5": {
      "target": "d1744e136396b363e5844ed5f928e40067b5784a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d1744e136396b363e5844ed5f928e40067b5784a/"
    },
    "refs/tags/v2.6.34-rc6": {
      "target": "1b5265c5905fdc68873d37c902adc4aec2cbd6a3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1b5265c5905fdc68873d37c902adc4aec2cbd6a3/"
    },
    "refs/tags/v2.6.34-rc7": {
      "target": "3884cbfac5cf610a7116ca90a970a11cf495bb83",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3884cbfac5cf610a7116ca90a970a11cf495bb83/"
    },
    "refs/tags/v2.6.35": {
      "target": "d786bf1b8352facdde8cc28725e1e6e067e1854d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d786bf1b8352facdde8cc28725e1e6e067e1854d/"
    },
    "refs/tags/v2.6.35-rc1": {
      "target": "77cb4411c86825cf693e338fada52dfca3345668",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/77cb4411c86825cf693e338fada52dfca3345668/"
    },
    "refs/tags/v2.6.35-rc2": {
      "target": "06006d039fbf390b87b5ee76d16187b5cb9b3f4d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/06006d039fbf390b87b5ee76d16187b5cb9b3f4d/"
    },
    "refs/tags/v2.6.35-rc3": {
      "target": "f51882c32dcbec0befa48e12682f1651c4772b7b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f51882c32dcbec0befa48e12682f1651c4772b7b/"
    },
    "refs/tags/v2.6.35-rc4": {
      "target": "55acf6b6533581f03df07ddd166c6631bc304845",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/55acf6b6533581f03df07ddd166c6631bc304845/"
    },
    "refs/tags/v2.6.35-rc5": {
      "target": "1dd55af06d281ffbc945cd7bb21cfb98d6e84a99",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1dd55af06d281ffbc945cd7bb21cfb98d6e84a99/"
    },
    "refs/tags/v2.6.35-rc6": {
      "target": "37a3bd34ba0892f0f1c1a0a03b06d61ed3c554c1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/37a3bd34ba0892f0f1c1a0a03b06d61ed3c554c1/"
    },
    "refs/tags/v2.6.36": {
      "target": "25427f38d3b791d986812cb81c68df38e8249ef8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/25427f38d3b791d986812cb81c68df38e8249ef8/"
    },
    "refs/tags/v2.6.36-rc1": {
      "target": "8ed88d401f908a594cd74a4f2513b0fabd32b699",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8ed88d401f908a594cd74a4f2513b0fabd32b699/"
    },
    "refs/tags/v2.6.36-rc2": {
      "target": "58d3707b8891f71d4891e6b36129eeacd3ba63f4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/58d3707b8891f71d4891e6b36129eeacd3ba63f4/"
    },
    "refs/tags/v2.6.36-rc3": {
      "target": "40f7ec041a61c6b6d419e418818c79f7c23a1007",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/40f7ec041a61c6b6d419e418818c79f7c23a1007/"
    },
    "refs/tags/v2.6.36-rc4": {
      "target": "8607f6908a65fbd41d8eee6d0572425182eced69",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8607f6908a65fbd41d8eee6d0572425182eced69/"
    },
    "refs/tags/v2.6.36-rc5": {
      "target": "f4d2c86897046fb2dd9680b3446dfcc17a11e7f4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f4d2c86897046fb2dd9680b3446dfcc17a11e7f4/"
    },
    "refs/tags/v2.6.36-rc6": {
      "target": "93590f17a2e3ba2aed400c7608263b97da62b6d4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/93590f17a2e3ba2aed400c7608263b97da62b6d4/"
    },
    "refs/tags/v2.6.36-rc7": {
      "target": "f3f2d2543afa76bcc13a58fc6a1ff723f28890da",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f3f2d2543afa76bcc13a58fc6a1ff723f28890da/"
    },
    "refs/tags/v2.6.36-rc8": {
      "target": "7619e63f48822b2c68d0e108677340573873fb93",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7619e63f48822b2c68d0e108677340573873fb93/"
    },
    "refs/tags/v2.6.37": {
      "target": "4a7895f41220ed60e97b736ac1b92e589e67b263",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4a7895f41220ed60e97b736ac1b92e589e67b263/"
    },
    "refs/tags/v2.6.37-rc1": {
      "target": "579cc3d8b5840f3355bad58e7ab23eae04ff9cb6",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/579cc3d8b5840f3355bad58e7ab23eae04ff9cb6/"
    },
    "refs/tags/v2.6.37-rc2": {
      "target": "a1e4aaecda4a52e850e90d4f95ed2f59955ed057",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a1e4aaecda4a52e850e90d4f95ed2f59955ed057/"
    },
    "refs/tags/v2.6.37-rc3": {
      "target": "8a198c8196ac3716e1856978d3e18cca5d800ef3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8a198c8196ac3716e1856978d3e18cca5d800ef3/"
    },
    "refs/tags/v2.6.37-rc4": {
      "target": "78d3494aa9e5cbd28600b5440655b263ab1c0efd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/78d3494aa9e5cbd28600b5440655b263ab1c0efd/"
    },
    "refs/tags/v2.6.37-rc5": {
      "target": "f9459ed00525f440d25d7c88fe8d52abd3746066",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f9459ed00525f440d25d7c88fe8d52abd3746066/"
    },
    "refs/tags/v2.6.37-rc6": {
      "target": "3f4ddf273ce92382ad3ce55fde3d773bd9e4bddd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3f4ddf273ce92382ad3ce55fde3d773bd9e4bddd/"
    },
    "refs/tags/v2.6.37-rc7": {
      "target": "6261d1ea5bdbc9baf84c192242c82e63cdb02788",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6261d1ea5bdbc9baf84c192242c82e63cdb02788/"
    },
    "refs/tags/v2.6.37-rc8": {
      "target": "5720e551140dd586156e9875b0d1ef528e9c5f59",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5720e551140dd586156e9875b0d1ef528e9c5f59/"
    },
    "refs/tags/v2.6.38": {
      "target": "fbeb94b65cf784ed8bf852131e28c9fb5c4c760f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fbeb94b65cf784ed8bf852131e28c9fb5c4c760f/"
    },
    "refs/tags/v2.6.38-rc1": {
      "target": "aa6741cead5660406a453c01b6e37cbe06d52433",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/aa6741cead5660406a453c01b6e37cbe06d52433/"
    },
    "refs/tags/v2.6.38-rc2": {
      "target": "182a43159a3ea400594ca05ce7ff5052950a1010",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/182a43159a3ea400594ca05ce7ff5052950a1010/"
    },
    "refs/tags/v2.6.38-rc3": {
      "target": "70f1e06182fab9290a1f7775ade996e4854dec3a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/70f1e06182fab9290a1f7775ade996e4854dec3a/"
    },
    "refs/tags/v2.6.38-rc4": {
      "target": "901069c5c5b155322539a94cf337c378848e435a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/901069c5c5b155322539a94cf337c378848e435a/"
    },
    "refs/tags/v2.6.38-rc5": {
      "target": "656372ebed8acf941bb63d08abb250c9896785a8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/656372ebed8acf941bb63d08abb250c9896785a8/"
    },
    "refs/tags/v2.6.38-rc6": {
      "target": "86483dddb24b8d5624c38362f820211c694473ba",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/86483dddb24b8d5624c38362f820211c694473ba/"
    },
    "refs/tags/v2.6.38-rc7": {
      "target": "295dd79b22916ed71a641dd80ee4d8b07c624feb",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/295dd79b22916ed71a641dd80ee4d8b07c624feb/"
    },
    "refs/tags/v2.6.38-rc8": {
      "target": "44986a8d6d6d024d9422dbb91635600862921ed3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/44986a8d6d6d024d9422dbb91635600862921ed3/"
    },
    "refs/tags/v2.6.39": {
      "target": "8b0753a3df28c21b0570fa21362c5f1b3b4f59bf",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8b0753a3df28c21b0570fa21362c5f1b3b4f59bf/"
    },
    "refs/tags/v2.6.39-rc1": {
      "target": "5cb75c01f17f69d6ac102d58766f7fb30269a5c9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5cb75c01f17f69d6ac102d58766f7fb30269a5c9/"
    },
    "refs/tags/v2.6.39-rc2": {
      "target": "76af3715abf44176ae1ac9cf5cb9f3861d39eee1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/76af3715abf44176ae1ac9cf5cb9f3861d39eee1/"
    },
    "refs/tags/v2.6.39-rc3": {
      "target": "074bd9a1615dd4c0108b0424a50770d1b605b53e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/074bd9a1615dd4c0108b0424a50770d1b605b53e/"
    },
    "refs/tags/v2.6.39-rc4": {
      "target": "c8e761132679d935b5d3edd06e48db2bd3eb918a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c8e761132679d935b5d3edd06e48db2bd3eb918a/"
    },
    "refs/tags/v2.6.39-rc5": {
      "target": "149b78658628901a6e578566a45f159d0b38ce2f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/149b78658628901a6e578566a45f159d0b38ce2f/"
    },
    "refs/tags/v2.6.39-rc6": {
      "target": "8630f22a089f0d777893ee6a53eb3e6acef06044",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8630f22a089f0d777893ee6a53eb3e6acef06044/"
    },
    "refs/tags/v2.6.39-rc7": {
      "target": "2c2138749b9dc99a1f26b212eb373458ea179e06",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2c2138749b9dc99a1f26b212eb373458ea179e06/"
    },
    "refs/tags/v3.0": {
      "target": "4204bcde7c0b93c5e127eb868e17b337a513cf34",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4204bcde7c0b93c5e127eb868e17b337a513cf34/"
    },
    "refs/tags/v3.0-rc1": {
      "target": "2a23a510142a1ab597f0214e4fadb3c7350bbb8d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2a23a510142a1ab597f0214e4fadb3c7350bbb8d/"
    },
    "refs/tags/v3.0-rc2": {
      "target": "eb73a032155ba51adb229be1a2ad3ac08951b485",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/eb73a032155ba51adb229be1a2ad3ac08951b485/"
    },
    "refs/tags/v3.0-rc3": {
      "target": "a73e7344f7a7712628f40f851fa835ea82025d7b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a73e7344f7a7712628f40f851fa835ea82025d7b/"
    },
    "refs/tags/v3.0-rc4": {
      "target": "4921f03412bd0a91cb242b259618b13dfae1acf6",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4921f03412bd0a91cb242b259618b13dfae1acf6/"
    },
    "refs/tags/v3.0-rc5": {
      "target": "c699212fde57ea66926c4a0b6ebbe4d62c52e0b9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c699212fde57ea66926c4a0b6ebbe4d62c52e0b9/"
    },
    "refs/tags/v3.0-rc6": {
      "target": "8a79d68b5a847beef3f70b2ed16b2a757572cc5e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8a79d68b5a847beef3f70b2ed16b2a757572cc5e/"
    }
  },
  "next_branch": "refs/tags/v3.0-rc7"
}
    """,  # NoQA: B950  # NoQA: B950
    "snapshot/cabcc7d7bf639bbe1cc3b41989e1806618dd5764/?branches_count=1000&branches_from=refs/tags/v3.0-rc7": r"""
{
  "id": "cabcc7d7bf639bbe1cc3b41989e1806618dd5764",
  "branches": {
    "refs/tags/v3.0-rc7": {
      "target": "394d6f903ac6889fa50ca5c19111ae659d524b4c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/394d6f903ac6889fa50ca5c19111ae659d524b4c/"
    },
    "refs/tags/v3.1": {
      "target": "a732360658fc100add6ded23ac6c8013b50cef1f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a732360658fc100add6ded23ac6c8013b50cef1f/"
    },
    "refs/tags/v3.1-rc1": {
      "target": "dd073a5b13967fa83038ed1594ecae5faaed71a2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/dd073a5b13967fa83038ed1594ecae5faaed71a2/"
    },
    "refs/tags/v3.1-rc10": {
      "target": "bc9dac81d1d3442713e5b91ed7cda1646df9730e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/bc9dac81d1d3442713e5b91ed7cda1646df9730e/"
    },
    "refs/tags/v3.1-rc2": {
      "target": "fbba4cec036b7027b84c00547361bd8127254b1e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fbba4cec036b7027b84c00547361bd8127254b1e/"
    },
    "refs/tags/v3.1-rc3": {
      "target": "ae30e7e9c4d8cd135ec429191f9eec746709eccc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ae30e7e9c4d8cd135ec429191f9eec746709eccc/"
    },
    "refs/tags/v3.1-rc4": {
      "target": "897e5ed317d229e937731d6389f8f51c7e29e62e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/897e5ed317d229e937731d6389f8f51c7e29e62e/"
    },
    "refs/tags/v3.1-rc5": {
      "target": "dba5cf0231a4c3091c0f3b7236c6978dab8cbc97",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/dba5cf0231a4c3091c0f3b7236c6978dab8cbc97/"
    },
    "refs/tags/v3.1-rc6": {
      "target": "8ff02915cf6ca8f5ecda04e0a2150507df89846b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8ff02915cf6ca8f5ecda04e0a2150507df89846b/"
    },
    "refs/tags/v3.1-rc7": {
      "target": "4cf670b4bbdb31291ffa52ad3f65acd8cd2eb20d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4cf670b4bbdb31291ffa52ad3f65acd8cd2eb20d/"
    },
    "refs/tags/v3.1-rc8": {
      "target": "e7f2a88548df497544d4fafca79836bbd8b6342e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e7f2a88548df497544d4fafca79836bbd8b6342e/"
    },
    "refs/tags/v3.1-rc9": {
      "target": "38a181c9f494c81d7d2327861621c0e04018bc6a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/38a181c9f494c81d7d2327861621c0e04018bc6a/"
    },
    "refs/tags/v3.10": {
      "target": "bc1b510b8979ecc322f8d930dde56658967b7355",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/bc1b510b8979ecc322f8d930dde56658967b7355/"
    },
    "refs/tags/v3.10-rc1": {
      "target": "1fd7e50e9cbdabce5a0ff00cfb8767b948d86eb7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1fd7e50e9cbdabce5a0ff00cfb8767b948d86eb7/"
    },
    "refs/tags/v3.10-rc2": {
      "target": "a8c6d53c4d84b3a5eb182758a9cdceceba4691da",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a8c6d53c4d84b3a5eb182758a9cdceceba4691da/"
    },
    "refs/tags/v3.10-rc3": {
      "target": "74f675aa9f9c2f2d52de4a00144b64f5a36637ac",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/74f675aa9f9c2f2d52de4a00144b64f5a36637ac/"
    },
    "refs/tags/v3.10-rc4": {
      "target": "846830e71929e8728b0209e04259ca3e2b2f99c9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/846830e71929e8728b0209e04259ca3e2b2f99c9/"
    },
    "refs/tags/v3.10-rc5": {
      "target": "ee3e35b1aa057646c166780d13bfe8e57ed5d4cf",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ee3e35b1aa057646c166780d13bfe8e57ed5d4cf/"
    },
    "refs/tags/v3.10-rc6": {
      "target": "c598c40bcc18e93e63e3f251d40234478195d03a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c598c40bcc18e93e63e3f251d40234478195d03a/"
    },
    "refs/tags/v3.10-rc7": {
      "target": "e6995de3dfc1b122c2ce08522909ced160c960d9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e6995de3dfc1b122c2ce08522909ced160c960d9/"
    },
    "refs/tags/v3.11": {
      "target": "bedaca4e311e2c2abe0a215ee2b25c133e435211",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/bedaca4e311e2c2abe0a215ee2b25c133e435211/"
    },
    "refs/tags/v3.11-rc1": {
      "target": "8d339e724c3dc1ff81ad341b201e5b273c594d75",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8d339e724c3dc1ff81ad341b201e5b273c594d75/"
    },
    "refs/tags/v3.11-rc2": {
      "target": "c97e1dd9202cbfdf214208c934a021ac73b5a7d4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c97e1dd9202cbfdf214208c934a021ac73b5a7d4/"
    },
    "refs/tags/v3.11-rc3": {
      "target": "92f65bf9283c618ce157027b2fbcee82156b4c5c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/92f65bf9283c618ce157027b2fbcee82156b4c5c/"
    },
    "refs/tags/v3.11-rc4": {
      "target": "458a49f2e1c379ab07a3fd59894ae7050360b1fc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/458a49f2e1c379ab07a3fd59894ae7050360b1fc/"
    },
    "refs/tags/v3.11-rc5": {
      "target": "e49ee8a9d47753e7cd0f6e1827a95c436bc7544c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e49ee8a9d47753e7cd0f6e1827a95c436bc7544c/"
    },
    "refs/tags/v3.11-rc6": {
      "target": "2ea699d98cd6f9e9b813c24542d581dedacdc659",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2ea699d98cd6f9e9b813c24542d581dedacdc659/"
    },
    "refs/tags/v3.11-rc7": {
      "target": "99c0aead9dbef77b6a09f11d34b6867f3b352efe",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/99c0aead9dbef77b6a09f11d34b6867f3b352efe/"
    },
    "refs/tags/v3.12": {
      "target": "71572a0f766f454071277744c98cab00dad5efc9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/71572a0f766f454071277744c98cab00dad5efc9/"
    },
    "refs/tags/v3.12-rc1": {
      "target": "c59c557446a6ddb3260df207048d7b0a876d981c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c59c557446a6ddb3260df207048d7b0a876d981c/"
    },
    "refs/tags/v3.12-rc2": {
      "target": "e271e342a6202340459998d713188d1ad6d32cb6",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e271e342a6202340459998d713188d1ad6d32cb6/"
    },
    "refs/tags/v3.12-rc3": {
      "target": "59ca6c45dd68f79fc0470aafcf98fc64efc0d24d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/59ca6c45dd68f79fc0470aafcf98fc64efc0d24d/"
    },
    "refs/tags/v3.12-rc4": {
      "target": "3b6a2139d8b69bfc3a95a7b5252967a548c88597",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3b6a2139d8b69bfc3a95a7b5252967a548c88597/"
    },
    "refs/tags/v3.12-rc5": {
      "target": "3ec4b24ffb064e6fa91b61471e874027b8e9fbd2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3ec4b24ffb064e6fa91b61471e874027b8e9fbd2/"
    },
    "refs/tags/v3.12-rc6": {
      "target": "19ef23a1f211ed17cad3bfbe0539ec6f38c8ff63",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/19ef23a1f211ed17cad3bfbe0539ec6f38c8ff63/"
    },
    "refs/tags/v3.12-rc7": {
      "target": "af03676423fe9b4beefafcb8cd4a80164d4b0bd7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/af03676423fe9b4beefafcb8cd4a80164d4b0bd7/"
    },
    "refs/tags/v3.13": {
      "target": "0e2c4c21a56b8f820bab218566b02ea728909f45",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0e2c4c21a56b8f820bab218566b02ea728909f45/"
    },
    "refs/tags/v3.13-rc1": {
      "target": "1c2570ad1851fffa1e34d8b74d2349b09a3e8c2a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1c2570ad1851fffa1e34d8b74d2349b09a3e8c2a/"
    },
    "refs/tags/v3.13-rc2": {
      "target": "74c87395184969ce9f20e902f756cc7256ad2c20",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/74c87395184969ce9f20e902f756cc7256ad2c20/"
    },
    "refs/tags/v3.13-rc3": {
      "target": "1ae260b036bb223c11cfcd46698ea46a90cd336c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1ae260b036bb223c11cfcd46698ea46a90cd336c/"
    },
    "refs/tags/v3.13-rc4": {
      "target": "00c58b88ff7f7f6503e2021def13498a6846a849",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/00c58b88ff7f7f6503e2021def13498a6846a849/"
    },
    "refs/tags/v3.13-rc5": {
      "target": "5be6b0c63d1d45c8ad134b19053aaf79e9161218",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5be6b0c63d1d45c8ad134b19053aaf79e9161218/"
    },
    "refs/tags/v3.13-rc6": {
      "target": "13ecbebf2553bb6c5f19457e78a6c8f2a1732151",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/13ecbebf2553bb6c5f19457e78a6c8f2a1732151/"
    },
    "refs/tags/v3.13-rc7": {
      "target": "66d8d623aca0b26934afe85f5b05eb930d0166d0",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/66d8d623aca0b26934afe85f5b05eb930d0166d0/"
    },
    "refs/tags/v3.13-rc8": {
      "target": "210fd8c5c324253b0d6c852d161017aeb65f70bd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/210fd8c5c324253b0d6c852d161017aeb65f70bd/"
    },
    "refs/tags/v3.14": {
      "target": "f2378b14895ac79c325abe3c933744a26465e570",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f2378b14895ac79c325abe3c933744a26465e570/"
    },
    "refs/tags/v3.14-rc1": {
      "target": "08499ec1a4c8d859c71564d56890442f2f1ac8ed",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/08499ec1a4c8d859c71564d56890442f2f1ac8ed/"
    },
    "refs/tags/v3.14-rc2": {
      "target": "97112cce341931fb992ce6afc9d3a9beff9ad81c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/97112cce341931fb992ce6afc9d3a9beff9ad81c/"
    },
    "refs/tags/v3.14-rc3": {
      "target": "0f0e0c41b94db3caa68d5111347e699f6cf84e70",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0f0e0c41b94db3caa68d5111347e699f6cf84e70/"
    },
    "refs/tags/v3.14-rc4": {
      "target": "94de4401e66c81c0a55db01011f770a73eaa79c6",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/94de4401e66c81c0a55db01011f770a73eaa79c6/"
    },
    "refs/tags/v3.14-rc5": {
      "target": "5b2ddfdea1e0dc3e0ce8c427b187a86ba12c1e01",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5b2ddfdea1e0dc3e0ce8c427b187a86ba12c1e01/"
    },
    "refs/tags/v3.14-rc6": {
      "target": "4a5d07cddee1adcf08db8036f32062912ae8654b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4a5d07cddee1adcf08db8036f32062912ae8654b/"
    },
    "refs/tags/v3.14-rc7": {
      "target": "0722b25c4f25a8e9af18044f6e907c85e24272cf",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0722b25c4f25a8e9af18044f6e907c85e24272cf/"
    },
    "refs/tags/v3.14-rc8": {
      "target": "38b8c3135dd6ee173ba865c772162375bb4b6ab0",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/38b8c3135dd6ee173ba865c772162375bb4b6ab0/"
    },
    "refs/tags/v3.15": {
      "target": "8a4cee2f8086eee3cb0971e6ff46579e87d843eb",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8a4cee2f8086eee3cb0971e6ff46579e87d843eb/"
    },
    "refs/tags/v3.15-rc1": {
      "target": "8e3557db5cb41a5f92cd3535c48a553376614b8d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8e3557db5cb41a5f92cd3535c48a553376614b8d/"
    },
    "refs/tags/v3.15-rc2": {
      "target": "eaede3609686fa034e9eb1c7928de40437817147",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/eaede3609686fa034e9eb1c7928de40437817147/"
    },
    "refs/tags/v3.15-rc3": {
      "target": "42cb88a8267096a0a1a05209dabf41a320555e4f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/42cb88a8267096a0a1a05209dabf41a320555e4f/"
    },
    "refs/tags/v3.15-rc4": {
      "target": "d75c1fc4bcb7a77ad05a6ade92604781b52cc8c4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d75c1fc4bcb7a77ad05a6ade92604781b52cc8c4/"
    },
    "refs/tags/v3.15-rc5": {
      "target": "f79f332993983a0eacd3712d2558b695fafac835",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f79f332993983a0eacd3712d2558b695fafac835/"
    },
    "refs/tags/v3.15-rc6": {
      "target": "e13ef130784db89eb073c51d2c64173031f72521",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e13ef130784db89eb073c51d2c64173031f72521/"
    },
    "refs/tags/v3.15-rc7": {
      "target": "6b80f7f7e624b97f196fb7e69a70f5dbc60f8fb8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6b80f7f7e624b97f196fb7e69a70f5dbc60f8fb8/"
    },
    "refs/tags/v3.15-rc8": {
      "target": "822729b4fa5db475925a364b454c0e701a2f74f1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/822729b4fa5db475925a364b454c0e701a2f74f1/"
    },
    "refs/tags/v3.16": {
      "target": "f14680f1692a9ec2a5b1b716a7a0c03dd391106f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f14680f1692a9ec2a5b1b716a7a0c03dd391106f/"
    },
    "refs/tags/v3.16-rc1": {
      "target": "b79831e695f2e100b056794ed586eca25ee13ddb",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b79831e695f2e100b056794ed586eca25ee13ddb/"
    },
    "refs/tags/v3.16-rc2": {
      "target": "d2d9c4d67d98c18cbe96b84ba8cb3edc2a1eb59e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d2d9c4d67d98c18cbe96b84ba8cb3edc2a1eb59e/"
    },
    "refs/tags/v3.16-rc3": {
      "target": "dee9ef9e4858ca40b659cbec3204f3dd2ab4acca",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/dee9ef9e4858ca40b659cbec3204f3dd2ab4acca/"
    },
    "refs/tags/v3.16-rc4": {
      "target": "5f55a97431e1c38ecfcaf7a7657443b2ae0c1fe5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5f55a97431e1c38ecfcaf7a7657443b2ae0c1fe5/"
    },
    "refs/tags/v3.16-rc5": {
      "target": "6fd7a6ce3b2b440f56db45f028feac9142ce5063",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6fd7a6ce3b2b440f56db45f028feac9142ce5063/"
    },
    "refs/tags/v3.16-rc6": {
      "target": "3274219cbad6f00ba16bde698ca6bea1270d044a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3274219cbad6f00ba16bde698ca6bea1270d044a/"
    },
    "refs/tags/v3.16-rc7": {
      "target": "895deb688afba7cec559697f632e97337c152f06",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/895deb688afba7cec559697f632e97337c152f06/"
    },
    "refs/tags/v3.17": {
      "target": "b6b09084ce3842fce3d4ba4d663d6019116bbb09",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b6b09084ce3842fce3d4ba4d663d6019116bbb09/"
    },
    "refs/tags/v3.17-rc1": {
      "target": "cda2a78ee899e0a509d2a6a71bd229748d754c88",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cda2a78ee899e0a509d2a6a71bd229748d754c88/"
    },
    "refs/tags/v3.17-rc2": {
      "target": "4dfe46edb15c8b9b9accb3653a139e259be1b36b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4dfe46edb15c8b9b9accb3653a139e259be1b36b/"
    },
    "refs/tags/v3.17-rc3": {
      "target": "3c41a10e18914987b1b722c02f72f0f956be3904",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3c41a10e18914987b1b722c02f72f0f956be3904/"
    },
    "refs/tags/v3.17-rc4": {
      "target": "13360dd4297b38addd25d433d174d5b0a93a202f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/13360dd4297b38addd25d433d174d5b0a93a202f/"
    },
    "refs/tags/v3.17-rc5": {
      "target": "0c399959441c6f829f62730a0c71e68aa9491bfc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0c399959441c6f829f62730a0c71e68aa9491bfc/"
    },
    "refs/tags/v3.17-rc6": {
      "target": "d9613c658b9f84a86999ab4bfbd7af005a7d9227",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d9613c658b9f84a86999ab4bfbd7af005a7d9227/"
    },
    "refs/tags/v3.17-rc7": {
      "target": "823d115bbd2ec5655a43e9eb3e63eb0ba65956ad",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/823d115bbd2ec5655a43e9eb3e63eb0ba65956ad/"
    },
    "refs/tags/v3.18": {
      "target": "5a4a5d52f29d08d923ce8d232b0b497da674dd2c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5a4a5d52f29d08d923ce8d232b0b497da674dd2c/"
    },
    "refs/tags/v3.18-rc1": {
      "target": "78e7b65fc586396ec62cd1e694a46f7065bee097",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/78e7b65fc586396ec62cd1e694a46f7065bee097/"
    },
    "refs/tags/v3.18-rc2": {
      "target": "9ce4874112bdd096ff397711cf5f8951a4bac673",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9ce4874112bdd096ff397711cf5f8951a4bac673/"
    },
    "refs/tags/v3.18-rc3": {
      "target": "9661ddfd3ad0bb4cee8ec6659331271eec7bd738",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9661ddfd3ad0bb4cee8ec6659331271eec7bd738/"
    },
    "refs/tags/v3.18-rc4": {
      "target": "5fe21907a952c3fa0afca6cd8237dd3386cff37f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5fe21907a952c3fa0afca6cd8237dd3386cff37f/"
    },
    "refs/tags/v3.18-rc5": {
      "target": "d9d0c8bfbf103d15a640cf04349bf762644614f2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d9d0c8bfbf103d15a640cf04349bf762644614f2/"
    },
    "refs/tags/v3.18-rc6": {
      "target": "d1bc5bd19d31306a08b3f29e20caba92feae30f1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d1bc5bd19d31306a08b3f29e20caba92feae30f1/"
    },
    "refs/tags/v3.18-rc7": {
      "target": "35e8bc566ae51bfdeafbb9328377cb20b0afd2a7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/35e8bc566ae51bfdeafbb9328377cb20b0afd2a7/"
    },
    "refs/tags/v3.19": {
      "target": "e24f071559c7a928a3033e9fe9f68e52f4f6ec01",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e24f071559c7a928a3033e9fe9f68e52f4f6ec01/"
    },
    "refs/tags/v3.19-rc1": {
      "target": "30c0735c980547061c94ed4ff4e3e6a0b9611ac4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/30c0735c980547061c94ed4ff4e3e6a0b9611ac4/"
    },
    "refs/tags/v3.19-rc2": {
      "target": "9d3d0a958af86c1f52c59ed8932b8915b07c04c2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9d3d0a958af86c1f52c59ed8932b8915b07c04c2/"
    },
    "refs/tags/v3.19-rc3": {
      "target": "1746eaa6b7f74b6a25d43fd57f51b6bf0fafaeae",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1746eaa6b7f74b6a25d43fd57f51b6bf0fafaeae/"
    },
    "refs/tags/v3.19-rc4": {
      "target": "b615659d48a613fd2cc6fd8b5f35194ff00d0b61",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b615659d48a613fd2cc6fd8b5f35194ff00d0b61/"
    },
    "refs/tags/v3.19-rc5": {
      "target": "49e185e202071d50250678cd8c3ca8ad667d47a4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/49e185e202071d50250678cd8c3ca8ad667d47a4/"
    },
    "refs/tags/v3.19-rc6": {
      "target": "d84422ebc1dc4397355e56ea36a8605089afd12a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d84422ebc1dc4397355e56ea36a8605089afd12a/"
    },
    "refs/tags/v3.19-rc7": {
      "target": "9d9ad3a9bd1afe27724b287052dccc0cf87eb6d3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9d9ad3a9bd1afe27724b287052dccc0cf87eb6d3/"
    },
    "refs/tags/v3.2": {
      "target": "d018bf5252f1041f8839022e3ab38a1bc40d3868",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d018bf5252f1041f8839022e3ab38a1bc40d3868/"
    },
    "refs/tags/v3.2-rc1": {
      "target": "4ea8b996e5dd18145d945667fbcd9a6af234a30c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4ea8b996e5dd18145d945667fbcd9a6af234a30c/"
    },
    "refs/tags/v3.2-rc2": {
      "target": "68b7d2ca89b1ce708f40ebd45d19fcf982cc6b38",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/68b7d2ca89b1ce708f40ebd45d19fcf982cc6b38/"
    },
    "refs/tags/v3.2-rc3": {
      "target": "e0f79b438798057b03d1a06704d63c8e4297b053",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e0f79b438798057b03d1a06704d63c8e4297b053/"
    },
    "refs/tags/v3.2-rc4": {
      "target": "96c50634c8fc57b7a4a5d84691a18c173268251e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/96c50634c8fc57b7a4a5d84691a18c173268251e/"
    },
    "refs/tags/v3.2-rc5": {
      "target": "10fc295c8cd3c5b9d7d271b4875679ee294afbae",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/10fc295c8cd3c5b9d7d271b4875679ee294afbae/"
    },
    "refs/tags/v3.2-rc6": {
      "target": "5292d98194346528607da828962c7e8677c27e15",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5292d98194346528607da828962c7e8677c27e15/"
    },
    "refs/tags/v3.2-rc7": {
      "target": "e33abea7c57a228c8fdf77c892c230cb840425ac",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e33abea7c57a228c8fdf77c892c230cb840425ac/"
    },
    "refs/tags/v3.3": {
      "target": "353f301f06485635f3a7789c2787da50a34c49b5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/353f301f06485635f3a7789c2787da50a34c49b5/"
    },
    "refs/tags/v3.3-rc1": {
      "target": "0803b590443c714345e41c77b6e2f1d6c92243ec",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0803b590443c714345e41c77b6e2f1d6c92243ec/"
    },
    "refs/tags/v3.3-rc2": {
      "target": "8bd4e2a2ba139fcdab2ab9bbdd76e4df4d66d7fd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8bd4e2a2ba139fcdab2ab9bbdd76e4df4d66d7fd/"
    },
    "refs/tags/v3.3-rc3": {
      "target": "72cdd65517c6463e5578040976c166b33bcfc14a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/72cdd65517c6463e5578040976c166b33bcfc14a/"
    },
    "refs/tags/v3.3-rc4": {
      "target": "b87f9498998dbefd82fa7b3d6414454600481889",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b87f9498998dbefd82fa7b3d6414454600481889/"
    },
    "refs/tags/v3.3-rc5": {
      "target": "13aae9ead734a49a5486907f7dc290b2ea50a6b9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/13aae9ead734a49a5486907f7dc290b2ea50a6b9/"
    },
    "refs/tags/v3.3-rc6": {
      "target": "d9234c2ff6f2a50075dd37fb859e17bfb185d17c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d9234c2ff6f2a50075dd37fb859e17bfb185d17c/"
    },
    "refs/tags/v3.3-rc7": {
      "target": "b2e4909898710f070173a108f25e6e135b5bdebd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b2e4909898710f070173a108f25e6e135b5bdebd/"
    },
    "refs/tags/v3.4": {
      "target": "27406e978f596a646b87941cba5247eb0ccc8916",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/27406e978f596a646b87941cba5247eb0ccc8916/"
    },
    "refs/tags/v3.4-rc1": {
      "target": "b84516a747a9754fc77668a23cbf6203cf2c05d2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b84516a747a9754fc77668a23cbf6203cf2c05d2/"
    },
    "refs/tags/v3.4-rc2": {
      "target": "4029716364bee5d89323c24dc343c0dade05c5e8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4029716364bee5d89323c24dc343c0dade05c5e8/"
    },
    "refs/tags/v3.4-rc3": {
      "target": "0690bdc7e124d9ef5da67cc75f0ea9952a5427d7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0690bdc7e124d9ef5da67cc75f0ea9952a5427d7/"
    },
    "refs/tags/v3.4-rc4": {
      "target": "2263a9c8844371b175e061728bd7acb8fd188c4f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2263a9c8844371b175e061728bd7acb8fd188c4f/"
    },
    "refs/tags/v3.4-rc5": {
      "target": "c3bf00ad560c0ccf0e6a36d922cc99e9de10fb47",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c3bf00ad560c0ccf0e6a36d922cc99e9de10fb47/"
    },
    "refs/tags/v3.4-rc6": {
      "target": "2b1e9fec303b3f2f138edcefc5f048d666e56166",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2b1e9fec303b3f2f138edcefc5f048d666e56166/"
    },
    "refs/tags/v3.4-rc7": {
      "target": "5842651b3b71b4873d020e3e16e238b76fdfa90c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5842651b3b71b4873d020e3e16e238b76fdfa90c/"
    },
    "refs/tags/v3.5": {
      "target": "f34fad35283333561c88f4507c4fa1849f3451a1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f34fad35283333561c88f4507c4fa1849f3451a1/"
    },
    "refs/tags/v3.5-rc1": {
      "target": "04f4dfa6bbb4df376ed50599de43bf5181241602",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/04f4dfa6bbb4df376ed50599de43bf5181241602/"
    },
    "refs/tags/v3.5-rc2": {
      "target": "41827bc48b4ead01918d9a8ee2f7307a886c3a1a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/41827bc48b4ead01918d9a8ee2f7307a886c3a1a/"
    },
    "refs/tags/v3.5-rc3": {
      "target": "111ecbd538d2d3542e2e54a666f04c11bb14ed83",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/111ecbd538d2d3542e2e54a666f04c11bb14ed83/"
    },
    "refs/tags/v3.5-rc4": {
      "target": "379f9f503b99360d6fa269947eefc67492641c28",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/379f9f503b99360d6fa269947eefc67492641c28/"
    },
    "refs/tags/v3.5-rc5": {
      "target": "b26e78e3ef49a32fe2267fb194d740cabec35b78",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b26e78e3ef49a32fe2267fb194d740cabec35b78/"
    },
    "refs/tags/v3.5-rc6": {
      "target": "a94b851ad29264f83dd48229090cf95e6394b0d5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a94b851ad29264f83dd48229090cf95e6394b0d5/"
    },
    "refs/tags/v3.5-rc7": {
      "target": "ac143858efa89977a70ff375ac9c1bedba17691d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ac143858efa89977a70ff375ac9c1bedba17691d/"
    },
    "refs/tags/v3.6": {
      "target": "255595ebd4692ca21dcdbeef2b648204d0b8ee63",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/255595ebd4692ca21dcdbeef2b648204d0b8ee63/"
    },
    "refs/tags/v3.6-rc1": {
      "target": "f85125cc169d2379a55f170270ef3f0adf260409",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f85125cc169d2379a55f170270ef3f0adf260409/"
    },
    "refs/tags/v3.6-rc2": {
      "target": "d1a4d2f3a903dbd3b274cac81d53c138748ef309",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d1a4d2f3a903dbd3b274cac81d53c138748ef309/"
    },
    "refs/tags/v3.6-rc3": {
      "target": "c8d47765843dba9357904ee626a46f949e1ee2ca",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c8d47765843dba9357904ee626a46f949e1ee2ca/"
    },
    "refs/tags/v3.6-rc4": {
      "target": "f8042a627ac90fb99c89a89cc7a38650528ab76f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f8042a627ac90fb99c89a89cc7a38650528ab76f/"
    },
    "refs/tags/v3.6-rc5": {
      "target": "4678fc83dcdd811ec2b893b087d2a08c8530d7b2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4678fc83dcdd811ec2b893b087d2a08c8530d7b2/"
    },
    "refs/tags/v3.6-rc6": {
      "target": "895fdd1a25e399348c1fdc9be85dd847c302c275",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/895fdd1a25e399348c1fdc9be85dd847c302c275/"
    },
    "refs/tags/v3.6-rc7": {
      "target": "cbdc1ccf2692ee9d0f50fe4911fa474b7ac2882c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cbdc1ccf2692ee9d0f50fe4911fa474b7ac2882c/"
    },
    "refs/tags/v3.7": {
      "target": "d4f28c153f820851f9c765ac9bc12dffe90accae",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d4f28c153f820851f9c765ac9bc12dffe90accae/"
    },
    "refs/tags/v3.7-rc1": {
      "target": "b93229f2cc1b8bb260135c3da61733ed92568fd3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b93229f2cc1b8bb260135c3da61733ed92568fd3/"
    },
    "refs/tags/v3.7-rc2": {
      "target": "966b4b0842a6cbbfe8ce8165865b7c4fa993df32",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/966b4b0842a6cbbfe8ce8165865b7c4fa993df32/"
    },
    "refs/tags/v3.7-rc3": {
      "target": "47193149164f56a5b2cebe803cdf6bde56971323",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/47193149164f56a5b2cebe803cdf6bde56971323/"
    },
    "refs/tags/v3.7-rc4": {
      "target": "cf5c3c9cee716beddb7f2539a7f7bc0415239ce6",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cf5c3c9cee716beddb7f2539a7f7bc0415239ce6/"
    },
    "refs/tags/v3.7-rc5": {
      "target": "167e404cd31e98ab4fa56d8a4d774e4d9fdb5191",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/167e404cd31e98ab4fa56d8a4d774e4d9fdb5191/"
    },
    "refs/tags/v3.7-rc6": {
      "target": "60c890ec8f52b5d6369164dbb2212b2273786866",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/60c890ec8f52b5d6369164dbb2212b2273786866/"
    },
    "refs/tags/v3.7-rc7": {
      "target": "c4f5fad00840768d96c6999947ca50b8c91e70a6",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c4f5fad00840768d96c6999947ca50b8c91e70a6/"
    },
    "refs/tags/v3.7-rc8": {
      "target": "61cbb5be523d9b12da6596d5db90965d6ecd70d4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/61cbb5be523d9b12da6596d5db90965d6ecd70d4/"
    },
    "refs/tags/v3.8": {
      "target": "a788f85fbb206352319787d9dc1a5e62d87f74d4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a788f85fbb206352319787d9dc1a5e62d87f74d4/"
    },
    "refs/tags/v3.8-rc1": {
      "target": "6391acbd01acb5cde61dd54baf9caef2ae58d98c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6391acbd01acb5cde61dd54baf9caef2ae58d98c/"
    },
    "refs/tags/v3.8-rc2": {
      "target": "bfad7d20d7080bdba4c0e864fcbdf55baa2d1de2",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/bfad7d20d7080bdba4c0e864fcbdf55baa2d1de2/"
    },
    "refs/tags/v3.8-rc3": {
      "target": "d7150026d19ce56fb285b3d812cd6d5e0bfbe262",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d7150026d19ce56fb285b3d812cd6d5e0bfbe262/"
    },
    "refs/tags/v3.8-rc4": {
      "target": "16fc6c0d406528d25f8ef0ec2c2a543b46a5fba1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/16fc6c0d406528d25f8ef0ec2c2a543b46a5fba1/"
    },
    "refs/tags/v3.8-rc5": {
      "target": "b67f19a416919ae48c98105a7e32886b15f6e0d8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b67f19a416919ae48c98105a7e32886b15f6e0d8/"
    },
    "refs/tags/v3.8-rc6": {
      "target": "442d6ffcf0920b0631fb9b50d665a991c9e16e73",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/442d6ffcf0920b0631fb9b50d665a991c9e16e73/"
    },
    "refs/tags/v3.8-rc7": {
      "target": "afa17d15e4cf8ee70ad475627721bf4385a4a387",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/afa17d15e4cf8ee70ad475627721bf4385a4a387/"
    },
    "refs/tags/v3.9": {
      "target": "f73366b472547dd3fe6f4fb71d003773bb19d1d6",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f73366b472547dd3fe6f4fb71d003773bb19d1d6/"
    },
    "refs/tags/v3.9-rc1": {
      "target": "d29d73e064b457fbb6077e01e9516c96bc075279",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d29d73e064b457fbb6077e01e9516c96bc075279/"
    },
    "refs/tags/v3.9-rc2": {
      "target": "e4ae8588b9e3b14b60d746ec1c9381a565c2baa3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e4ae8588b9e3b14b60d746ec1c9381a565c2baa3/"
    },
    "refs/tags/v3.9-rc3": {
      "target": "c7664f420fc8d902676e1c8a34acfc5777bf8ce8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c7664f420fc8d902676e1c8a34acfc5777bf8ce8/"
    },
    "refs/tags/v3.9-rc4": {
      "target": "d3318b414811d5393b4535ab6cb5aeddb57a4e5b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d3318b414811d5393b4535ab6cb5aeddb57a4e5b/"
    },
    "refs/tags/v3.9-rc5": {
      "target": "c67bf5361e7e66a0ff1f4caf95f89347d55dfb89",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c67bf5361e7e66a0ff1f4caf95f89347d55dfb89/"
    },
    "refs/tags/v3.9-rc6": {
      "target": "2c8edfa75a6f7de3abc6061c5f665735de07dbee",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2c8edfa75a6f7de3abc6061c5f665735de07dbee/"
    },
    "refs/tags/v3.9-rc7": {
      "target": "c03c4b22b8b29c522d986a1e5440e35eb2d08aa7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c03c4b22b8b29c522d986a1e5440e35eb2d08aa7/"
    },
    "refs/tags/v3.9-rc8": {
      "target": "eec112d8fed9b29bd17c023975efd68b7f8b020b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/eec112d8fed9b29bd17c023975efd68b7f8b020b/"
    },
    "refs/tags/v4.0": {
      "target": "f68c461e712a798548ba3d976523209e7d775418",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f68c461e712a798548ba3d976523209e7d775418/"
    },
    "refs/tags/v4.0-rc1": {
      "target": "2e0c59488ed2cd9f563dbbc0c8dadf1ff6d985d8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2e0c59488ed2cd9f563dbbc0c8dadf1ff6d985d8/"
    },
    "refs/tags/v4.0-rc2": {
      "target": "548a887e1f1437a60c5a3f6cbe6a076a8741aacc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/548a887e1f1437a60c5a3f6cbe6a076a8741aacc/"
    },
    "refs/tags/v4.0-rc3": {
      "target": "4288c7642bbc65e806cba72ea7d6fe3d210e381e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4288c7642bbc65e806cba72ea7d6fe3d210e381e/"
    },
    "refs/tags/v4.0-rc4": {
      "target": "327b326bcb2d2426d3727ba473d89c30a55b0dbb",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/327b326bcb2d2426d3727ba473d89c30a55b0dbb/"
    },
    "refs/tags/v4.0-rc5": {
      "target": "44138dfa09632152f3f52720bc07158d74949414",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/44138dfa09632152f3f52720bc07158d74949414/"
    },
    "refs/tags/v4.0-rc6": {
      "target": "560a83984c7d32dd40027c9920d3997d519e94a7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/560a83984c7d32dd40027c9920d3997d519e94a7/"
    },
    "refs/tags/v4.0-rc7": {
      "target": "2191b3832b01c9d2cee0c1a85c46c187a8d02035",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2191b3832b01c9d2cee0c1a85c46c187a8d02035/"
    },
    "refs/tags/v4.1": {
      "target": "8c36ccaef5b771db296df5d8ef1ecf03037b80f8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8c36ccaef5b771db296df5d8ef1ecf03037b80f8/"
    },
    "refs/tags/v4.1-rc1": {
      "target": "0ff6850711723750817e8f28d60d6555f32f2a4d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0ff6850711723750817e8f28d60d6555f32f2a4d/"
    },
    "refs/tags/v4.1-rc2": {
      "target": "d657ab5e916bf043d76080c003ab8eb9efe5cb9e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d657ab5e916bf043d76080c003ab8eb9efe5cb9e/"
    },
    "refs/tags/v4.1-rc3": {
      "target": "15712d02596a983465c650a31c40e0ed3597e90c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/15712d02596a983465c650a31c40e0ed3597e90c/"
    },
    "refs/tags/v4.1-rc4": {
      "target": "8ba8113a05841e52fdd34ed0055b9c551068312c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8ba8113a05841e52fdd34ed0055b9c551068312c/"
    },
    "refs/tags/v4.1-rc5": {
      "target": "03c4f667368424eb7ee88eaf8cea11587ee7c64f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/03c4f667368424eb7ee88eaf8cea11587ee7c64f/"
    },
    "refs/tags/v4.1-rc6": {
      "target": "986aef29e5bff76d9844d62218b9b7c264531a7f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/986aef29e5bff76d9844d62218b9b7c264531a7f/"
    },
    "refs/tags/v4.1-rc7": {
      "target": "b6b66985b1569bffc4f2043335eb774ea2d42538",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b6b66985b1569bffc4f2043335eb774ea2d42538/"
    },
    "refs/tags/v4.1-rc8": {
      "target": "ead046ad8f469fc0c7cd719126a76c7db24bbbf4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ead046ad8f469fc0c7cd719126a76c7db24bbbf4/"
    },
    "refs/tags/v4.10": {
      "target": "850bc05248749f47b0c0a64af52cfe213bdec385",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/850bc05248749f47b0c0a64af52cfe213bdec385/"
    },
    "refs/tags/v4.10-rc1": {
      "target": "2b3b80e8b9daba3e8e12f23f1acde4bd0ec88427",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2b3b80e8b9daba3e8e12f23f1acde4bd0ec88427/"
    },
    "refs/tags/v4.10-rc2": {
      "target": "eaa30079ff1896aa415b0fd5b153fccc11ac570e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/eaa30079ff1896aa415b0fd5b153fccc11ac570e/"
    },
    "refs/tags/v4.10-rc3": {
      "target": "cb32faa36a6b317fffeb566758519a5d4aa09750",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cb32faa36a6b317fffeb566758519a5d4aa09750/"
    },
    "refs/tags/v4.10-rc4": {
      "target": "b0131e43d2880f9bda208b0f17570f20280995e9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b0131e43d2880f9bda208b0f17570f20280995e9/"
    },
    "refs/tags/v4.10-rc5": {
      "target": "01a49c26cbeb8c074c5e5ea212a67d92d62307dc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/01a49c26cbeb8c074c5e5ea212a67d92d62307dc/"
    },
    "refs/tags/v4.10-rc6": {
      "target": "a4ee79b7932655354695e1c449a95cccbe3d0679",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a4ee79b7932655354695e1c449a95cccbe3d0679/"
    },
    "refs/tags/v4.10-rc7": {
      "target": "d8c71924693d90834b2de38a061f5200b1dc6245",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d8c71924693d90834b2de38a061f5200b1dc6245/"
    },
    "refs/tags/v4.10-rc8": {
      "target": "f1bff1ad193c13715d0f6c8f5a3451a4798c5d62",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f1bff1ad193c13715d0f6c8f5a3451a4798c5d62/"
    },
    "refs/tags/v4.11": {
      "target": "f22c06ae9a78e9d0a3c6d1c8e546bdddfff332b5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f22c06ae9a78e9d0a3c6d1c8e546bdddfff332b5/"
    },
    "refs/tags/v4.11-rc1": {
      "target": "4314d82b9e7401b130537174f9c89984425113f8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4314d82b9e7401b130537174f9c89984425113f8/"
    },
    "refs/tags/v4.11-rc2": {
      "target": "c76a1ee152bc8a85ee33acebc48d0db1fafddb0c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c76a1ee152bc8a85ee33acebc48d0db1fafddb0c/"
    },
    "refs/tags/v4.11-rc3": {
      "target": "6b0b7bb1a7f968e2f4965d5f6d17b8970b29897e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6b0b7bb1a7f968e2f4965d5f6d17b8970b29897e/"
    },
    "refs/tags/v4.11-rc4": {
      "target": "56071a8c579258685faa4f8f7a67643dfee3042d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/56071a8c579258685faa4f8f7a67643dfee3042d/"
    },
    "refs/tags/v4.11-rc5": {
      "target": "a54e39dbdd34e6d6b0f971a16233efa320357d5c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a54e39dbdd34e6d6b0f971a16233efa320357d5c/"
    },
    "refs/tags/v4.11-rc6": {
      "target": "346dd7d0b1b42d184cd20ff83384b24a6c9826d8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/346dd7d0b1b42d184cd20ff83384b24a6c9826d8/"
    },
    "refs/tags/v4.11-rc7": {
      "target": "e797c36f320f82da365be7282a89e3e52b1e1ff4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e797c36f320f82da365be7282a89e3e52b1e1ff4/"
    },
    "refs/tags/v4.11-rc8": {
      "target": "903bda2c64f7ad1b51b05670998f4393a73f8492",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/903bda2c64f7ad1b51b05670998f4393a73f8492/"
    },
    "refs/tags/v4.12": {
      "target": "8ecf9c5f3a2a8b2d8032bb1bc2f6abc260fd46bf",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8ecf9c5f3a2a8b2d8032bb1bc2f6abc260fd46bf/"
    },
    "refs/tags/v4.12-rc1": {
      "target": "3cf38b2cdc0a6452d9a51c13840cefde4fce6cd7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3cf38b2cdc0a6452d9a51c13840cefde4fce6cd7/"
    },
    "refs/tags/v4.12-rc2": {
      "target": "93e7baa205eab1fca25218d69af4adffae994524",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/93e7baa205eab1fca25218d69af4adffae994524/"
    },
    "refs/tags/v4.12-rc3": {
      "target": "992074ed78d9b1a05d18c656fa16b51f4d25005e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/992074ed78d9b1a05d18c656fa16b51f4d25005e/"
    },
    "refs/tags/v4.12-rc4": {
      "target": "38bb76c574be09462e64875f160d9189b84a0660",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/38bb76c574be09462e64875f160d9189b84a0660/"
    },
    "refs/tags/v4.12-rc5": {
      "target": "d612c4ce21e76aafc04bfd38d9400fde145090dd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d612c4ce21e76aafc04bfd38d9400fde145090dd/"
    },
    "refs/tags/v4.12-rc6": {
      "target": "1e455240998832dacdf1d27f67dbfc25c3f453c3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1e455240998832dacdf1d27f67dbfc25c3f453c3/"
    },
    "refs/tags/v4.12-rc7": {
      "target": "c5fc112f12a4e33d750f2ceea94a7d8fc415f4f1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c5fc112f12a4e33d750f2ceea94a7d8fc415f4f1/"
    },
    "refs/tags/v4.13": {
      "target": "269e8c00ebbf99adfc6a60af813dc43a378e7c06",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/269e8c00ebbf99adfc6a60af813dc43a378e7c06/"
    },
    "refs/tags/v4.13-rc1": {
      "target": "c5a15c6f450709c4dc4517250cea9aff202ed9bd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c5a15c6f450709c4dc4517250cea9aff202ed9bd/"
    },
    "refs/tags/v4.13-rc2": {
      "target": "9ae81923f5fc397e5f0cf2836926a7d1cab26aa5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9ae81923f5fc397e5f0cf2836926a7d1cab26aa5/"
    },
    "refs/tags/v4.13-rc3": {
      "target": "f7965d62aac0f3728499ca12d77525afdda2ede9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f7965d62aac0f3728499ca12d77525afdda2ede9/"
    },
    "refs/tags/v4.13-rc4": {
      "target": "c3c4ea1b1d61a2e4923ba77837767bc8bbebf12b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/c3c4ea1b1d61a2e4923ba77837767bc8bbebf12b/"
    },
    "refs/tags/v4.13-rc5": {
      "target": "2ef4d46d02937a82e6a2446d41f209a998f3b7fd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2ef4d46d02937a82e6a2446d41f209a998f3b7fd/"
    },
    "refs/tags/v4.13-rc6": {
      "target": "383584ac020b8c34f88f31aa66062628c80fd46d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/383584ac020b8c34f88f31aa66062628c80fd46d/"
    },
    "refs/tags/v4.13-rc7": {
      "target": "b2123ad40c157bcafa5ca501f77b41dccbdba098",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b2123ad40c157bcafa5ca501f77b41dccbdba098/"
    },
    "refs/tags/v4.14": {
      "target": "e3d97e8db5c45dfe3e4c70252acc5f4edea172b7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e3d97e8db5c45dfe3e4c70252acc5f4edea172b7/"
    },
    "refs/tags/v4.14-rc1": {
      "target": "9a7101adf0fbde19a069b6ebca1dde8954daef29",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9a7101adf0fbde19a069b6ebca1dde8954daef29/"
    },
    "refs/tags/v4.14-rc2": {
      "target": "eab260db417449f5c6500319c9d2f3fc33087051",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/eab260db417449f5c6500319c9d2f3fc33087051/"
    },
    "refs/tags/v4.14-rc3": {
      "target": "fe3de06298537055613b96424db9ede87266cd7c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fe3de06298537055613b96424db9ede87266cd7c/"
    },
    "refs/tags/v4.14-rc4": {
      "target": "f82861df06a7d05108bc837aecfed97534784929",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f82861df06a7d05108bc837aecfed97534784929/"
    },
    "refs/tags/v4.14-rc5": {
      "target": "6653fe5d1e2bf528bde9b86658eb31ab039ad60b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6653fe5d1e2bf528bde9b86658eb31ab039ad60b/"
    },
    "refs/tags/v4.14-rc6": {
      "target": "cbbcc7e4ec6e68930eb608960dae4159d5aaa0a3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cbbcc7e4ec6e68930eb608960dae4159d5aaa0a3/"
    },
    "refs/tags/v4.14-rc7": {
      "target": "0242b86da86562ca68d816dc7a7ebb6dafff0fe5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0242b86da86562ca68d816dc7a7ebb6dafff0fe5/"
    },
    "refs/tags/v4.14-rc8": {
      "target": "b60ed69f0d467074fba635278de2df54daf43d3b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b60ed69f0d467074fba635278de2df54daf43d3b/"
    },
    "refs/tags/v4.15": {
      "target": "f5b7ec87f701326704545ebfc8fe4a2aa42086aa",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f5b7ec87f701326704545ebfc8fe4a2aa42086aa/"
    },
    "refs/tags/v4.15-rc1": {
      "target": "2fd4bf77c3d8ccceaf6bdaa507ee4cf94e1a2ca5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2fd4bf77c3d8ccceaf6bdaa507ee4cf94e1a2ca5/"
    },
    "refs/tags/v4.15-rc2": {
      "target": "9f92a4865a74b206034b1358b226a1237ad7e6c9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9f92a4865a74b206034b1358b226a1237ad7e6c9/"
    },
    "refs/tags/v4.15-rc3": {
      "target": "fa0dd402a43dd6a02af9f3b8690bc1c9d4090a26",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fa0dd402a43dd6a02af9f3b8690bc1c9d4090a26/"
    },
    "refs/tags/v4.15-rc4": {
      "target": "e0e0c107ce60c16ec503033024744889ebd8f9d5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e0e0c107ce60c16ec503033024744889ebd8f9d5/"
    },
    "refs/tags/v4.15-rc5": {
      "target": "624b3b85d15a7e77a5e49f9496e42878cbb3bc8e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/624b3b85d15a7e77a5e49f9496e42878cbb3bc8e/"
    },
    "refs/tags/v4.15-rc6": {
      "target": "dabc25ee28d36fe572d2093224db0d02e5dc8da4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/dabc25ee28d36fe572d2093224db0d02e5dc8da4/"
    },
    "refs/tags/v4.15-rc7": {
      "target": "e43ec6ccc6aa9628bf2563eec17b9ce8ba9b30f7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e43ec6ccc6aa9628bf2563eec17b9ce8ba9b30f7/"
    },
    "refs/tags/v4.15-rc8": {
      "target": "9870bc69c9a9974f71b3063efc7d55397359d845",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9870bc69c9a9974f71b3063efc7d55397359d845/"
    },
    "refs/tags/v4.15-rc9": {
      "target": "7d05020f7ea88483dde607796a072b97611c37f5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7d05020f7ea88483dde607796a072b97611c37f5/"
    },
    "refs/tags/v4.16": {
      "target": "3fb78e8be9d9428fbc4c016f8b031b01a6fdd63d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3fb78e8be9d9428fbc4c016f8b031b01a6fdd63d/"
    },
    "refs/tags/v4.16-rc1": {
      "target": "5781e793a06a17edb3616114129f664f7b73161f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5781e793a06a17edb3616114129f664f7b73161f/"
    },
    "refs/tags/v4.16-rc2": {
      "target": "8a1eab637309319aee4caa10b196eb7327b549ae",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8a1eab637309319aee4caa10b196eb7327b549ae/"
    },
    "refs/tags/v4.16-rc3": {
      "target": "671e0d17675c17d89086e353935bf718e8096b29",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/671e0d17675c17d89086e353935bf718e8096b29/"
    },
    "refs/tags/v4.16-rc4": {
      "target": "25876d753c615f7b595a8da1746fa1a8832c7fa3",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/25876d753c615f7b595a8da1746fa1a8832c7fa3/"
    },
    "refs/tags/v4.16-rc5": {
      "target": "7169c8da1f5d4497ca62c507a735ed1e5dc67af8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7169c8da1f5d4497ca62c507a735ed1e5dc67af8/"
    },
    "refs/tags/v4.16-rc6": {
      "target": "49d8069446d565d0a796edc048b273620e74bdee",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/49d8069446d565d0a796edc048b273620e74bdee/"
    },
    "refs/tags/v4.16-rc7": {
      "target": "a1fbb3e9a14d2073b62404ff87bf3d1c077579ea",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a1fbb3e9a14d2073b62404ff87bf3d1c077579ea/"
    },
    "refs/tags/v4.17": {
      "target": "42308afa917ad74c85ad01bad5ffbe04e7cc6f76",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/42308afa917ad74c85ad01bad5ffbe04e7cc6f76/"
    },
    "refs/tags/v4.17-rc1": {
      "target": "a79c33a10ce2764c62fb8af6cbb571752d55c1c0",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a79c33a10ce2764c62fb8af6cbb571752d55c1c0/"
    },
    "refs/tags/v4.17-rc2": {
      "target": "2010cd502831cd3e64db447fd892821446d92d0a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2010cd502831cd3e64db447fd892821446d92d0a/"
    },
    "refs/tags/v4.17-rc3": {
      "target": "51b71dd7b97fb2f1605cba6277e72322fdf35d04",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/51b71dd7b97fb2f1605cba6277e72322fdf35d04/"
    },
    "refs/tags/v4.17-rc4": {
      "target": "757758d955574686dd47a12b9ff821ef496b21b8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/757758d955574686dd47a12b9ff821ef496b21b8/"
    },
    "refs/tags/v4.17-rc5": {
      "target": "2710afc17affbc818249bf9a378423da6a97adb5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2710afc17affbc818249bf9a378423da6a97adb5/"
    },
    "refs/tags/v4.17-rc6": {
      "target": "f5a91149feade0213a7ec9dac8615f41f2f15027",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f5a91149feade0213a7ec9dac8615f41f2f15027/"
    },
    "refs/tags/v4.17-rc7": {
      "target": "0504ece5232c8e66b5c2ef6c97c485e058ec7064",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0504ece5232c8e66b5c2ef6c97c485e058ec7064/"
    },
    "refs/tags/v4.18": {
      "target": "85262cb1fbc5cc267b3c86bc162a02dedd219548",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/85262cb1fbc5cc267b3c86bc162a02dedd219548/"
    },
    "refs/tags/v4.18-rc1": {
      "target": "7f40f7193eaee45287374eed75ecfd5ba893f709",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7f40f7193eaee45287374eed75ecfd5ba893f709/"
    },
    "refs/tags/v4.18-rc2": {
      "target": "fb87fcb6b1ae28324b266e48281eed938d6f9f42",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fb87fcb6b1ae28324b266e48281eed938d6f9f42/"
    },
    "refs/tags/v4.18-rc3": {
      "target": "b976618f9aab4783db47d1ab6d40201807e388e8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b976618f9aab4783db47d1ab6d40201807e388e8/"
    },
    "refs/tags/v4.18-rc4": {
      "target": "a4246a64e18f55e7dced22acf8a71f3b8401bd18",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a4246a64e18f55e7dced22acf8a71f3b8401bd18/"
    },
    "refs/tags/v4.18-rc5": {
      "target": "1383782787b75342dcbc1b0c67b510a0ac1c1c11",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1383782787b75342dcbc1b0c67b510a0ac1c1c11/"
    },
    "refs/tags/v4.18-rc6": {
      "target": "8734478edfd20d9ee223ea24afb38187d2a8c939",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8734478edfd20d9ee223ea24afb38187d2a8c939/"
    },
    "refs/tags/v4.18-rc7": {
      "target": "71257c9661c8053cc071fcd7eb2d14549161509c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/71257c9661c8053cc071fcd7eb2d14549161509c/"
    },
    "refs/tags/v4.18-rc8": {
      "target": "9331534f9b0bc629c3e77677439636736447cf84",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9331534f9b0bc629c3e77677439636736447cf84/"
    },
    "refs/tags/v4.19": {
      "target": "2241b8bcf2b5f1b01ebb1cbd1231bbbb72230064",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2241b8bcf2b5f1b01ebb1cbd1231bbbb72230064/"
    },
    "refs/tags/v4.19-rc1": {
      "target": "0eaf7a1a2dcfb1d287af95e4a5d4fb7d7be94cbf",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0eaf7a1a2dcfb1d287af95e4a5d4fb7d7be94cbf/"
    },
    "refs/tags/v4.19-rc2": {
      "target": "f703a4f77147a05a73d889992bd6b02a152d84d6",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f703a4f77147a05a73d889992bd6b02a152d84d6/"
    },
    "refs/tags/v4.19-rc3": {
      "target": "549008abfd88f89951ac70c8bf78a27a726d7478",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/549008abfd88f89951ac70c8bf78a27a726d7478/"
    },
    "refs/tags/v4.19-rc4": {
      "target": "5f552f3827718b54ef7a5737d4915258cccdff65",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5f552f3827718b54ef7a5737d4915258cccdff65/"
    },
    "refs/tags/v4.19-rc5": {
      "target": "cb786f94f5dacf1c20dc0b1f6b0c11edaa791554",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cb786f94f5dacf1c20dc0b1f6b0c11edaa791554/"
    },
    "refs/tags/v4.19-rc6": {
      "target": "ee3051afbc525f33f4a921c971290cf3bb264fba",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ee3051afbc525f33f4a921c971290cf3bb264fba/"
    },
    "refs/tags/v4.19-rc7": {
      "target": "8c7d64c58fc97a3db97d6f2f6c30c6a70b55cc45",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8c7d64c58fc97a3db97d6f2f6c30c6a70b55cc45/"
    },
    "refs/tags/v4.19-rc8": {
      "target": "9b6bcb55ef5b3293c6bb647dbbe4d14f8082dfda",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9b6bcb55ef5b3293c6bb647dbbe4d14f8082dfda/"
    },
    "refs/tags/v4.2": {
      "target": "ac9d1de8ac6a4eeed646f2bfc7791b0b3f532e38",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ac9d1de8ac6a4eeed646f2bfc7791b0b3f532e38/"
    },
    "refs/tags/v4.2-rc1": {
      "target": "1e951912027ea6873da6985b91e50c47f645ae1a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1e951912027ea6873da6985b91e50c47f645ae1a/"
    },
    "refs/tags/v4.2-rc2": {
      "target": "7e67591243ec4ae951e69811b6b155c2c2ca8447",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7e67591243ec4ae951e69811b6b155c2c2ca8447/"
    },
    "refs/tags/v4.2-rc3": {
      "target": "a0ff0d2281af2d44f49e2b8f05199778c152a6a4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a0ff0d2281af2d44f49e2b8f05199778c152a6a4/"
    },
    "refs/tags/v4.2-rc4": {
      "target": "e84837c3908251707c7c161141fbf6173ee50c6a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e84837c3908251707c7c161141fbf6173ee50c6a/"
    },
    "refs/tags/v4.2-rc5": {
      "target": "1881e83be5b9d3ea40d7b72a81098c68d7577fdb",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1881e83be5b9d3ea40d7b72a81098c68d7577fdb/"
    },
    "refs/tags/v4.2-rc6": {
      "target": "52c056f2619406a65358601d23251c0b7fba8e0e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/52c056f2619406a65358601d23251c0b7fba8e0e/"
    },
    "refs/tags/v4.2-rc7": {
      "target": "abb8a3b0b6fdb4f9ffe6fa90109a0f5ec0a4754d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/abb8a3b0b6fdb4f9ffe6fa90109a0f5ec0a4754d/"
    },
    "refs/tags/v4.2-rc8": {
      "target": "3ea11d4a2e3e1fad39c07dfc1ce8c952cec0a2c4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3ea11d4a2e3e1fad39c07dfc1ce8c952cec0a2c4/"
    },
    "refs/tags/v4.20": {
      "target": "012caf0bb40788cf3c7b876bababe810007b9996",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/012caf0bb40788cf3c7b876bababe810007b9996/"
    },
    "refs/tags/v4.20-rc1": {
      "target": "29d82febb3f92e64c1c4826be4c130804c50707e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/29d82febb3f92e64c1c4826be4c130804c50707e/"
    },
    "refs/tags/v4.20-rc2": {
      "target": "f615b81baa686db72c59f3a4e5ad953cbdf6b9d7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f615b81baa686db72c59f3a4e5ad953cbdf6b9d7/"
    },
    "refs/tags/v4.20-rc3": {
      "target": "b9f688725cc0236a778e942fe4da2edd706a33af",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b9f688725cc0236a778e942fe4da2edd706a33af/"
    },
    "refs/tags/v4.20-rc4": {
      "target": "e8bc1380f83ffb81f6c4faccf38beba81b905a57",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e8bc1380f83ffb81f6c4faccf38beba81b905a57/"
    },
    "refs/tags/v4.20-rc5": {
      "target": "7930acef24ef5700d5f4185ec3ec633058ccf542",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/7930acef24ef5700d5f4185ec3ec633058ccf542/"
    },
    "refs/tags/v4.20-rc6": {
      "target": "cba6e81694201f001b4aa80c0366db5b6091666d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cba6e81694201f001b4aa80c0366db5b6091666d/"
    },
    "refs/tags/v4.20-rc7": {
      "target": "264cdddbff3029302e355cd7a00d7403e76dfa66",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/264cdddbff3029302e355cd7a00d7403e76dfa66/"
    },
    "refs/tags/v4.3": {
      "target": "d74151c670f51a5d1c7a5aa72dc1f3c0d90fba48",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d74151c670f51a5d1c7a5aa72dc1f3c0d90fba48/"
    },
    "refs/tags/v4.3-rc1": {
      "target": "8a49eff98fd60628619a11f5da0fbf21b9bbd7ac",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8a49eff98fd60628619a11f5da0fbf21b9bbd7ac/"
    },
    "refs/tags/v4.3-rc2": {
      "target": "027ad56d85e0d79cea005497b445c11ef644d46f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/027ad56d85e0d79cea005497b445c11ef644d46f/"
    },
    "refs/tags/v4.3-rc3": {
      "target": "89c6defeed8f0b0d61f45d081038f7e1e5023ffb",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/89c6defeed8f0b0d61f45d081038f7e1e5023ffb/"
    },
    "refs/tags/v4.3-rc4": {
      "target": "e4ceda2f73b1f6777b283caef4156c4f42a73831",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e4ceda2f73b1f6777b283caef4156c4f42a73831/"
    },
    "refs/tags/v4.3-rc5": {
      "target": "1590f2edcb51267c9a421eb1b124c669d6e2d1f7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1590f2edcb51267c9a421eb1b124c669d6e2d1f7/"
    },
    "refs/tags/v4.3-rc6": {
      "target": "f00790324dc5ed1c74c6308c9e00864b6e162df8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f00790324dc5ed1c74c6308c9e00864b6e162df8/"
    },
    "refs/tags/v4.3-rc7": {
      "target": "a5e28a33ce5244958f8a0f9e408b7191ed4a3fc1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a5e28a33ce5244958f8a0f9e408b7191ed4a3fc1/"
    },
    "refs/tags/v4.4": {
      "target": "b5be40b90dbaa6bd337f3b77de361bfc0723468b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b5be40b90dbaa6bd337f3b77de361bfc0723468b/"
    },
    "refs/tags/v4.4-rc1": {
      "target": "02f277aef4de9e9ed5313ff106edc58afc9bef44",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/02f277aef4de9e9ed5313ff106edc58afc9bef44/"
    },
    "refs/tags/v4.4-rc2": {
      "target": "28e2369c64672dea3aa687008ad30e7fffbe3e6d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/28e2369c64672dea3aa687008ad30e7fffbe3e6d/"
    },
    "refs/tags/v4.4-rc3": {
      "target": "5da763dbd025698748c4f98be1f39a65f38a329e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5da763dbd025698748c4f98be1f39a65f38a329e/"
    },
    "refs/tags/v4.4-rc4": {
      "target": "4b74253c4cd4f47aa99e52a0741a9084279fbd1a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4b74253c4cd4f47aa99e52a0741a9084279fbd1a/"
    },
    "refs/tags/v4.4-rc5": {
      "target": "75ac92eafc98c67b69a06e5e9f7847a2ed7fa0c8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/75ac92eafc98c67b69a06e5e9f7847a2ed7fa0c8/"
    },
    "refs/tags/v4.4-rc6": {
      "target": "fac30e915cefbb5cb8aeeebfc614828a90b0293b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/fac30e915cefbb5cb8aeeebfc614828a90b0293b/"
    },
    "refs/tags/v4.4-rc7": {
      "target": "f5fd7e2626260acc3437bc354279fe55ba26498e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f5fd7e2626260acc3437bc354279fe55ba26498e/"
    },
    "refs/tags/v4.4-rc8": {
      "target": "4df2717f13a5f6b6e9c21cf3676f4863e4a0530e",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4df2717f13a5f6b6e9c21cf3676f4863e4a0530e/"
    },
    "refs/tags/v4.5": {
      "target": "0fc80a1fc05dd08762fda6e8efafbb54e88608f4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0fc80a1fc05dd08762fda6e8efafbb54e88608f4/"
    },
    "refs/tags/v4.5-rc1": {
      "target": "b13adfcdf288a00a7e58556f326ecb56fbb53b92",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b13adfcdf288a00a7e58556f326ecb56fbb53b92/"
    },
    "refs/tags/v4.5-rc2": {
      "target": "aa73385a80d39b9db114d164f9b958660d4d0f7a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/aa73385a80d39b9db114d164f9b958660d4d0f7a/"
    },
    "refs/tags/v4.5-rc3": {
      "target": "2a458cef29af1ce7bb087e72ae2fdbe4c5ecc2dc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2a458cef29af1ce7bb087e72ae2fdbe4c5ecc2dc/"
    },
    "refs/tags/v4.5-rc4": {
      "target": "b726a6ccc1f4d80f1dcdb7ec93c466027dc9f6a8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b726a6ccc1f4d80f1dcdb7ec93c466027dc9f6a8/"
    },
    "refs/tags/v4.5-rc5": {
      "target": "e91f74b3b3c3aa2b9fc7720ff24ca63fefe467b4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e91f74b3b3c3aa2b9fc7720ff24ca63fefe467b4/"
    },
    "refs/tags/v4.5-rc6": {
      "target": "527cae8caa26632d8355074027a03af92931ce07",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/527cae8caa26632d8355074027a03af92931ce07/"
    },
    "refs/tags/v4.5-rc7": {
      "target": "4dbe92d445c77fed92d4bb1b655ea52c61205b78",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4dbe92d445c77fed92d4bb1b655ea52c61205b78/"
    },
    "refs/tags/v4.6": {
      "target": "a902836aade7b56f8dc31f50f9c6ca6d0fa01e45",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a902836aade7b56f8dc31f50f9c6ca6d0fa01e45/"
    },
    "refs/tags/v4.6-rc1": {
      "target": "5c1930df65df4be8312d705037ac9a8a33cc2a31",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5c1930df65df4be8312d705037ac9a8a33cc2a31/"
    },
    "refs/tags/v4.6-rc2": {
      "target": "66ddb3c4430579382992b5226ebd0f893707c08d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/66ddb3c4430579382992b5226ebd0f893707c08d/"
    },
    "refs/tags/v4.6-rc3": {
      "target": "2073bfeaa5ffaf82c23112898bfcfff89199a67d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2073bfeaa5ffaf82c23112898bfcfff89199a67d/"
    },
    "refs/tags/v4.6-rc4": {
      "target": "35c0e46ca81f4d7d67f7b33bb62de42d43c46cb0",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/35c0e46ca81f4d7d67f7b33bb62de42d43c46cb0/"
    },
    "refs/tags/v4.6-rc5": {
      "target": "8ef3ad9a813abdf0817acb0b2be30be70bf25a9b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/8ef3ad9a813abdf0817acb0b2be30be70bf25a9b/"
    },
    "refs/tags/v4.6-rc6": {
      "target": "e23626e59033cf3d43c646c4f3df426dcea5bead",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e23626e59033cf3d43c646c4f3df426dcea5bead/"
    },
    "refs/tags/v4.6-rc7": {
      "target": "edff27137675622d8e2e6597eb3c7381e6ffb664",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/edff27137675622d8e2e6597eb3c7381e6ffb664/"
    },
    "refs/tags/v4.7": {
      "target": "5563368c573d2a245d14737b9ff32dcdee4126dc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5563368c573d2a245d14737b9ff32dcdee4126dc/"
    },
    "refs/tags/v4.7-rc1": {
      "target": "2fa60a513cc5f5007310895202d1cf708610d2ee",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2fa60a513cc5f5007310895202d1cf708610d2ee/"
    },
    "refs/tags/v4.7-rc2": {
      "target": "f4b955c0f6674ed5294eef53b3d79801f0bfc466",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f4b955c0f6674ed5294eef53b3d79801f0bfc466/"
    },
    "refs/tags/v4.7-rc3": {
      "target": "a0b42bf4b0397d9cec150ac1d353f028d7b45e60",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a0b42bf4b0397d9cec150ac1d353f028d7b45e60/"
    },
    "refs/tags/v4.7-rc4": {
      "target": "4c3cf21c98a95873f3f6a267ee48527d255a84f7",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4c3cf21c98a95873f3f6a267ee48527d255a84f7/"
    },
    "refs/tags/v4.7-rc5": {
      "target": "284561c2e0216e12c426ea1db27b2a30189e729c",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/284561c2e0216e12c426ea1db27b2a30189e729c/"
    },
    "refs/tags/v4.7-rc6": {
      "target": "b659fd7eb2387752f2742c9690b0146cd5188d63",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b659fd7eb2387752f2742c9690b0146cd5188d63/"
    },
    "refs/tags/v4.7-rc7": {
      "target": "2a8653d68222a365505e2bc7413aa8a340cfc199",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2a8653d68222a365505e2bc7413aa8a340cfc199/"
    },
    "refs/tags/v4.8": {
      "target": "ddf6b1a7de691080a203e6ce1dff4331e1770825",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ddf6b1a7de691080a203e6ce1dff4331e1770825/"
    },
    "refs/tags/v4.8-rc1": {
      "target": "b7d4dff9dcaabd98371bbc6d4f171bf8712f2724",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b7d4dff9dcaabd98371bbc6d4f171bf8712f2724/"
    },
    "refs/tags/v4.8-rc2": {
      "target": "427b3a97331fba24e493cf2b9ed8c0e9748296f8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/427b3a97331fba24e493cf2b9ed8c0e9748296f8/"
    },
    "refs/tags/v4.8-rc3": {
      "target": "ea79b5025467abc3c0441103b29f89c5ee81f174",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ea79b5025467abc3c0441103b29f89c5ee81f174/"
    },
    "refs/tags/v4.8-rc4": {
      "target": "bc6679d273e0527adc36c295f71360be4feb40b5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/bc6679d273e0527adc36c295f71360be4feb40b5/"
    },
    "refs/tags/v4.8-rc5": {
      "target": "d03ec7f4fa91aab68d1744ca8be5be98e2e03b11",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d03ec7f4fa91aab68d1744ca8be5be98e2e03b11/"
    },
    "refs/tags/v4.8-rc6": {
      "target": "b77977e498a9cf49f86103b34f8f90d8b02bca0f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b77977e498a9cf49f86103b34f8f90d8b02bca0f/"
    },
    "refs/tags/v4.8-rc7": {
      "target": "ca855d9962cdcdccbfee79f734dec82d92576d9a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ca855d9962cdcdccbfee79f734dec82d92576d9a/"
    },
    "refs/tags/v4.8-rc8": {
      "target": "3f184b22bf0241b43cc46b16d5ede18ea2a5eef8",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3f184b22bf0241b43cc46b16d5ede18ea2a5eef8/"
    },
    "refs/tags/v4.9": {
      "target": "b05b1307892247c76ad6dc3f59921736279a347b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b05b1307892247c76ad6dc3f59921736279a347b/"
    },
    "refs/tags/v4.9-rc1": {
      "target": "27eb48d13e6981d9ada2c74f50a364a4414f73b9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/27eb48d13e6981d9ada2c74f50a364a4414f73b9/"
    },
    "refs/tags/v4.9-rc2": {
      "target": "f068b3f31cc8acb6e37785d15170dd9e95c7d161",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f068b3f31cc8acb6e37785d15170dd9e95c7d161/"
    },
    "refs/tags/v4.9-rc3": {
      "target": "44525af32f277001a751a51d59de12c5a733814f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/44525af32f277001a751a51d59de12c5a733814f/"
    },
    "refs/tags/v4.9-rc4": {
      "target": "f0b8a0182b2f3a7139d8d041e06bc11c1ba6e912",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/f0b8a0182b2f3a7139d8d041e06bc11c1ba6e912/"
    },
    "refs/tags/v4.9-rc5": {
      "target": "9d4d32609e7bb04693ab7bf313cd5856b5298c1f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9d4d32609e7bb04693ab7bf313cd5856b5298c1f/"
    },
    "refs/tags/v4.9-rc6": {
      "target": "edc31677f4f97b582fbc53659337dddb210a22fa",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/edc31677f4f97b582fbc53659337dddb210a22fa/"
    },
    "refs/tags/v4.9-rc7": {
      "target": "512983f23bc1346e6dbaf0b899257332659d9f5d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/512983f23bc1346e6dbaf0b899257332659d9f5d/"
    },
    "refs/tags/v4.9-rc8": {
      "target": "a9b7e3f1eada90250a6b2ab2ef3e0a846cb16831",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a9b7e3f1eada90250a6b2ab2ef3e0a846cb16831/"
    },
    "refs/tags/v5.0": {
      "target": "cfb40dbadb669988d32ec0f4f716f541902c7895",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cfb40dbadb669988d32ec0f4f716f541902c7895/"
    },
    "refs/tags/v5.0-rc1": {
      "target": "0dd18ffb54df420d2c7d25bc5259942dbbc2a610",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0dd18ffb54df420d2c7d25bc5259942dbbc2a610/"
    },
    "refs/tags/v5.0-rc2": {
      "target": "6f38b5d6cfd43dde3058a10c68baae9cf17af912",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6f38b5d6cfd43dde3058a10c68baae9cf17af912/"
    },
    "refs/tags/v5.0-rc3": {
      "target": "0c356ad876fa1d72f08a9eef059e6d6ebe3a6e4d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0c356ad876fa1d72f08a9eef059e6d6ebe3a6e4d/"
    },
    "refs/tags/v5.0-rc4": {
      "target": "9a57b893babc8ea26e887dd82d5d4b63c0dea9a4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9a57b893babc8ea26e887dd82d5d4b63c0dea9a4/"
    },
    "refs/tags/v5.0-rc5": {
      "target": "327a5e79e6eb0ef77bd4b808e1a4a282d1644d23",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/327a5e79e6eb0ef77bd4b808e1a4a282d1644d23/"
    },
    "refs/tags/v5.0-rc6": {
      "target": "5c9952325da3c4f39e0b117b91b7acef241c3afc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/5c9952325da3c4f39e0b117b91b7acef241c3afc/"
    },
    "refs/tags/v5.0-rc7": {
      "target": "d4588d950276a68c2f52bed31d6fb3e564f87919",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d4588d950276a68c2f52bed31d6fb3e564f87919/"
    },
    "refs/tags/v5.0-rc8": {
      "target": "497dce527948a0e9e471ddf45c66d542828c3561",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/497dce527948a0e9e471ddf45c66d542828c3561/"
    },
    "refs/tags/v5.1": {
      "target": "b0854f56c54832584b1173b997fa665328b50afd",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/b0854f56c54832584b1173b997fa665328b50afd/"
    },
    "refs/tags/v5.1-rc1": {
      "target": "0e8ffd573ccccb680aacabf14ca3282b60ab57dc",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/0e8ffd573ccccb680aacabf14ca3282b60ab57dc/"
    },
    "refs/tags/v5.1-rc2": {
      "target": "046cca6fd4d3017c1dc7463d795c706ca09724a6",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/046cca6fd4d3017c1dc7463d795c706ca09724a6/"
    },
    "refs/tags/v5.1-rc3": {
      "target": "e40bc7462f998035ea03e93daab2e5b27ed22149",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e40bc7462f998035ea03e93daab2e5b27ed22149/"
    },
    "refs/tags/v5.1-rc4": {
      "target": "1f3a6c30348e2205f75969e38c02f7d211fa01a9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/1f3a6c30348e2205f75969e38c02f7d211fa01a9/"
    },
    "refs/tags/v5.1-rc5": {
      "target": "e55675cb6f80500a8aa6db808643f8b66a84bd38",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e55675cb6f80500a8aa6db808643f8b66a84bd38/"
    },
    "refs/tags/v5.1-rc6": {
      "target": "cd9f1e5b687b85a321f0a9c0bbf95eaf9b8a6b3d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cd9f1e5b687b85a321f0a9c0bbf95eaf9b8a6b3d/"
    },
    "refs/tags/v5.1-rc7": {
      "target": "460ea0a56a22ae104fa81a09cd5e9e510cdc4d71",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/460ea0a56a22ae104fa81a09cd5e9e510cdc4d71/"
    },
    "refs/tags/v5.2": {
      "target": "14817899749faa879ecbeda203c73138c4d65e79",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/14817899749faa879ecbeda203c73138c4d65e79/"
    },
    "refs/tags/v5.2-rc1": {
      "target": "311e1ea1b63dc428575270b9d4801e3e6f96457b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/311e1ea1b63dc428575270b9d4801e3e6f96457b/"
    },
    "refs/tags/v5.2-rc2": {
      "target": "ba4529a907755b8810bec37ba26a43fe8d9ac1b4",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ba4529a907755b8810bec37ba26a43fe8d9ac1b4/"
    },
    "refs/tags/v5.2-rc3": {
      "target": "98af7fc496808efe10b304e6c817ea300f8fb56b",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/98af7fc496808efe10b304e6c817ea300f8fb56b/"
    },
    "refs/tags/v5.2-rc4": {
      "target": "a09f513e2ab90f4f5636f024dd2222763b98a5b1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a09f513e2ab90f4f5636f024dd2222763b98a5b1/"
    },
    "refs/tags/v5.2-rc5": {
      "target": "34781703e15434b965474c4b669b2eb603f3778f",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/34781703e15434b965474c4b669b2eb603f3778f/"
    },
    "refs/tags/v5.2-rc6": {
      "target": "ef2678862915da4aa7bb2dd8c9608859b358fc9a",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ef2678862915da4aa7bb2dd8c9608859b358fc9a/"
    },
    "refs/tags/v5.2-rc7": {
      "target": "dde2fa21d081aa5152511f1b70d53dfa0bdc0556",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/dde2fa21d081aa5152511f1b70d53dfa0bdc0556/"
    },
    "refs/tags/v5.3": {
      "target": "4dba490412e7f6c9f17a0afcf7b08f110817b004",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4dba490412e7f6c9f17a0afcf7b08f110817b004/"
    },
    "refs/tags/v5.3-rc1": {
      "target": "cc8d7c0499f1b15ae71a54981dd1b441888b0e08",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/cc8d7c0499f1b15ae71a54981dd1b441888b0e08/"
    },
    "refs/tags/v5.3-rc2": {
      "target": "d450c62c027d1e6c171e3ac86011c226e396d2cf",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/d450c62c027d1e6c171e3ac86011c226e396d2cf/"
    },
    "refs/tags/v5.3-rc3": {
      "target": "16a223436c28816262794acf683e095965607762",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/16a223436c28816262794acf683e095965607762/"
    },
    "refs/tags/v5.3-rc4": {
      "target": "4ca8eeb2a6203f252a8572aa640b88cd9736ccef",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/4ca8eeb2a6203f252a8572aa640b88cd9736ccef/"
    },
    "refs/tags/v5.3-rc5": {
      "target": "88f8355483766c4361228939b5a4d3aaab48f9a6",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/88f8355483766c4361228939b5a4d3aaab48f9a6/"
    },
    "refs/tags/v5.3-rc6": {
      "target": "add840e44afed5b308140f6d3383921335ce0a59",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/add840e44afed5b308140f6d3383921335ce0a59/"
    },
    "refs/tags/v5.3-rc7": {
      "target": "980d0783eab87f4dea7f03737641ead5121ec738",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/980d0783eab87f4dea7f03737641ead5121ec738/"
    },
    "refs/tags/v5.3-rc8": {
      "target": "ed9a681803ae0aac98b7932541dde8f17e3da543",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ed9a681803ae0aac98b7932541dde8f17e3da543/"
    },
    "refs/tags/v5.4": {
      "target": "6e815efe19a99a33b16cc720c3d3a727565a4fa1",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/6e815efe19a99a33b16cc720c3d3a727565a4fa1/"
    },
    "refs/tags/v5.4-rc1": {
      "target": "24cb7d7282e06bdeced7f657dff650d3e02f5716",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/24cb7d7282e06bdeced7f657dff650d3e02f5716/"
    },
    "refs/tags/v5.4-rc2": {
      "target": "a09a4100c0de95f0c617c7ed2f4b8b45271cd4d5",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a09a4100c0de95f0c617c7ed2f4b8b45271cd4d5/"
    },
    "refs/tags/v5.4-rc3": {
      "target": "3ad6a0c75d5b889c3c010429d4a7c985ff7a0da9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3ad6a0c75d5b889c3c010429d4a7c985ff7a0da9/"
    },
    "refs/tags/v5.4-rc4": {
      "target": "ec0c8d6258a65ee84f63cd0507aa444f54507d54",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ec0c8d6258a65ee84f63cd0507aa444f54507d54/"
    },
    "refs/tags/v5.4-rc5": {
      "target": "9e8d1fa52a51b42513fcba3b8aab4fa36d234e99",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9e8d1fa52a51b42513fcba3b8aab4fa36d234e99/"
    },
    "refs/tags/v5.4-rc6": {
      "target": "52bed4561efc5aa3ddeda5e62d083f398c217335",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/52bed4561efc5aa3ddeda5e62d083f398c217335/"
    },
    "refs/tags/v5.4-rc7": {
      "target": "368e906473dfe965c421677be67b4b1297eed6c9",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/368e906473dfe965c421677be67b4b1297eed6c9/"
    },
    "refs/tags/v5.4-rc8": {
      "target": "9b2a6e34b876d9c71fd0feec0e7ee07031bc1902",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/9b2a6e34b876d9c71fd0feec0e7ee07031bc1902/"
    },
    "refs/tags/v5.5-rc1": {
      "target": "3e4fa608e7227f8bde448547589146f40dd46081",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3e4fa608e7227f8bde448547589146f40dd46081/"
    },
    "refs/tags/v5.5-rc2": {
      "target": "ff88a16e78d46895b4ccb535462233fe99b17d21",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/ff88a16e78d46895b4ccb535462233fe99b17d21/"
    },
    "refs/tags/v5.5-rc3": {
      "target": "3da0952890f1bb3898c72abdafbf2d58ae15193d",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/3da0952890f1bb3898c72abdafbf2d58ae15193d/"
    },
    "refs/tags/v5.5-rc4": {
      "target": "31724e5503a9ffd71c440c31a1e3d6a50bd50186",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/31724e5503a9ffd71c440c31a1e3d6a50bd50186/"
    },
    "refs/tags/v5.5-rc5": {
      "target": "a6e9f2d278c7f7a7a7870f9ac7bc9662df8aa551",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/a6e9f2d278c7f7a7a7870f9ac7bc9662df8aa551/"
    },
    "refs/tags/v5.5-rc6": {
      "target": "e321afedf55565f8e6e4e2069c4046e5905a4b29",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/e321afedf55565f8e6e4e2069c4046e5905a4b29/"
    },
    "refs/tags/v5.5-rc7": {
      "target": "2bf590c16af975a0132e4a90967807d538cdc7de",
      "target_type": "release",
      "target_url": "https://archive.softwareheritage.org/api/1/release/2bf590c16af975a0132e4a90967807d538cdc7de/"
    }
  },
  "next_branch": null
}
    """,  # NoQA: B950  # NoQA: B950
    "origin/https://github.com/NixOS/nixpkgs/visits/?last_visit=50&per_page=10": r"""
[
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 49,
    "date": "2018-07-31T04:34:23.298931+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/49/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 48,
    "date": "2018-07-27T10:12:23.069912+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/48/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 47,
    "date": "2018-07-24T15:34:14.433021+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/47/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 46,
    "date": "2018-07-23T22:07:33.025195+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/46/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 45,
    "date": "2018-07-21T16:50:41.944584+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/45/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 44,
    "date": "2018-07-21T12:17:24.044885+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/44/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 43,
    "date": "2018-07-20T13:58:53.926748+00:00",
    "type": "git",
    "status": "full",
    "snapshot": "c5d63cd27b09068b1f6913a90dfe44f1276091c2",
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/43/",
    "snapshot_url": "https://archive.softwareheritage.org/api/1/snapshot/c5d63cd27b09068b1f6913a90dfe44f1276091c2/"
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 42,
    "date": "2018-07-20T00:38:20.356513+00:00",
    "type": "git",
    "status": "full",
    "snapshot": "456550ea74af4e2eecaa406629efaaf0b9b5f976",
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/42/",
    "snapshot_url": "https://archive.softwareheritage.org/api/1/snapshot/456550ea74af4e2eecaa406629efaaf0b9b5f976/"
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 41,
    "date": "2018-07-18T12:42:33.408489+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/41/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 40,
    "date": "2018-07-17T16:40:42.302353+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/40/",
    "snapshot_url": null
  }
]
    """,  # NoQA: B950  # NoQA: B950
    "origin/https://github.com/NixOS/nixpkgs/visits/?last_visit=40&per_page=10": r"""
[
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 39,
    "date": "2018-07-16T23:24:15.061318+00:00",
    "type": "git",
    "status": "full",
    "snapshot": "1c2d51e9cda9958736394454ce8875e5913a1925",
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/39/",
    "snapshot_url": "https://archive.softwareheritage.org/api/1/snapshot/1c2d51e9cda9958736394454ce8875e5913a1925/"
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 38,
    "date": "2018-07-16T10:43:14.651509+00:00",
    "type": "git",
    "status": "full",
    "snapshot": "8fc8f205ba2d16a798b85467bd9f3cc6b938cc5c",
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/38/",
    "snapshot_url": "https://archive.softwareheritage.org/api/1/snapshot/8fc8f205ba2d16a798b85467bd9f3cc6b938cc5c/"
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 37,
    "date": "2018-07-15T18:45:41.744499+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/37/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 36,
    "date": "2018-07-15T10:41:05.001029+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/36/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 35,
    "date": "2018-07-14T10:42:58.232545+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/35/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 34,
    "date": "2018-07-14T05:31:45.565907+00:00",
    "type": "git",
    "status": "full",
    "snapshot": "588fa42fefd27047474072da57ac2c83c3a481a7",
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/34/",
    "snapshot_url": "https://archive.softwareheritage.org/api/1/snapshot/588fa42fefd27047474072da57ac2c83c3a481a7/"
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 33,
    "date": "2018-07-12T22:17:10.133596+00:00",
    "type": "git",
    "status": "full",
    "snapshot": "9f9c7b2d52038b1d1fb6b1ff7387f0446c3f9da0",
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/33/",
    "snapshot_url": "https://archive.softwareheritage.org/api/1/snapshot/9f9c7b2d52038b1d1fb6b1ff7387f0446c3f9da0/"
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 32,
    "date": "2018-07-12T14:45:54.291987+00:00",
    "type": "git",
    "status": "partial",
    "snapshot": null,
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/32/",
    "snapshot_url": null
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 31,
    "date": "2018-07-11T18:30:43.843317+00:00",
    "type": "git",
    "status": "full",
    "snapshot": "3d4c9c5c9a7de4d174b8a02d23bf4a55b2d3a911",
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/31/",
    "snapshot_url": "https://archive.softwareheritage.org/api/1/snapshot/3d4c9c5c9a7de4d174b8a02d23bf4a55b2d3a911/"
  },
  {
    "origin": "https://github.com/NixOS/nixpkgs",
    "visit": 30,
    "date": "2018-07-10T07:37:24.991560+00:00",
    "type": "git",
    "status": "full",
    "snapshot": "100de51846f317e6ab48da79d985cefa6fdefe42",
    "origin_visit_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/visit/30/",
    "snapshot_url": "https://archive.softwareheritage.org/api/1/snapshot/100de51846f317e6ab48da79d985cefa6fdefe42/"
  }
]
    """,  # NoQA: B950  # NoQA: B950
    "origin/https://github.com/NixOS/nixpkgs/visit/latest/": r"""
{
  "origin": "https://github.com/NixOS/nixpkgs",
  "visit": 128,
  "date": "2021-09-02T20:20:31.231786+00:00",
  "status": "full",
  "snapshot": "6e1fe7858066ff1a6905080ac6503a3a12b84f59",
  "type": "git",
  "metadata": {},
  "origin_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/NixOS/nixpkgs/get/",
  "snapshot_url": "https://archive.softwareheritage.org/api/1/snapshot/6e1fe7858066ff1a6905080ac6503a3a12b84f59/"
}
    """,  # NoQA: B950  # NoQA: B950
    "origin/search/foo%20bar%20baz%20qux/?with_visit=true": r"""
[
  {
    "url": "https://github.com/foo-bar-baz-qux/mygithubpage",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/mygithubpage/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/foo-bar-baz-qux.github.io",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/foo-bar-baz-qux.github.io/visits/"
  },
  {
    "url": "https://github.com/tunnckoCore/foo-bar-baz-qux",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/tunnckoCore/foo-bar-baz-qux/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/aml-project",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/aml-project/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/ci_test",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/ci_test/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/extreme-computing",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/extreme-computing/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/cs205-homework",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/cs205-homework/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/rstan",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/rstan/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/cs207",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/cs207/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/2015lab1",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/2015lab1/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/stan",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/stan/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/atom-script",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/atom-script/visits/"
  },
  {
    "url": "https://github.com/foobarbazquxquux/gordon-test",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foobarbazquxquux/gordon-test/visits/"
  },
  {
    "url": "https://github.com/foobarbazquxquux/helios",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foobarbazquxquux/helios/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/2016",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/2016/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/devtools",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/devtools/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/dplyr",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/dplyr/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/paletter",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/paletter/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/ggplot2",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/ggplot2/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/stm",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/stm/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/concept-to-clinic",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/concept-to-clinic/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/scales",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/scales/visits/"
  },
  {
    "url": "https://www.npmjs.com/package/foo-bar-baz-qux",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://www.npmjs.com/package/foo-bar-baz-qux/visits/"
  },
  {
    "url": "https://bitbucket.org/foobarbazqux/rp.git",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://bitbucket.org/foobarbazqux/rp.git/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/knowledge-repo",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/knowledge-repo/visits/"
  },
  {
    "url": "https://github.com/foo-bar-baz-qux/EconML",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/foo-bar-baz-qux/EconML/visits/"
  }
]
    """,  # NoQA: B950  # NoQA: B950
    "origin/search/python/?limit=5": r"""
[
  {
    "url": "https://github.com/neon670/python.dev",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/neon670/python.dev/visits/"
  },
  {
    "url": "https://github.com/aur-archive/python-werkzeug",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/aur-archive/python-werkzeug/visits/"
  },
  {
    "url": "https://github.com/jsagon/jtradutor-web-python",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/jsagon/jtradutor-web-python/visits/"
  },
  {
    "url": "https://github.com/zjmwqx/ipythonCode",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/zjmwqx/ipythonCode/visits/"
  },
  {
    "url": "https://github.com/knutab/Python-BSM",
    "origin_visits_url": "https://archive.softwareheritage.org/api/1/origin/https://github.com/knutab/Python-BSM/visits/"
  }
]
    """,  # NoQA: B950  # NoQA: B950
}
